/* bwMonitor.c - rate based bwMonitor */

/*
modification history
--------------------
01b,17nov09,suh fixed spr15980 weighted load-balancing issues
01a,28sep08,sam created
*/

/*
DESCRIPTION
This module uses the kernel rate estimator for rate determination. The rate
estimators must have been installed in the packet path using the netfilter
target RATEEST.

This module implements a rate based next hop selection algorithm for multiapth
routing. This selection is based on the initial weights given by the user and
the current traffic load on each of its nexthop links in a multipath route.

The NXTHOP_LOAD_TOLERANCE is the % tolerance of the load-balancer. This defaults
to 10.
*/

#include "linux/module.h"
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/netdevice.h>
#include <linux/gen_stats.h>
#include <net/route.h>
#include <net/ip_fib.h>
#include  <net/netfilter/xt_rateest.h>

#define CONFIG_BW_MONITOR 1

#define NXTHOP_LOAD_TOLERANCE   80
#define MAX_BANDWIDTH 1024

/* link states */
#define NXTHOP_LINK_OVERUSED     1
#define NXTHOP_LINK_UNDERUSED    2
#define NXTHOP_LINK_NORMALUSED   3
#define NXTHOP_LINK_MAX          4

#undef BWMON_DEBUG
#ifdef BWMON_DEBUG
#define BWMON_PRINTK(a...)     do { if (debug) printk(KERN_DEBUG a); } while (0)
#else
#define BWMON_PRINTK(a...)
#endif

extern int (*bwmon_multipath_weight_check) (const struct flowi *flp,
                                             struct fib_result *res);

static int bwMonRateInterval = 2;
struct timer_list bwRateTimer;
static int lastHopSelected=0;
static int multipathSelect = 0;
static int weightedHopPrefer = 0;
static char weightedHop[16] = "";
static int loadTolerance = NXTHOP_LOAD_TOLERANCE;
static int enable_debug_msgs = 0;
static int maxBandwidth = 1024 ;
static int usableBwLimit = (NXTHOP_LOAD_TOLERANCE * (MAX_BANDWIDTH )) / 100;

#ifdef CONFIG_SYSCTL
static struct ctl_table_header * pBwMonitorSysctlHeader;
#endif /* CONFIG_SYSCTL */
#ifdef BWMON_DEBUG
static int debug = 0;

/*******************************************************************************
*
* bwMonTimerHandler - timer handler for bw monitor.
*
* This function calculates the data rate of all the interfaces registered.
*
* RETURNS: 0
*/

void  bwMonTimerHandler
    (
    unsigned long  arg
    )
    {
    struct net_device * pNetDevice=NULL;
    struct net_device_stats * pNetStats=NULL;
    unsigned long tx_rate =0;
    unsigned long rx_rate = 0;
    char rxName[16]={0};
    char txName[16]={0};

    /* loop through the list of devices and calculate the rate of each one */
	for (pNetDevice = dev_base; pNetDevice; pNetDevice = pNetDevice->next)
        {
        struct xt_rateest *est1=NULL;

        tx_rate = 0;
        rx_rate = 0;

        /* continue if the interface is down */
        if ((pNetDevice->flags & IFF_UP) == 0)
            continue;

        /* if the interface does not have statistics counter, then continue */
	    if (pNetDevice->get_stats == NULL)
            continue;

		pNetStats = pNetDevice->get_stats(pNetDevice);
        if (pNetStats == NULL)
            continue;

        snprintf(txName, sizeof(txName), "tx_%s", pNetDevice->name);
        est1 = xt_rateest_lookup (txName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            tx_rate = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        snprintf(rxName, sizeof(rxName), "rx_%s", pNetDevice->name);
        est1 = xt_rateest_lookup (rxName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            rx_rate = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        BWMON_PRINTK ("RATEEST: if:%s rx:%lu tx:%lu total: %lu\n",
                      pNetDevice->name, rx_rate, tx_rate, tx_rate + rx_rate);

        }

    /* reset the timer */
    del_timer(&bwRateTimer);
    bwRateTimer.expires = jiffies + bwMonRateInterval * HZ;
	add_timer(&bwRateTimer);

    return;
    }
#endif /* BWMON_DEBUG */

/*******************************************************************************
*
* bwMonMultipathNxtHopSelect - selects the next-hop for a multipath route.
*
* This function selects the nexthop based on the current linkusage.
*
* RETURNS: 0
*/

int bwMonMultipathNxtHopSelect
    (
    const struct flowi * pFlow,       /* traffic description */
    struct fib_result *  pFib         /* multipath route */
    )
    {
    struct fib_nh * pNxtHop=NULL;
    int    currHop=0;
    int    selHop=0;
    unsigned long tx_bps = 0;
    unsigned long rx_bps = 0;
    char rxName[16]={0};
    char txName[16]={0};

    if(enable_debug_msgs)
        printk("bwMonMultipathNxtHopSelect:: checking rates \n");

    for (currHop=0; currHop < pFib->fi->fib_nhs; currHop++)
        {
        struct xt_rateest *est1=NULL;

        memset (txName, '\0', sizeof(txName));
        memset (rxName, '\0', sizeof(rxName));

        pNxtHop = &pFib->fi->fib_nh[currHop];
        snprintf(txName, sizeof(txName), "tx_%s", pNxtHop->nh_dev->name);
        est1 = xt_rateest_lookup (txName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            tx_bps = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        snprintf(rxName, sizeof(rxName), "rx_%s", pNxtHop->nh_dev->name);
        est1 = xt_rateest_lookup (rxName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            rx_bps = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        pNxtHop->currBwShare = tx_bps + rx_bps;

        if((currHop == weightedHopPrefer) && (pNxtHop->currBwShare <= usableBwLimit))
           {
           selHop = currHop;

           if(enable_debug_msgs == 2)
              {
              printk("hop :%d dev:%s usableBwLimit = %d currBwShare = %d lastHopSelected = %d weightedHopPrefer = %d \n",
              currHop, pNxtHop->nh_dev->name, usableBwLimit, pNxtHop->currBwShare,lastHopSelected,weightedHopPrefer);
              }

           if(enable_debug_msgs)
              printk("1. selecting hop: %d lastHopSelected = %d \n ", selHop, lastHopSelected );
           break;
           }
        else
           {
            if(enable_debug_msgs == 2)
              printk("4. hop :%d dev:%s usableBwLimit = %d currBwShare = %d lastHopSelected = %d weightedHopPrefer = %d \n",
              currHop, pNxtHop->nh_dev->name, usableBwLimit, pNxtHop->currBwShare,lastHopSelected,weightedHopPrefer);
           }

	if(currHop == (pFib->fi->fib_nhs - 1))
            {
            if(currHop == weightedHopPrefer)
                {
                selHop = 0;
                if(enable_debug_msgs)
                   printk("2. selecting hop: %d lastHopSelected = %d \n ", selHop, lastHopSelected );
                }
            else
                {
                selHop = currHop;
                if(enable_debug_msgs)
                   printk("3. selecting hop: %d lastHopSelected = %d \n ", selHop, lastHopSelected );
                }
            }
        }

    pFib->nh_sel = selHop;
    lastHopSelected = selHop;

    /* update the timestamp of the selected hop */
    pNxtHop = &pFib->fi->fib_nh[selHop];
    pNxtHop->nh_timestamp = jiffies;


    return 0;
    }

#ifdef CONFIG_SYSCTL

/*******************************************************************************
*
* procMultipathSelect - handler for /proc/sys/net/bwmonitor/multipath_select
*
* This function installs/uninstalls the multipath selection hook.
*
* RETURNS: 0
*/
static int procMultipathSelect
    (
    ctl_table   * pTable,
    int           write,
    struct file * pFile,
    void __user * pBuffer,
    size_t      * pLen,
    loff_t      * pPos
    )
    {
    int result;

    result = proc_dointvec (pTable, write, pFile, pBuffer, pLen, pPos);

    if (write && result == 0)
        {
        if (multipathSelect)
            {
            /* Install the hook */
            bwmon_multipath_weight_check = bwMonMultipathNxtHopSelect;
            BWMON_PRINTK ("bwMonitor multipath selection enabled multipathSelect = %d \n", multipathSelect);
            printk ("bwMonitor multipath selection enabled \n");
            }
        else
            {
            /* Uninstall the hook */
            bwmon_multipath_weight_check = NULL;
            BWMON_PRINTK ("bwMonitor multipath selection disabled\n");
            printk("bwMonitor multipath selection disabled\n");
            }
        }
    return result;
    }

static int procHopPreferSelect
    (
    ctl_table   * pTable,
    int           write,
    struct file * pFile,
    void __user * pBuffer,
    size_t      * pLen,
    loff_t      * pPos
    )
    {
    int result;

    result = proc_dointvec (pTable, write, pFile, pBuffer, pLen, pPos);

    if (write && result == 0)
        {
        if (weightedHopPrefer)
		printk("weightedHopPrefer set to %d \n",weightedHopPrefer);
	}

    return result;
    }


static int procSetLoadTolerance
    (
    ctl_table   * pTable,
    int           write,
    struct file * pFile,
    void __user * pBuffer,
    size_t      * pLen,
    loff_t      * pPos
    )
    {
    int result;

    result = proc_dointvec (pTable, write, pFile, pBuffer, pLen, pPos);

    if (write && result == 0)
        {
        if ((loadTolerance) && ((loadTolerance > 0) && (loadTolerance <= 100) ))
	   {
                usableBwLimit = ((loadTolerance * maxBandwidth ) / 100 );
	   }
	}

    return result;
    }

static ctl_table bwMonitorSysctlTable[] =
    {
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "multipath_select",
        .mode     = 0644,
        .data     = &multipathSelect,
        .maxlen   = sizeof (int),
        .proc_handler = procMultipathSelect,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "weighted_hop_prefer",
        .mode     = 0644,
/*
        .data     = weightedHop,
*/
        .data     = &weightedHopPrefer,
        .maxlen   = sizeof (int),
        .proc_handler = procHopPreferSelect,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "load_tolerance",
        .mode     = 0644,
        .data     = &loadTolerance ,
        .maxlen   = sizeof (int),
        .proc_handler = procSetLoadTolerance,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "debugon",
        .mode     = 0644,
        .data     = &enable_debug_msgs,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "maxBandwidth",
        .mode     = 0644,
        .data     = &maxBandwidth,
        .maxlen   = sizeof (int),
        .proc_handler = procSetLoadTolerance,
        },
#ifdef BWMON_DEBUG
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "debug",
        .mode     = 0644,
        .data     = &debug,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "rate_interval",
        .mode     = 0644,
        .data     = &bwMonRateInterval,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
#endif /* BWMON_DEBUG */
        { 0 }
    };

static ctl_table bwMonitorNetTable[] =
    {
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "bwmonitor",
        .mode = 0555,
        .child = bwMonitorSysctlTable
        },
        { 0 }
    };

static ctl_table bwMonitorRootTable[] =
    {
        {
        .ctl_name = CTL_NET,
        .procname = "net",
        .mode = 0555,
        .child = bwMonitorNetTable
        },
        { 0 }
    };

/*******************************************************************************
*
* bwMonitorSysctlRegister - registers sysctl entries
*
* This routine registers net.bwmonitor sysctl entries.
*
* RETURNS: 0 or -1
*/
static __init int bwMonitorSysctlRegister (void)
    {
    pBwMonitorSysctlHeader = register_sysctl_table (bwMonitorRootTable);
    if (pBwMonitorSysctlHeader == NULL)
        {
        BWMON_PRINTK ("bwMonitor sysctl registration failed\n");
        printk("bwMonitor sysctl registration failed\n");
        return -1;
        }

    BWMON_PRINTK ("bwMonitor sysctl registered\n");
    printk("bwMonitor sysctl registered\n");
    return 0;
    }

/*******************************************************************************
*
* bwMonitorSysctlRegister - registers sysctl entries
*
* This routine unregisters net.bwmonitor sysctl entries.
*
* RETURNS: 0 or -1
*/
static __exit int bwMonitorSysctlUnregister (void)
    {
    if (pBwMonitorSysctlHeader == NULL)
        {
        BWMON_PRINTK ("bwMonitor sysctl not registered\n");
        printk("bwMonitor sysctl not registered\n");
        return -1;
        }

    unregister_sysctl_table (pBwMonitorSysctlHeader);
    pBwMonitorSysctlHeader = NULL;

    BWMON_PRINTK ("Unregistered bwMonitor sysctl\n");
    printk("Unregistered bwMonitor sysctl\n");
    return 0;
    }
#endif /* CONFIG_SYSCTL */

/*******************************************************************************
*
* bwMonitorInit - initializes bandwidth monitor
*
* This routine initializes bandwidth monitor.
*
* RETURNS: 0 or -1
*/
int __init bwMonitorInit (void)
    {
#ifdef BWMON_DEBUG
    /* initialize the timer to calculate traffic rate */
	init_timer(&bwRateTimer);
	bwRateTimer.data = 0;
	bwRateTimer.function = bwMonTimerHandler;
	bwRateTimer.expires = jiffies + bwMonRateInterval * HZ;
	add_timer(&bwRateTimer);
#endif /* BWMON_DEBUG */

        if (multipathSelect)
            {
            bwmon_multipath_weight_check = bwMonMultipathNxtHopSelect;
            }

#ifdef CONFIG_SYSCTL
    bwMonitorSysctlRegister();

    printk("CONFIG_SYSCTL enabled ...\n");

#endif /* CONFIG_SYSCTL */
    BWMON_PRINTK ("Initialized bandwidth monitor ...\n");

    printk("Initialized bandwidth monitor ...\n");

    return (0);
    }

/*******************************************************************************
*
* bwMonitorExit - de-initalize the bandwidth monitor
*
* This routine deinitializes the bandwidth monitor.
*
* RETURNS: NONE
*/
void __exit bwMonitorExit (void)
    {
#ifdef BWMON_DEBUG
    /* de-initialize the timer traffic rate timer */
	del_timer(&bwRateTimer);
#endif /* BWMON_DEBUG */

#ifdef CONFIG_SYSCTL
    bwMonitorSysctlUnregister();
#endif /* CONFIG_SYSCTL */
    bwmon_multipath_weight_check = NULL;

    printk("Removed bandwidth monitor ...\n");
    }

MODULE_AUTHOR("Samba <samba@teamf1.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("rate based multipath selector");
module_init(bwMonitorInit);
module_exit(bwMonitorExit);

