/* Shared library add-on to ip6tables to add IP6 range matching support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <arpa/inet.h>

#include <ip6tables.h>
#include <linux/netfilter_ipv6/ip6t_ip6range.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"ip6range match v%s options:\n"
"[!] --src-range ip-ip        Match source IP in the specified range\n"
"[!] --dst-range ip-ip        Match destination IP in the specified range\n"
"\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "src-range", 1, 0, '1' },
	{ "dst-range", 1, 0, '2' },
	{0}
};

static void
parse_ip6range(char *arg, struct ip6t_ip6range *range6)
{
    char * dash=NULL;

	dash = strchr(arg, '-');
	if (dash)
		*dash = '\0';
		
    if (inet_pton(AF_INET6, arg, &range6->min_addr) < 0) {
        exit_error(PARAMETER_PROBLEM, "invalid source address value"); 
    }  

    if (dash) {
        if (inet_pton(AF_INET6, dash+1, &range6->max_addr) < 0) {
            exit_error(PARAMETER_PROBLEM, "invalid source address value"); 
        }  
    } else
		range6->max_addr = range6->min_addr;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ip6t_entry *entry,
      unsigned int *nfcache,
      struct ip6t_entry_match **match)
{
	struct ip6t_ip6range_info *info = 
            (struct ip6t_ip6range_info *)(*match)->data;

	switch (c) {
	case '1':
		if (*flags & IP6RANGE_SRC)
			exit_error(PARAMETER_PROBLEM,
                       "ip6range match: Only use --src-range ONCE!");
		*flags |= IP6RANGE_SRC;

		info->flags |= IP6RANGE_SRC;
		check_inverse(optarg, &invert, &optind, 0);
		if (invert) {
			info->flags |= IP6RANGE_SRC_INV;
		}
		parse_ip6range(optarg, &info->src);		

		break;

	case '2':
		if (*flags & IP6RANGE_DST)
			exit_error(PARAMETER_PROBLEM,
				   "ip6range match: Only use --dst-range ONCE!");
		*flags |= IP6RANGE_DST;

		info->flags |= IP6RANGE_DST;
		check_inverse(optarg, &invert, &optind, 0);
		if (invert) {
			info->flags |= IP6RANGE_DST_INV;
        }
		parse_ip6range(optarg, &info->dst);		

		break;

	default:
		return 0;
	}
	return 1;
}

/* Final check; must have specified --src-range or --dst-range. */
static void
final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
                   "ip6range match: "
                   "You must specify `--src-range' or `--dst-range'");
}

static void
print_ip6range(const struct ip6t_ip6range *range6)
{
	printf("%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x-"
           "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
            ntohs(range6->min_addr.s6_addr16[0]),
            ntohs(range6->min_addr.s6_addr16[1]),
            ntohs(range6->min_addr.s6_addr16[2]),
            ntohs(range6->min_addr.s6_addr16[3]),
            ntohs(range6->min_addr.s6_addr16[4]),
            ntohs(range6->min_addr.s6_addr16[5]),
            ntohs(range6->min_addr.s6_addr16[6]),
            ntohs(range6->min_addr.s6_addr16[7]),
            ntohs(range6->max_addr.s6_addr16[0]),
            ntohs(range6->max_addr.s6_addr16[1]),
            ntohs(range6->max_addr.s6_addr16[2]),
            ntohs(range6->max_addr.s6_addr16[3]),
            ntohs(range6->max_addr.s6_addr16[4]),
            ntohs(range6->max_addr.s6_addr16[5]),
            ntohs(range6->max_addr.s6_addr16[6]),
            ntohs(range6->max_addr.s6_addr16[7]));
}

/* Prints out the info. */
static void
print(const struct ip6t_ip6 *ip,
      const struct ip6t_entry_match *match,
      int numeric)
{
	struct ip6t_ip6range_info *info = (struct ip6t_ip6range_info *)match->data;

	if (info->flags & IP6RANGE_SRC) {
		printf("source IP range ");
		if (info->flags & IP6RANGE_SRC_INV)
			printf("! ");
		print_ip6range(&info->src);
	}
	if (info->flags & IP6RANGE_DST) {
		printf("destination IP range ");
		if (info->flags & IP6RANGE_DST_INV)
			printf("! ");
		print_ip6range(&info->dst);
	}
}

/* Saves the union ipt_info in parsable form to stdout. */
static void
save(const struct ip6t_ip6 *ip, const struct ip6t_entry_match *match)
{
	struct ip6t_ip6range_info *info = (struct ip6t_ip6range_info *)match->data;

	if (info->flags & IP6RANGE_SRC) {
		if (info->flags & IP6RANGE_SRC_INV)
			printf("! ");
		printf("--src-range ");
		print_ip6range(&info->src);
		if (info->flags & IP6RANGE_DST)
			fputc(' ', stdout);
	}
	if (info->flags & IP6RANGE_DST) {
		if (info->flags & IP6RANGE_DST_INV)
			printf("! ");
		printf("--dst-range ");
		print_ip6range(&info->dst);
	}
}

static struct ip6tables_match ip6range = { 
	.next		= NULL,
	.name		= "ip6range",
	.version	= IPTABLES_VERSION,
	.size		= IP6T_ALIGN(sizeof(struct ip6t_ip6range_info)),
	.userspacesize	= IP6T_ALIGN(sizeof(struct ip6t_ip6range_info)),
	.help		= &help,
	.parse		= &parse,
	.final_check	= &final_check,
	.print		= &print,
	.save		= &save,
	.extra_opts	= opts
};

void _init(void)
{
	register_match6(&ip6range);
}
