/* ipt_BWPROF.c - bandwidth profile module */

/* Copyright (c) 2007, TeamF1, Inc. */

/*
modification history
--------------------
01b,26dec07,suh ported for 32-bit userland on a 64-bit kernel
01a,28Feb07,rnp created
*/

/* Shared library add-on to iptables to add BWPROF target support. */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <syslog.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
/* For 64bit kernel / 32bit userspace */
#include "../include/linux/netfilter_ipv4/ipt_BWPROF.h"

/* max 32 here since it is used in flags(32-bit) */
#define IPT_BWPROF_OPT_LIMIT        0x01
#define IPT_BWPROF_OPT_DIR          0x02
#define IPT_BWPROF_OPT_TYPE         0x03
#define IPT_BWPROF_OPT_IF           0x04
#define IPT_BWPROF_OPT_CT_TC_INDEX  0x05
#define IPT_BWPROF_OPT_LOG_LEVEL    0x06
#define IPT_BWPROF_OPT_TIME_START   0x09
#define IPT_BWPROF_OPT_TIME_STOP    0x0a
#define IPT_BWPROF_OPT_DAYS         0x0b
#define IPT_BWPROF_OPT_ACTION       0x0c

/* these are not user-options, but are used in flags */
#define IPT_BWPROF_OPT_DIR_NONE     0x1d

#define IPT_BWPROF_OPT_TIME         0x1f


/* 
 * Function which prints out usage message.
 */
static void
help(void)
{
	printf(
    "BWPROF target options(include some LOG and TIME target options):\n"
    "  --limit minbw:maxbw         Set bandwidth range in kbps\n"
    "  --dir string                Set direction - ORIGINAL, REPLY\n"
    "  --type string               Set type - ALL, PER_SRC_IP, PER_DST_IP\n"
    "  --if interfaces             Set interface names(comma separated)\n"
    "  --ct_tc_index value         Set connection tc_index value(for dbg)\n"
    "  --log-level level           LOG: Level of logging\n"
    "  --log-prefix prefix         LOG: Prefix log messages with this prefix.\n"
    "  --log-format format         LOG: Log format.\n"
    "  --timestart value           TIME: HH:MM (default 00:00).\n"
    "  --timestop value            TIME: HH:MM (default 23:59).\n"
    "  --days value                TIME: a list of days to apply.\n"
    "  --action value              TIME: sched,log,accept/block(1+1+0/1)\n"
    "\n");
}

static struct option opts[] = {
	{ "limit", 1, 0, IPT_BWPROF_OPT_LIMIT },
	{ "dir", 1, 0, IPT_BWPROF_OPT_DIR },
	{ "type", 1, 0, IPT_BWPROF_OPT_TYPE },
	{ "if", 1, 0, IPT_BWPROF_OPT_IF },
	{ "ct_tc_index", 1, 0, IPT_BWPROF_OPT_CT_TC_INDEX },
	{ "log-level", 1, 0, IPT_BWPROF_OPT_LOG_LEVEL },
	{ "log-prefix", 1, 0, IPT_BWPROF_OPT_LOG_PREFIX },
	{ "log-format", 1, 0, IPT_BWPROF_OPT_LOG_FORMAT },
	{ "timestart", 1, 0, IPT_BWPROF_OPT_TIME_START },
	{ "timestop", 1, 0, IPT_BWPROF_OPT_TIME_STOP },
	{ "days", 1, 0, IPT_BWPROF_OPT_DAYS },
	{ "action", 1, 0, IPT_BWPROF_OPT_ACTION },
	{ 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
struct ipt_bwprof_target_info *bwprofinfo
    = (struct ipt_bwprof_target_info *)(t)->data;

/* default logLevel is LOG_DEBUG */
bwprofinfo->logLevel = LOG_DEBUG;
}

struct name_list {
	const char *valname;
	unsigned int val;
};

static struct name_list bwprof_dir_names[] = 
{ 
    { "ORIGINAL", BWPROF_DIR_ORIGINAL },
    { "REPLY"   , BWPROF_DIR_REPLY },
    { "BOTH"    , BWPROF_DIR_BOTH },
    { "NONE"    , BWPROF_DIR_NONE },
};

static struct name_list bwprof_type_names[] = 
{ 
    { "ALL"        , BWPROF_TYPE_ALL },
    { "PER_SRC_IP" , BWPROF_TYPE_PER_SRC_IP },
    { "PER_DST_IP" , BWPROF_TYPE_PER_DST_IP }
};

static struct name_list bwprof_log_names[] =
{ 
    { "alert",   LOG_ALERT },
    { "crit",    LOG_CRIT },
    { "debug",   LOG_DEBUG },
    { "emerg",   LOG_EMERG },
    { "info",    LOG_INFO },
    { "notice",  LOG_NOTICE },
    { "warning", LOG_WARNING }
};

static struct name_list bwprof_action_names[] =
{ 
    { "accept",             BWPROF_ACTION_ACCEPT },
    { "block",              BWPROF_ACTION_BLOCK },
    { "log_accept",         BWPROF_ACTION_LOG_ACCEPT },
    { "log_block",          BWPROF_ACTION_LOG_BLOCK },
    { "sched_accept",       BWPROF_ACTION_SCHED_ACCEPT },
    { "sched_block",        BWPROF_ACTION_SCHED_BLOCK },
    { "sched_log_accept",   BWPROF_ACTION_SCHED_LOG_ACCEPT },
    { "sched_log_block",    BWPROF_ACTION_SCHED_LOG_BLOCK },
};

static u_int8_t
parse_custom(const char *valname, struct name_list * names, int nameListSize)
{
	unsigned int val = -1;
	unsigned int set = 0;

	if (string_to_number(valname, 0, 0, &val) == -1) {
		unsigned int i = 0;

		for (i = 0; i < nameListSize/sizeof(struct name_list); i++) 
        {
			if (strncasecmp(valname, names[i].valname, strlen(valname)) == 0) {
				if (set++)
					exit_error(PARAMETER_PROBLEM,
                               "bwprof value `%s' ambiguous", valname);
				val = names[i].val;
			}
		}

		if (!set)
			exit_error(PARAMETER_PROBLEM, " `%s' unknown", valname);
	}

	return (u_int8_t)val;
}

/* copied from libipt_TIME.c */
/**
 * param: part1, a pointer on a string 2 chars maximum long string, that will contain the hours.
 * param: part2, a pointer on a string 2 chars maximum long string, that will contain the minutes.
 * param: str_2_parse, the string to parse.
 * return: 1 if ok, 0 if error.
 */
static int
split_time(char **part1, char **part2, const char *str_2_parse)
{
	unsigned short int i,j=0;
	char *rpart1 = *part1;
	char *rpart2 = *part2;
	unsigned char found_column = 0;

	/* Check the length of the string */
	if (strlen(str_2_parse) > 5)
		return 0;
	/* parse the first part until the ':' */
	for (i=0; i<2; i++)
	{
		if (str_2_parse[i] == ':')
			found_column = 1;
		else
			rpart1[i] = str_2_parse[i];
	}
	if (!found_column)
		i++;
	j=i;
	/* parse the second part */
	for (; i<strlen(str_2_parse); i++)
	{
		rpart2[i-j] = str_2_parse[i];
	}
	/* if we are here, format should be ok. */
	return 1;
}

static int
parse_number(char *str, int num_min, int num_max, int *number)
{
	/* if the number starts with 0, replace it with a space else
	string_to_number() will interpret it as octal !! */
	if (strlen(str) == 0)
		return 0;

	if ((str[0] == '0') && (str[1] != '\0'))
		str[0] = ' ';

	return string_to_number(str, num_min, num_max, number);
}
static void
parse_time_string(int *hour, int *minute, const char *time)
{
	char *hours;
	char *minutes;
	hours = (char *)malloc(3);
	minutes = (char *)malloc(3);
	memset(hours, 0, 3);
	memset(minutes, 0, 3);

	if (split_time((char **)&hours, (char **)&minutes, time) == 1)
	{
		*hour = 0;
		*minute = 0;
		if ((parse_number((char *)hours, 0, 23, hour) != -1) &&
		    (parse_number((char *)minutes, 0, 59, minute) != -1))
		{
			free(hours);
			free(minutes);
			return;
		}
	}

	free(hours);
	free(minutes);

	/* If we are here, there was a problem ..*/
	exit_error(PARAMETER_PROBLEM,
		   "invalid time `%s' specified, should be HH:MM format", time);
}

/* return 1->ok, return 0->error */
static int
parse_day(int *days, int from, int to, const char *string)
{
	char *dayread;
	char *days_str[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	unsigned short int days_of_week[7] = {64, 32, 16, 8, 4, 2, 1};
	unsigned int i;

	dayread = (char *)malloc(4);
	bzero(dayread, 4);
	if ((to-from) != 3) {
		free(dayread);
		return 0;
	}
	for (i=from; i<to; i++)
		dayread[i-from] = string[i];
	for (i=0; i<7; i++)
		if (strcmp(dayread, days_str[i]) == 0)
		{
			*days |= days_of_week[i];
			free(dayread);
			return 1;
		}
	/* if we are here, we didn't read a valid day */
	free(dayread);
	return 0;
}

static void
parse_days_string(int *days, const char *daystring)
{
	int len;
	int i=0;
	char *err = "invalid days `%s' specified, should be Sun,Mon,Tue... format";

	len = strlen(daystring);
	if (len < 3)
		exit_error(PARAMETER_PROBLEM, err, daystring);	
	while(i<len)
	{
		if (parse_day(days, i, i+3, daystring) == 0)
			exit_error(PARAMETER_PROBLEM, err, daystring);
		i += 4;
	}
}

static void
divide_time(int fulltime, int *hours, int *minutes)
{
	*hours = fulltime / 60;
	*minutes = fulltime % 60;
}

static void
print_days(int daynum)
{
	char *days[7] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
	unsigned short int days_of_week[7] = {64, 32, 16, 8, 4, 2, 1};
	unsigned short int i, nbdays=0;

	for (i=0; i<7; i++) {
		if ((days_of_week[i] & daynum) == days_of_week[i])
		{
			if (nbdays>0)
				printf(",%s", days[i]);
			else
				printf("%s", days[i]);
			++nbdays;
		}
	}
	printf(" ");
}
/* END: copied from libipt_TIME.c */

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
	 const struct ipt_entry *entry,
	 struct ipt_entry_target **target)
{
	struct ipt_bwprof_target_info *bwprofinfo
		= (struct ipt_bwprof_target_info *)(*target)->data;

	switch (c) {
	case IPT_BWPROF_OPT_LIMIT:
        {
        char * buf1 = NULL, * buf2 = NULL;
        buf1 = strdup(optarg);
        if ((buf2 = strchr(buf1, ':')) != NULL)
            { *buf2 = '\0'; buf2++; }
#ifdef KERNEL_64_USERSPACE_32
		if (string_to_number_ll(buf1, 0, 0, &bwprofinfo->minbw))
#else
		if (string_to_number_l(buf1, 0, 0, &bwprofinfo->minbw))
#endif
            exit_error(PARAMETER_PROBLEM, "Bad min. bandwidth value `%s'",
                       buf1);
        if (buf2 != NULL)
            {
#ifdef KERNEL_64_USERSPACE_32
            if (string_to_number_ll(buf2, 0, 0, &bwprofinfo->maxbw))
#else
            if (string_to_number_l(buf2, 0, 0, &bwprofinfo->maxbw))
#endif
                exit_error(PARAMETER_PROBLEM, "Bad max. bandwidth value `%s'",
                           buf2);
            }
		if (*flags & FLAG(IPT_BWPROF_OPT_LIMIT))
			exit_error(PARAMETER_PROBLEM,
			           "BWPROF target: Can't specify --limit twice");
		*flags |= FLAG(IPT_BWPROF_OPT_LIMIT);
        free(buf1);
		break;
        }
	case IPT_BWPROF_OPT_DIR:
        bwprofinfo->dir = parse_custom(optarg, bwprof_dir_names,
                                       sizeof(bwprof_dir_names));
		if (*flags & FLAG(IPT_BWPROF_OPT_DIR))
			exit_error(PARAMETER_PROBLEM,
			           "BWPROF target: Can't specify --dir twice");
		*flags |= FLAG(IPT_BWPROF_OPT_DIR);
        if (bwprofinfo->dir == BWPROF_DIR_NONE)
            *flags |= FLAG(IPT_BWPROF_OPT_DIR_NONE);
		break;
	case IPT_BWPROF_OPT_TYPE:
        bwprofinfo->type = parse_custom(optarg, bwprof_type_names,
                                        sizeof(bwprof_type_names));
		if (*flags & FLAG(IPT_BWPROF_OPT_TYPE))
			exit_error(PARAMETER_PROBLEM,
			           "BWPROF target: Can't specify --type twice");
		*flags |= FLAG(IPT_BWPROF_OPT_TYPE);
		break;
	case IPT_BWPROF_OPT_IF:
        {
        char * ptr = NULL;
        char * buffer = strdup(optarg);
        unsigned int i;
        for (ptr = strtok(buffer, ","), i = 0; 
             ptr && (i < BWPROF_MAX_IF); 
             ptr = strtok(NULL, ","), i++) {
            strncpy(bwprofinfo->ifName[i], ptr, BWPROF_MAX_IF_NAME_SZ);
            }

        if (ptr != NULL)
            exit_error(PARAMETER_PROBLEM, "Too many interfaces(max: %d)", 
                       BWPROF_MAX_IF);
        free(buffer);
        /* check for presence of atleast 1 name */
        if (bwprofinfo->ifName[0][0] != '\0')
            *flags |= FLAG(IPT_BWPROF_OPT_IF);
        break;
        }
	case IPT_BWPROF_OPT_CT_TC_INDEX:
		if (string_to_number_l(optarg, 0, 0, &bwprofinfo->ct_tc_index))
        exit_error(PARAMETER_PROBLEM, "Bad ct_tc_index value `%s'", optarg);
		if (*flags & FLAG(IPT_BWPROF_OPT_CT_TC_INDEX))
			exit_error(PARAMETER_PROBLEM,
			           "BWPROF target: Can't specify --ct_tc_index twice");
		*flags |= FLAG(IPT_BWPROF_OPT_CT_TC_INDEX);
		break;
	case IPT_BWPROF_OPT_LOG_LEVEL:
        bwprofinfo->logLevel = parse_custom(optarg, bwprof_log_names,
                                            sizeof(bwprof_log_names));
		if (*flags & FLAG(IPT_BWPROF_OPT_LOG_LEVEL))
			exit_error(PARAMETER_PROBLEM,
			           "BWPROF target: Can't specify --log-level twice");
		*flags |= FLAG(IPT_BWPROF_OPT_LOG_LEVEL);
        /* LOG target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_LOG);
		break;
	case IPT_BWPROF_OPT_LOG_PREFIX:
        {
		if (*flags & FLAG(IPT_BWPROF_OPT_LOG_PREFIX))
			exit_error(PARAMETER_PROBLEM, "Can't specify --log-prefix twice");

		if (strlen(optarg) > sizeof(bwprofinfo->logPrefix) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum prefix length %u for --log-prefix",
				   (unsigned int)sizeof(bwprofinfo->logPrefix) - 1);

		if (strlen(optarg) != strlen(strtok(optarg, "\n")))
			exit_error(PARAMETER_PROBLEM,
				   "Newlines not allowed in --log-prefix");

		strcpy(bwprofinfo->logPrefix, optarg);
		*flags |= FLAG(IPT_BWPROF_OPT_LOG_PREFIX);
        /* LOG target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_LOG);
		break;
        }
	case IPT_BWPROF_OPT_LOG_FORMAT:
        {
		if (*flags & FLAG(IPT_BWPROF_OPT_LOG_FORMAT))
			exit_error(PARAMETER_PROBLEM, "Can't specify --log-format twice");

		if (strlen(optarg) > sizeof(bwprofinfo->logFormat) - 1)
			exit_error(PARAMETER_PROBLEM,
				   "Maximum format length %u for --log-format",
				   (unsigned int)sizeof(bwprofinfo->logFormat) - 1);

		if (strlen(optarg) != strlen(strtok(optarg, "\n")))
			exit_error(PARAMETER_PROBLEM,
				   "Newlines not allowed in --log-format");

		strcpy(bwprofinfo->logFormat, optarg);
		*flags |= FLAG(IPT_BWPROF_OPT_LOG_FORMAT);
        /* LOG target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_LOG);
		break;
        }
	case IPT_BWPROF_OPT_TIME_START:
        {
        int hours, minutes;
		if (*flags & FLAG(IPT_BWPROF_OPT_TIME_START))
            exit_error(PARAMETER_PROBLEM, "Can't specify --timestart twice");
		parse_time_string(&hours, &minutes, optarg);
		bwprofinfo->timeStart = (hours * 60) + minutes;
		*flags |= FLAG(IPT_BWPROF_OPT_TIME_START);
        /* TIME target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_TIME);
		break;
        }
	case IPT_BWPROF_OPT_TIME_STOP:
        {
        int hours, minutes;
		if (*flags & FLAG(IPT_BWPROF_OPT_TIME_STOP))
            exit_error(PARAMETER_PROBLEM, "Can't specify --timestart twice");
		parse_time_string(&hours, &minutes, optarg);
		bwprofinfo->timeStop = (hours * 60) + minutes;
		*flags |= FLAG(IPT_BWPROF_OPT_TIME_STOP);
        /* TIME target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_TIME);
		break;
        }
	case IPT_BWPROF_OPT_DAYS:
        {
        int days = 0;
		if (*flags & FLAG(IPT_BWPROF_OPT_DAYS))
            exit_error(PARAMETER_PROBLEM, "Can't specify --days twice");
		parse_days_string(&days, optarg);
		bwprofinfo->timeDays = days;
		*flags |= FLAG(IPT_BWPROF_OPT_DAYS);
        /* TIME target option is specified */
		*flags |= FLAG(IPT_BWPROF_OPT_TIME);
		break;
        }
	case IPT_BWPROF_OPT_ACTION:
        {
		if (*flags & FLAG(IPT_BWPROF_OPT_ACTION))
            exit_error(PARAMETER_PROBLEM, "Can't specify --action twice");
		bwprofinfo->action = parse_custom(optarg, bwprof_action_names,
                                          sizeof(bwprof_action_names));
		*flags |= FLAG(IPT_BWPROF_OPT_ACTION);
		break;
        }
	default:
		return 0;
	}

    bwprofinfo->flags = *flags;
	return 1;
}

static void
final_check(unsigned int flags)
{
    if (!(flags & FLAG(IPT_BWPROF_OPT_DIR_NONE)))
        {
        if (!(flags & FLAG(IPT_BWPROF_OPT_LIMIT)))
            exit_error(PARAMETER_PROBLEM,
                       "BWPROF target: Parameter --limit is required");

        if (!(flags & FLAG(IPT_BWPROF_OPT_IF)))
            exit_error(PARAMETER_PROBLEM,
                       "BWPROF target: Parameter --if is required");
        }
}

#ifdef KERNEL_64_USERSPACE_32
static void
print_int_hex(unsigned long long i)
{
	printf("0x%llx ", i);
}

static void
print_int_dec(unsigned long long i)
{
	printf("%lld ", i);
}
#else
static void
print_int_hex(unsigned long i)
{
	printf("0x%lx", i);
}

static void
print_int_dec(unsigned long i)
{
	printf("%ld", i);
}
#endif

static  void print_if(char ifName[BWPROF_MAX_IF][BWPROF_MAX_IF_NAME_SZ])
{
	unsigned int i = 0;
    for (i = 0; i < BWPROF_MAX_IF; i++) {
        printf("%s", ifName[i]);
        if ((i+1 < BWPROF_MAX_IF) && (ifName[i+1][0] != '\0'))
            printf(",");
        else
            break;
    }
}

static  void print_custom(int val, struct name_list * names, int nameListSize)
{
	unsigned int i = 0;
    for (i = 0; i < nameListSize/sizeof(struct name_list); i++) {
        if (val == names[i].val) {
            printf("%s", names[i].valname);
            break;
        }
    }
    if (i == nameListSize/sizeof(struct name_list))
        printf("UNKNOWN(%u)", val);
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
	 const struct ipt_entry_target *target,
	 int numeric)
{
	const struct ipt_bwprof_target_info *bwprofinfo =
		(const struct ipt_bwprof_target_info *)target->data;
	printf("BWPROF ");
	print_custom(bwprofinfo->action, bwprof_action_names, 
                 sizeof(bwprof_action_names));
	printf(" ");
	print_custom(bwprofinfo->type,bwprof_type_names, sizeof(bwprof_type_names));
	printf(" [");
	print_int_dec(bwprofinfo->minbw);
	printf(":");
	print_int_dec(bwprofinfo->maxbw);
	printf("]kbps on ");
	print_if(bwprofinfo->ifName);
	printf(" in ");
	print_custom(bwprofinfo->dir, bwprof_dir_names, sizeof(bwprof_dir_names));
	printf(" direction");
    if (bwprofinfo->ct_tc_index)
    {
        printf(" ct_tc_index=");
        print_int_hex(bwprofinfo->ct_tc_index);
    }
    if (bwprofinfo->flags & FLAG(IPT_BWPROF_OPT_LOG))
    {
        printf(" LOG with prefix %s", bwprofinfo->logPrefix);
        printf(" at level=");
        print_custom(bwprofinfo->logLevel, bwprof_log_names, 
                     sizeof(bwprof_log_names));
        if (bwprofinfo->logFormat[0])
            printf(" using format=%s", bwprofinfo->logFormat);
    }
    if (bwprofinfo->flags & FLAG(IPT_BWPROF_OPT_TIME))
    {
        int hour, minute;
        printf(" TIME ");
        divide_time(bwprofinfo->timeStart, &hour, &minute);
        printf("from %d:%d ", hour, minute);
        divide_time(bwprofinfo->timeStop, &hour, &minute);
        printf("to %d:%d ", hour, minute);
        printf("on ");
        if (bwprofinfo->timeDays == 127)
            printf("all days ");
        else
            print_days(bwprofinfo->timeDays);
    }
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
	const struct ipt_bwprof_target_info *bwprofinfo =
		(const struct ipt_bwprof_target_info *)target->data;

	printf("--limit ");
	print_int_dec(bwprofinfo->minbw);
	printf(":");
	print_int_dec(bwprofinfo->maxbw);
	printf("--if ");
	print_if(&bwprofinfo->ifName[0]);
	printf("--dir ");
	print_custom(bwprofinfo->dir, bwprof_dir_names, sizeof(bwprof_dir_names));
	printf("--type ");
	print_custom(bwprofinfo->type,bwprof_type_names, sizeof(bwprof_type_names));
	printf("--ct_tc_index ");
	print_int_hex(bwprofinfo->ct_tc_index);
}

static
struct iptables_target bwprof = {
	.next		= NULL,
	.name		= "BWPROF",
	.version	= IPTABLES_VERSION,
	.revision	= 0,
	.size		= IPT_ALIGN(sizeof(struct ipt_bwprof_target_info)),
	.userspacesize	= IPT_ALIGN(sizeof(struct ipt_bwprof_target_info)),
	.help		= &help,
	.init		= &init,
	.parse		= &parse,
	.final_check	= &final_check,
	.print		= &print,
	.save		= &save,
	.extra_opts	= opts
};

void _init(void)
{
	register_target(&bwprof);
}
