
#include "wrapper-cvmx-includes.h"
#include "ethernet-headers.h"
#include "ethernet-offload.h"

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>


#ifdef CVM_IPFWD_DRIVER_OFFLOAD

/* Shared pointer to the forwarding cache */
extern cvm_ipfwd_flow_bucket_t *cvm_gbl_ipfwd_cache;
extern int (*tf1_fib_ct_refresh_hook)(struct ip_conntrack * ct);
int cvm_ipfwd_ct_match_find (struct ip_conntrack_tuple * tuple, 
                             cvm_ipfwd_flow_bucket_t * bucket,
                             unsigned int timeout);
extern unsigned int tcp_ct_conn_timeout (struct ip_conntrack * ct);
extern unsigned int udp_ct_conn_timeout (struct ip_conntrack * ct);


/**
 * =============================================
 * = Cavium IP forwarding/NAT offload support. =
 * =============================================
 *
 * This feature is enabled by defining CVM_IPFWD_DRIVER_OFFLOAD
 * and setting ipfwd_driver_offload module parameter to 1. 
 *
 * Once enabled, the "active" TCP/UDP flows are forwarded (and
 * NATted) in the ethernet driver. 
 *
 */

/**
 *
 * IPFWD/NAT OFFLOAD INLINE FUNCTIONS
 *
 */

/**
 * Helper function to calculate IPv4 header checksum.
 * 
 * @param *ip pointer to the beginning of IP header
 * 
 * @return 16-bit one's complement IPv4 checksum.
 * No alignment requirements.
 */
uint16_t cvm_ipfwd_calculate_ip_header_checksum(uint16_t *ip)
{
	uint64_t sum;
	uint16_t *ptr = ip;
	uint8_t *bptr = (uint8_t*) ip;

	sum  = ptr[0];
	sum += ptr[1];
	sum += ptr[2];
	sum += ptr[3];
	sum += ptr[4];
	// Skip checksum field
	sum += ptr[6];
	sum += ptr[7];
	sum += ptr[8];
	sum += ptr[9];

	// Check for options
	if(cvmx_unlikely(bptr[0] != 0x45)) goto slow_cksum_calc;

	return_from_slow_cksum_calc:

	sum = (uint16_t) sum + (sum >> 16);
	sum = (uint16_t) sum + (sum >> 16);
	return((uint16_t) (sum ^ 0xffff));

	slow_cksum_calc:
	// Add IPv4 options into the checksum (if present)
	{
		uint64_t len = (bptr[0] & 0xf) - 5;
		ptr = &ptr[len<<1];

		while(len-- > 0)
		{
			sum += *ptr++;
			sum += *ptr++;
		}
	}

	goto return_from_slow_cksum_calc;
}

/**
 * Helper function to allocate flow_info buffer 
 *
 * @return pointer from the FPA pool (or NULL if pool is empty)
 */
void *cvm_ipfwd_ptr_alloc(void)
{
	void *ptr = cvmx_fpa_alloc(CVMX_FPA_FLOW_INFO_POOL);
	return(ptr);
}
#ifdef CONFIG_TF1_HW_OFFLOAD
/* @Brief: Modifies the cookie value of skb 
 * @param1 : skb of the packet
 * @return : 1 on sucess
 *
 */
int mangleCookieValue(struct sk_buff *skb)
    {
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
    cvm_packet_info_t *pinfo=NULL;
    if(skb == NULL )
        return 1;
    pinfo = (cvm_packet_info_t *) ((((uint64_t) skb->end + sizeof(struct skb_shared_info)) + 7) & ~7ull);
    pinfo->cookie = 0;
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
    return 1;
    }
/**
 * Function returns verdict on marking the packets for caching its flow path 
 * @param1 - ip header of the packet.
 * Returns 1 if the packets flow needs to be cached
 *
 */

int isMarkable(struct iphdr *ip)
    {
    unsigned int ports;
    unsigned short *pData;
    ports = *((uint32_t*) ip + ip->ihl);
    pData = (unsigned short*)&ports;
    //return 1;
    return !(findInBlockList(*pData) || findInBlockList(*(pData+1)));
    }
#endif /*CONFIG_TF1_HW_OFFLOAD*/

/**
 * Helper function to free flow_info buffers.
 *
 * Buffers are freed into a queue and eventually freed to
 * the FPA pool.
 *
 * @param *ptr flow_info buffer pointer to be freed
 *
 */
void cvm_ipfwd_ptr_free(void *ptr)
{
	uint64_t coreid = cvmx_get_core_num();
	cvm_ipfwd_fifo_t *fifo = &free_fifo[coreid];
	void **tmpptr = &fifo->cvm_ipfwd_ptr_fifo[fifo->idx];

	if(*tmpptr)
		cvmx_fpa_free(*tmpptr, CVMX_FPA_FLOW_INFO_POOL, 0);

	*tmpptr = ptr;

	/* increment free fifo ptr */
	fifo->idx = (fifo->idx+1) & CVM_IPFWD_PTR_FIFO_MASK;
}

/**
 * Helper function to flush the freeing fifo.
 *
 */
void cvm_ipfwd_ptr_flush(void)
{
	int i,j;
	cvm_ipfwd_fifo_t *fifo;
	void **tmpptr = NULL;

	for(i=0;i<NR_CPUS;i++)
	{
		fifo = &free_fifo[i];
		fifo->idx = 0;
		for(j=0;j<CVM_IPFWD_PTR_FIFO_SIZE;j++)
		{
			tmpptr = &fifo->cvm_ipfwd_ptr_fifo[j];
			if(*tmpptr)
				cvmx_fpa_free(*tmpptr, CVMX_FPA_FLOW_INFO_POOL, 0);
			*tmpptr = NULL;
		}
	}
}
#ifdef CONFIG_TF1_HW_OFFLOAD
/**
 * Helper function for flushing the hardware flow cache info from sysctl handler
 * of hardware_offload_control and cvm_offload_flush.
 * 
 */ 


void flushFlowInfo()
{
	int i,j;
	cvm_ipfwd_flow_bucket_t *bucket;
	cvm_ipfwd_flow_entry_t *flow_entry;
	cvm_ipfwd_ptr_flush();

	/* Flushes out dynamically registered blockList ports.
	 * Statically registered ports will be freed on respective conntrack
	 * module de-initialization.
	 */
	//flushBlockList(1); 
	for (i=0;i<CVM_IPFWD_LUT_BUCKETS;i++)
	{
		bucket = &cvm_gbl_ipfwd_cache[i];
		spin_lock(&bucket->hdr.wr_lock);
		for (j=0;j<CVM_IPFWD_ENTRIES_PER_BUCKET;j++)
		{
			flow_entry = &bucket->entry[j];
			if (flow_entry->s.flow_info)
				cvm_ipfwd_ptr_free(cvmx_phys_to_ptr((uint64_t)
					           flow_entry->s.flow_info<<7));
			flow_entry->s.flow_info = 0;
			flow_entry->u.word0 = 0;
			flow_entry->u.word1 = 0;
			flow_entry->u.word2 = 0;
		}
		spin_unlock(&bucket->hdr.wr_lock);
	}
}
#endif /*CONFIG_TF1_HW_OFFLOAD*/ 


/**
 * Helper function to check if the timestamp was just created 
 *
 * @param created timestamp to check
 *
 * @return 1: true -- 0: false
 *
 */
int cvm_ipfwd_check_timestamp_just_created(uint64_t created)
{
	if(created > (CVM_IPFWD_TIMER_GET_TIME-4))
		return(1);
	else
		return(0);
}

/**
 * Helper function to check the timestamp.
 *
 * @param created timestamp to check
 *
 * @return 
 *    CVM_IPFWD_EXPIRED - timestamp is expired
 *    CVM_IPFWD_OLD     - timestamp is old (=not new and close to expiration)
 *    CVM_IPFWD_NEW     - timestamp is new
 */
int cvm_ipfwd_check_timestamp(uint64_t created)
{
	uint64_t now;

	now = CVM_IPFWD_TIMER_GET_TIME;

	if(now < created)
		return(CVM_IPFWD_EXPIRED);	 /* Wrap around -- treat as expired */

	if(now < (created + CVM_IPFWD_TIMER_NEW_LIFETIME))
		return(CVM_IPFWD_NEW);

	if(now < (created + CVM_IPFWD_TIMER_TOTAL_LIFETIME))
		return(CVM_IPFWD_OLD);

	return(CVM_IPFWD_EXPIRED);
}

/**
 * Function to randomly return TRUE/FALSE.
 */
int cvm_ipfwd_random_select(void)
{
	return(((cvmx_get_cycle() >> CVM_IPFWD_RND_SHIFT) & CVM_IPFWD_RND_MSK) == CVM_IPFWD_RND_VAL);
}


/*
 * Refresh the conntrack entries, if a corresponding flow entry is present.
 */
int cvm_ipfwd_ct_refresh (struct ip_conntrack * ct)
{
	uint16_t index1 = 0;
	uint16_t index2 = 0;
	uint8_t	 protocol = 0;
	cvm_ipfwd_flow_bucket_t * bucket = NULL;
	struct ip_conntrack_tuple * tuple = NULL;
	int idx = 0;
	unsigned int timeout = 0;

	if (!ct) return 0;

	/*
 	 * if ct->fib_bucket_index == 0, we have to compare 
 	 * CVM_IPFWD_LUT_BUCKETS * CVM_IPFWD_ENTRIES_PER_BUCKET
 	 * entries to find the matching flow entry. To avoid
 	 * this we just skip the lookup.
 	 */	 
	if (ct->fib_bucket_index <= 0)
		return 0;

	tuple = &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple;
	index1 = ct->fib_bucket_index & 0xffff0000 >> 16;
	protocol = tuple->dst.protonum;

	if ((protocol != IPPROTO_TCP) && (protocol != IPPROTO_UDP))
		return 0;

	switch (protocol) {
		case IPPROTO_TCP:
			timeout = tcp_ct_conn_timeout (ct);
			break;
		case IPPROTO_UDP:
			timeout = udp_ct_conn_timeout (ct);
			break;
		default:
			break;
	}

	if (index1 < CVM_IPFWD_LUT_BUCKETS) {
		bucket = &cvm_gbl_ipfwd_cache[index1];
		spin_lock(&bucket->hdr.wr_lock);
		if ((idx = cvm_ipfwd_ct_match_find(tuple, bucket, timeout))
						!= CVM_IPFWD_NOT_FOUND) {
				spin_unlock(&bucket->hdr.wr_lock);
				goto ct_refresh;
		}
		spin_unlock(&bucket->hdr.wr_lock);
	}
		
	tuple = &ct->tuplehash[IP_CT_DIR_REPLY].tuple;
	index2 = ct->fib_bucket_index & 0x0000ffff;

	if (index2 < CVM_IPFWD_LUT_BUCKETS) {
		bucket = &cvm_gbl_ipfwd_cache[index2];
		spin_lock(&bucket->hdr.wr_lock);
		if ((idx = cvm_ipfwd_ct_match_find(tuple, bucket, timeout))
						!= CVM_IPFWD_NOT_FOUND) {
				spin_unlock(&bucket->hdr.wr_lock);
				goto ct_refresh;
		}
		spin_unlock(&bucket->hdr.wr_lock); 	
	}

	return 0;
	
	ct_refresh:

	del_timer(&ct->timeout);
	ct->timeout.expires = jiffies + timeout;
	add_timer(&ct->timeout);

	return 1;
}

/*
 * find the flow cache entry for the given conntrack tuple. 
 */
int cvm_ipfwd_ct_match_find (struct ip_conntrack_tuple * tuple, 
							 cvm_ipfwd_flow_bucket_t * bucket,
							 unsigned int timeout)
{
	uint32_t saddr = 0;
	uint32_t daddr = 0;
	uint16_t sport = 0; 
	uint16_t dport = 0; 
	uint8_t	 protocol = 0;
	int idx;
	cvm_ipfwd_flow_entry_t *entry;

	saddr = tuple->src.ip;
	daddr = tuple->dst.ip;
	protocol = tuple->dst.protonum;
	sport = tuple->src.u.tcp.port;
	dport = tuple->dst.u.tcp.port;

	for(idx = 0; idx < CVM_IPFWD_ENTRIES_PER_BUCKET; idx++)
		{
		cvm_ipfwd_flow_info_t * info = NULL;

		entry = &bucket->entry[idx];

		/* Test current entry -- compare 5-tuple */
		if ((saddr != entry->s.ip_saddr) || (daddr != entry->s.ip_daddr) ||
			(protocol != entry->s.ip_proto) || (sport != entry->s.l4_sport) ||
			(dport != entry->s.l4_dport))
			continue;

		info = (cvm_ipfwd_flow_info_t*)
				cvmx_phys_to_ptr((uint64_t)entry->s.flow_info << 7);

		if ((jiffies - info->lastused) < timeout)
			return(idx);
		}

	/* If we get here, entry was not found */
	return(CVM_IPFWD_NOT_FOUND);
}

/**
 * Fast-path cache look-up function.
 *
 * Search a cache bucket to find an entry that matches
 * input packet's 5-tuple.
 *
 * @param *bucket ptr to hash bucket
 * @param *ip ptr to IP header of the input packet 
 * @param *entry flow_entry structure to be populated (output)
 *
 * @return index of the matching entry [0 .. (CVM_IPFWD_ENTRIES_PER_BUCKET-1)]
 * @return CVM_IPFWD_NOT_FOUND entry not found
 *
 */
int cvm_ipfwd_find_entry(cvm_ipfwd_flow_bucket_t *bucket,
                         struct iphdr* ip,
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
                         uint16_t vlan_TCI, 
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
                         cvm_ipfwd_flow_entry_t *entry)
{
	uint32_t saddr = ip->saddr, daddr = ip->daddr;
	uint32_t ports = *(((uint32_t*) ip) + ip->ihl);
	int count, idx;
	uint8_t  proto = ip->protocol;

	for(idx = 0; idx < CVM_IPFWD_ENTRIES_PER_BUCKET; idx++)
	{
		count = CVM_IPFWD_ENTRY_RETRIES;

		/* Read current entry */
		do
		{
			*entry = bucket->entry[idx];
		} while((--count >= 0) && (entry->s.id0 != entry->s.id2));

		if(count >= 0)
		{
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
			if (entry->s.vlan_info != vlan_TCI)
				continue;
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
			/* Test current entry -- compare 5-tuple */
			if((saddr != entry->s.ip_saddr) || (daddr != entry->s.ip_daddr) ||
			   (proto != entry->s.ip_proto) ||
			   (ports != ((entry->s.l4_sport << 16) | entry->s.l4_dport)))
				continue;

			/* Found */
			/* Prefetch flow info */
			if(entry->s.flow_info)
				CVMX_PREFETCH0(cvmx_phys_to_ptr((uint64_t)entry->s.flow_info<<7));
			return(idx);
		}
	}

	/* If we get here, entry was not found */
	return(CVM_IPFWD_NOT_FOUND);
}


/**
 *
 * IPFWD/NAT OFFLOAD FUNCTIONS
 *
 */

/**
 * Slow-path: Check bucket header timestamp to see whether bucket can be
 * updated. 
 *
 * @param *bucket ptr to hash bucket
 *
 * @return 0: this bucket cannot be updated
 * @return 1: this bucket can be updated
 */
int cvm_ipfwd_check_bucket_timestamp(cvm_ipfwd_flow_bucket_t *bucket)
{
	uint64_t old = bucket->hdr.timestamp;
	uint64_t new = (cvmx_get_cycle() >> 16) & 0xffff;

	if((new - old) < 8)
		return(0);

	/* Time to punt packet to Linux --
	 * update bucket header timestamp (without locking.) */
	*(uint16_t*)&bucket->hdr.timestamp = new;
	CVMX_SYNCWS;
	return(1);
}

/**
 * Slow-path: Check bucket timestamp.
 *
 * @param *bucket ptr to hash bucket
 *
 */
void cvm_ipfwd_mark_packet(struct sk_buff *skb, cvmx_wqe_t *work, cvm_ipfwd_flow_bucket_t *bucket)
{
	struct iphdr *ip = (struct iphdr*) (skb->data + work->word2.s.ip_offset);
	struct tcphdr *tcp;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    struct vlan_ethhdr *vhdr = (struct vlan_ethhdr*)(skb->data);
#endif
    if(!isMarkable(ip)){
        return;
    }
#endif/*CONFIG_TF1_HW_OFFLOAD*/

	cvm_packet_info_t *info = (cvm_packet_info_t *) 
		((((uint64_t) skb->end + sizeof(struct skb_shared_info)) + 7) & ~7ull);
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    info->reserved0=((vhdr->h_vlan_proto == htons(ETH_P_8021Q))?1:0);

	if (info->reserved0)
		info->reserved1 = vhdr->h_vlan_TCI;
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/

	bucket->hdr.reserved = work->tag & CVM_IPFWD_LUT_MASK;
	info->bucket = bucket;
	info->cookie = CVM_IPFWD_COOKIE_VALUE;
	info->ip.saddr = ip->saddr;
	info->ip.daddr = ip->daddr;
	info->ip.protocol = ip->protocol;
	info->ip.tot_len = ip->tot_len;
	info->ip.ihl = ip->ihl;
	((uint32_t*)&info->ip)[ip->ihl] = ((uint32_t*)ip)[ip->ihl];

	/* If TCP, cache also sequence number and acknowlegment number */
	if(ip->protocol == IP_PROTOCOL_TCP)
	{
		tcp = (struct tcphdr *) ((uint32_t*)ip+ip->ihl);
		info->seq = tcp->seq;
		info->ack_seq = tcp->ack_seq;
	}
}

/**
 * Walks through all entries in a bucket.  Returns an index to the
 * oldest entry and populates a flow_entry structure.
 *
 * Note! Bucket must be locked prior to calling this function 
 *
 * @param *bucket bucket pointer
 * @param *entry pointer to a flow_entry structure (output).  The
 *               matching flow_entry will be copied into this structure.
 * @return idx matching entry index (within bucket)
 */
int cvm_ipfwd_get_free_entry(cvm_ipfwd_flow_bucket_t *bucket, cvm_ipfwd_flow_entry_t *entry)
{
	int idx = 0, i;
	uint64_t min = ~0ull;
	cvm_ipfwd_flow_entry_t *entry_ptr;

	for(i=0;i<CVM_IPFWD_ENTRIES_PER_BUCKET;i++)
	{
		entry_ptr = &bucket->entry[i];
		if(entry_ptr->s.timestamp < min)
		{
			min = entry_ptr->s.timestamp;
			idx = i;
		}
	}

	/* Copy found entry */
	*entry = bucket->entry[idx];

	return(idx);
}

/**
 * Format flow entry fields, based on an IP packet.
 *
 * Note! Bucket must be locked prior to calling this function 
 *
 * @param *ip pointer to an IP header
 * @param *entry pointer to a flow entry
 * @return none
 */
void cvm_ipfwd_setup_entry(struct iphdr *ip, cvm_ipfwd_flow_entry_t *entry)
{
	uint8_t id = entry->s.id0 + 1;	/* Cache current id count */
	uint32_t ports;

	/* Free flow_info buf if one exists */
	if(entry->s.flow_info)
		cvm_ipfwd_ptr_free(cvmx_phys_to_ptr((uint64_t)entry->s.flow_info<<7));

	/* Clear entry */
	entry->u.word0 = 0;
	entry->u.word1 = 0;
	entry->u.word2 = 0;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    entry->u.word3 = 0;
#endif
#endif/*CONFIG_TF1_HW_OFFLOAD*/

	entry->s.id0 = id;
	entry->s.id1 = id;
	entry->s.id2 = id;
	entry->s.ip_proto = ip->protocol;
	entry->s.ip_saddr = ip->saddr;
	entry->s.ip_daddr = ip->daddr;
	ports = *((uint32_t*)ip + ip->ihl);

	entry->s.l4_sport = ports >> 16;
	entry->s.l4_dport = ports & 0xffff;
}

/**
 * Setup flow info structure.  Copies Ethernet dst/src, IP src/dst/proto,
 * and L4 ports from a packet pointed to by the skb into the flow info
 * structure.
 *
 * @param *info pointer to flow info structure
 * @param *skb  pointer to sk_buff
 * @param *dev  pointer to output device structure
 * @return none
 */
void cvm_ipfwd_setup_flow_info(cvm_ipfwd_flow_info_t *info, struct sk_buff *skb, struct net_device *dev)
{
	cvm_ipfwd_eh_t *eh = (cvm_ipfwd_eh_t *) skb->data;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    struct vlan_ethhdr *vhdr = (struct vlan_ethhdr*) skb->data;
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	struct iphdr *ip;
 int ipoffset = 14;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q 
    if(vhdr->h_vlan_proto == htons(ETH_P_8021Q))
    {
     info->vlan_present = 1;
     info->vlan_info = vhdr->h_vlan_TCI;
     ipoffset += 4;
    } 
     else
    {
    info->vlan_present = 0;
    } 
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	info->enet_src_hi = eh->enet_src_hi;
	info->enet_src_lo = eh->enet_src_lo;
	info->enet_dst_hi = eh->enet_dst_hi;
	info->enet_dst_lo = eh->enet_dst_lo;

    ip = (struct iphdr*) (skb->data + ipoffset);
	info->ip_saddr = ip->saddr;
	info->ip_daddr = ip->daddr;
	info->ip_proto = ip->protocol;
	info->ports = *((uint32_t*) ip + ip->ihl);
	info->dev = dev;
	info->lastused = jiffies;
}

/**
 * Cache flow routine.
 *
 * If packet info (at the end of the packet buffer) is valid, a new
 * flow entry (and flow info) will be created for the packet.
 *
 * The following must be true for the flow to be cached:
 *    - packet_info->cookie must be valid
 *    - packet_info->bucket must be valid (LUT bucket ptr)
 *    - packet length must not have changed during its traversal
 *      through the Linux stack
 *    - If TCP, sequence number and sequence acknowlegement numbers
 *      must not have changed
 *
 * If any of the above conditions is not true, then the flow associated with
 * this packet will not be cached.
 *
 * @param *skb pointer to sk_buff
 * @param *dev pointer to output interface's device structe
 */
int cvm_ipfwd_cache_flow(struct sk_buff *skb, struct net_device *dev, cvm_packet_info_t *info)
{
	int idx, status;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    int vlan_present=0;
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	cvm_ipfwd_flow_entry_t entry;
	cvm_ipfwd_flow_bucket_t *bucket;
        struct iphdr *ipnew;

	/* iph points to the new header -- ip points to the cached header */
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    struct vlan_ethhdr *vhdr = (struct vlan_ethhdr*) skb->data; 

    vlan_present = ((vhdr->h_vlan_proto == htons(ETH_P_8021Q))?1:0);

    if(vlan_present == 1)
     ipnew = (struct iphdr *)(skb->data + sizeof(struct ethhdr) +
                                                 4);
    else
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
     ipnew = (struct iphdr *)(skb->data + sizeof(struct ethhdr));
    

    struct iphdr *ip = &info->ip;
	struct tcphdr *l4 = (struct tcphdr *) ((uint32_t*)ipnew + ipnew->ihl);

	///* Check whether this buffer has valid packet_info */
	if((info->cookie != CVM_IPFWD_COOKIE_VALUE) || (info->bucket == NULL))
	{
		return(0);
	}

	bucket = info->bucket;

	/* Verify length */
	if(ip->tot_len != ipnew->tot_len)
		return(0);

	/* If TCP, check whether sequence number or acknowlegment number changed */
	if(ipnew->protocol == IP_PROTOCOL_TCP)
	{
		if(info->seq != l4->seq)
			return(0);
		if(info->ack_seq != l4->ack_seq)
			return(0);
	}

#if 0
	if(spin_trylock(&bucket->hdr.wr_lock))
	{
		/* Lock is taken -- abort */
		return(0);
	}
#else
	/* packet_info valid -> lock bucket */
	spin_lock(&bucket->hdr.wr_lock);
#endif

	/* See if a matching entry already exists */
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    if (info->reserved0)
    idx = cvm_ipfwd_find_entry(bucket,ip , info->reserved1 ,&entry);
    else
    idx = cvm_ipfwd_find_entry(bucket, ip, 0, &entry);
#else
	idx = cvm_ipfwd_find_entry(bucket, ip, &entry);
#endif
#else
	idx = cvm_ipfwd_find_entry(bucket, ip, &entry);
#endif /*CONFIG_TF1_HW_OFFLOAD*/

	if(idx == CVM_IPFWD_NOT_FOUND)
	{
		struct ip_conntrack * ct=NULL;

		/* Find available entry (replace oldest) */
		idx = cvm_ipfwd_get_free_entry(bucket, &entry);

		/* If the entry-to-be-replaced was just created, don't replace it */
		if(cvm_ipfwd_check_timestamp_just_created(entry.s.timestamp))
		{
			spin_unlock(&bucket->hdr.wr_lock);
			return(0);
		}

		/* Set entry fields in free flow info ptr if noe exists */
		cvm_ipfwd_setup_entry(ip, &entry);	   
		entry.s.index = ((cvm_ipfwd_flow_bucket_t *)info->bucket)->hdr.reserved;
		ct = (struct ip_conntrack *)skb->nfct;
		if (ct) {
			if (ct->fib_bucket_index > 0) 
				ct->fib_bucket_index |= entry.s.index << 16;
			else
				ct->fib_bucket_index = entry.s.index;
		}

#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
		if(info->reserved0){
			entry.s.vlan_info = info->reserved1;
		}
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	} else	{
		status = cvm_ipfwd_check_timestamp(entry.s.timestamp);
		if(status == CVM_IPFWD_NEW)
			goto done;

		/* OLD & EXPIRED => setup new flow info (at least for now -- add
		 * conditional update later, if necessary) */
		if(entry.s.flow_info)
		{
			cvm_ipfwd_ptr_free(cvmx_phys_to_ptr((uint64_t)entry.s.flow_info<<7));
			entry.s.flow_info = 0;
		}
	}

	/* See if need to allocate info buffer */
	if(entry.s.flow_info == 0)
	{
		entry.s.flow_info = ((uint64_t) cvmx_ptr_to_phys(cvm_ipfwd_ptr_alloc())) >> 7;
		if(entry.s.flow_info == 0)
		{
			DEBUGPRINT("%s: Failed to allocate flow info buffer\n", __func__); 
			spin_unlock(&bucket->hdr.wr_lock);
			return(0);
		}

		/* Setup info */
		cvm_ipfwd_setup_flow_info(cvmx_phys_to_ptr((uint64_t)entry.s.flow_info<<7), skb, dev);
	}

	/* Refresh time stamp */
	entry.s.timestamp = CVM_IPFWD_TIMER_GET_TIME;

	/* Store entry */
	bucket->entry[idx] = entry;

	done:
	spin_unlock(&bucket->hdr.wr_lock);
	return(0);
}

#ifdef CONFIG_PROC_FS
int cvm_mac_sprintf(char * pMac, char * macstr, int macstrlen)
{
  unsigned char mac[6];
  memset (mac, 0x00, sizeof(mac));
  memcpy (mac, pMac, sizeof(mac));
  snprintf (macstr,	macstrlen, "%02X:%02X:%02X:%02X:%02X:%02X",
			mac[0], mac[1],mac[2], mac[3], mac[4],mac[5]);
  return 0;
}

static struct proc_dir_entry * proc_net_ipfastfwd = NULL;
struct ipfwd_iter_state {
	unsigned int bucket;
};

static  void * ipfwd_get_first(struct seq_file *seq)
{
	struct ipfwd_iter_state *st = seq->private;
	cvm_ipfwd_flow_bucket_t *bucket = NULL;

	st->bucket = 0;
	bucket = &cvm_gbl_ipfwd_cache[st->bucket];
	return (void *)bucket;
}

static void *ipfwd_get_next(struct seq_file *seq, void *head)
{
	struct ipfwd_iter_state *st = seq->private;
	cvm_ipfwd_flow_bucket_t *bucket;

	if (++st->bucket >= CVM_IPFWD_LUT_BUCKETS)
		return NULL;

	bucket = &cvm_gbl_ipfwd_cache[st->bucket];
	return (void *)bucket;
}

static void *ipfwd_get_idx(struct seq_file *seq, loff_t pos)
{
	cvm_ipfwd_flow_bucket_t *bucket = (cvm_ipfwd_flow_bucket_t *)ipfwd_get_first(seq);

	if (bucket)
		while (pos && (bucket = ipfwd_get_next(seq, bucket)))
			pos--;
	return pos ? NULL : (void *)bucket;
}

static void *ipfwd_seq_start(struct seq_file *seq, loff_t *pos)
{
	return ipfwd_get_idx(seq, *pos);
}

static void *ipfwd_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	(*pos)++;
	return ipfwd_get_next(s, v);
}

static void ipfwd_seq_stop(struct seq_file *s, void *v)
{
}

static int ipfwd_seq_show(struct seq_file *s, void *v)
{
	cvm_ipfwd_flow_bucket_t *bucket = (cvm_ipfwd_flow_bucket_t *)v;
	int j;
	cvm_ipfwd_flow_entry_t *flow_entry;
	cvm_ipfwd_flow_info_t * info;
	char src_mac[64];
	char dst_mac[64];
	int status = 0;
	char statusStr[32];

	spin_lock(&bucket->hdr.wr_lock);
	for (j=0; j<CVM_IPFWD_ENTRIES_PER_BUCKET; j++)
	{
		uint16_t  vlan_info=0;    
		uint16_t  src_vlanId=0; 
		uint16_t  dest_vlanId=0; 

		flow_entry = &bucket->entry[j];
		if (flow_entry->u.word0 == 0)
			continue;

		info = (cvm_ipfwd_flow_info_t*)
				cvmx_phys_to_ptr((uint64_t)flow_entry->s.flow_info << 7);
		if (info == NULL)
			continue;

		status =  cvm_ipfwd_check_timestamp(flow_entry->s.timestamp);
		switch (status)
		{
			case CVM_IPFWD_EXPIRED:
				strcpy (statusStr, "expired");
				break;
			case CVM_IPFWD_OLD:
				strcpy (statusStr, "old");
				break;
			case CVM_IPFWD_NEW: 	
				strcpy (statusStr, "new");
				break;
			default:
				strcpy (statusStr, "unknown");
				break;	
		}

		vlan_info = ntohs(info->vlan_info);
		dest_vlanId = (vlan_info & VLAN_VID_MASK);

		vlan_info = ntohs(flow_entry->s.vlan_info); 
		src_vlanId = (vlan_info & VLAN_VID_MASK);

		memset (dst_mac, 0x00, sizeof(dst_mac));
		memset (src_mac, 0x00, sizeof(src_mac));

		cvm_mac_sprintf((char *)&info->enet_dst_hi,
				dst_mac, sizeof(dst_mac));
		cvm_mac_sprintf((char *)&info->enet_src_hi, src_mac,
				sizeof(src_mac));

		seq_printf(s,"id0=%d id1=%d id2=%d [proto=%d sport=%d saddr= "
		   	NIPQUAD_FMT " dport=%d daddr= "
		   	NIPQUAD_FMT " vlanId=%d] => [proto=%d sport=%d saddr= "
			NIPQUAD_FMT " dport=%d daddr= "
			NIPQUAD_FMT " vlanId=%d] created=%d status:%s "
			"ifname=%s src_mac:%s dst_mac:%s flow:%p entry: %p "
			" bucket: %p index=%d lastused:%ld\n",
			flow_entry->s.id0, flow_entry->s.id1,
			flow_entry->s.id2, 

			flow_entry->s.ip_proto,	
			flow_entry->s.l4_sport, NIPQUAD(flow_entry->s.ip_saddr),
			flow_entry->s.l4_dport,	NIPQUAD(flow_entry->s.ip_daddr),
			src_vlanId,

			info->ip_proto,	((info->ports & 0xffff0000) >> 16),
			NIPQUAD(info->ip_saddr), info->ports & 0x0000ffff,
			NIPQUAD(info->ip_daddr), dest_vlanId,

			flow_entry->s.timestamp, statusStr,
			info->dev->name, src_mac, dst_mac, info,
			flow_entry, bucket, bucket->hdr.reserved,
			info->lastused);
	}
	spin_unlock(&bucket->hdr.wr_lock);

	return 0;
}

static struct seq_operations ipfwd_seq_ops = {
	.start = ipfwd_seq_start,
	.next  = ipfwd_seq_next,
	.stop  = ipfwd_seq_stop,
	.show  = ipfwd_seq_show
};

static int ipfwd_open(struct inode *inode, struct file *file)
{
	struct seq_file *seq;
	struct ipfwd_iter_state *st;
	int ret;

	st = kmalloc(sizeof(struct ipfwd_iter_state), GFP_KERNEL);
	if (st == NULL)
		return -ENOMEM;
	ret = seq_open(file, &ipfwd_seq_ops);
	if (ret)
		goto out_free;
	seq          = file->private_data;
	seq->private = st;
	memset(st, 0, sizeof(struct ipfwd_iter_state));
	return ret;
out_free:
	kfree(st);
	return ret;
}

static const struct file_operations ipfwd_file_ops = {
	.owner   = THIS_MODULE,
	.open    = ipfwd_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release_private,
};

int ip_fastfwd_proc_tbl_init (void) {

	tf1_fib_ct_refresh_hook = cvm_ipfwd_ct_refresh;
	proc_net_ipfastfwd = proc_net_fops_create("ipfastfwd", 0440, &ipfwd_file_ops);
	return 0; 
}
#endif /* CONFIG_PROC_FS */

#endif /* CVM_IPFWD_DRIVER_OFFLOAD */

