#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/init.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/string.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <net/dst.h>
#include <linux/if_vlan.h>
#include <linux/sysctl.h>
#include <linux/errno.h>
#ifdef CONFIG_XFRM
	#include <linux/xfrm.h>
	#include <net/xfrm.h>
#endif  /* CONFIG_XFRM */

#include<linux/types.h>
#ifdef CVM_IPFWD_DRIVER_OFFLOAD
	#include <linux/tcp.h>
#endif /* CVM_IPFWD_DRIVER_OFFLOAD */
#include "ethernet-defines.h"
#include "ethernet-headers.h"

/**
 *
 * IPFWD/NAT OFFLOAD DATA STRCUTURES
 *
 */

/* Ethernet header structure */
typedef struct
{
	uint32_t enet_dst_hi;
	uint16_t enet_dst_lo;
	uint16_t enet_src_hi;
	uint32_t enet_src_lo;
	uint16_t enet_type;
} cvm_ipfwd_eh_t;

/* Flow entry structure (24 bytes) in the hash bucket */
typedef union
{
	struct
	{
		uint64_t word0;
		uint64_t word1;
		uint64_t word2;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
		uint32_t word3;
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	} u;

	struct
	{
		/* Word 0 */
		uint8_t          id0       : 2;	  /* synchronization id for word 0*/
		uint8_t          reserved0 : 6;
		uint8_t          ip_proto;		  /* orig IP protocol */
		uint16_t         l4_sport;		  /* orig L4 src port */
		uint32_t         ip_saddr;		  /* orig IP src addr */

		/* Word 1 */
		uint16_t         id1       : 2;	  /* synchronization id for word 1*/
		uint16_t         reserved1 : 14;
		uint16_t         l4_dport;		  /* orig L4 dst port */
		uint32_t         ip_daddr;		  /* orig IP dst addr */

		/* Word 2 */
		uint32_t         id2       : 2;	  /* synchronization id for word 2*/
		uint32_t         reserved2 : 2;
		uint32_t         timestamp : 28;  /* lifetime timestamp */
		uint32_t         flow_info;		  /* packet info ptr */
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
        /* Word 3 */
        uint16_t        vlan_info;        /* Stores the Vlan id and priority */
        uint16_t		index;			  /* bucket index */
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	} s;
} cvm_ipfwd_flow_entry_t;

/* Bucket header structure (64 bits) */
typedef struct
{
	spinlock_t           wr_lock;		 /* bucket lock */
	uint16_t             reserved;
	uint16_t             timestamp;		 /* bucket timetamp */
} cvm_ipfwd_flow_bucket_header_t;

/* Hash bucket (i.e., flow LUT entry) */
typedef struct
{
	cvm_ipfwd_flow_bucket_header_t hdr;
	cvm_ipfwd_flow_entry_t         entry[CVM_IPFWD_ENTRIES_PER_BUCKET];
} cvm_ipfwd_flow_bucket_t;

/* Flow info structure (pointed to by flow_info ptr in the flow entry
 * structure) */
typedef struct
{
	uint32_t          enet_dst_hi;	/* new ethernet dst addr */
	uint16_t          enet_dst_lo;
	uint16_t          enet_src_hi;	/* new ethernet src addr */
	uint32_t          enet_src_lo;
#ifdef CONFIG_TF1_HW_OFFLOAD
#ifdef CONFIG_VLAN_8021Q
    uint16_t          vlan_info;    /* vlan id,cfi and priority */
    uint16_t          vlan_present; /* is 1, If packet is of 8021Q type 
                                       else the value will be zero*/
#endif
#endif /*CONFIG_TF1_HW_OFFLOAD*/
	uint32_t          ip_saddr;		/* new ip src addr */
	uint32_t          ip_daddr;		/* new ip dst addr */
	uint8_t           ip_proto;		/* new ip protocol */
	uint32_t          ports;		/* new L4 ports */
	struct net_device *dev;
	unsigned long 	  lastused; 	
} cvm_ipfwd_flow_info_t;

/* Packet info structure (stored at the bottom of the packet buffer) */
typedef struct {
        uint64_t                reserved0;     /* vlan present/not */
	uint64_t                cookie;		   /* magic number */
	cvm_ipfwd_flow_bucket_t *bucket;	   /* bucket ptr */
	uint32_t                seq;		   /* TCP sequence number (if present) */
	uint32_t                ack_seq;	   /* TCP acknowledge number (if present) */
    uint32_t                reserved1;     /* old vlan info */
	struct iphdr            ip;			   /* IP header */
	uint64_t                reserved2[4];  /* L4 ports and reserved */
} cvm_packet_info_t;

/**
 *
 * IPFWD/NAT GLOBAL VARIABLES
 *
 */


typedef struct
{
	uint64_t res0 : 48;
	uint64_t idx  : 16;
	void     *cvm_ipfwd_ptr_fifo[CVM_IPFWD_PTR_FIFO_SIZE];
	uint8_t  res1[256 - (sizeof(uint64_t) + (sizeof(void *) * CVM_IPFWD_PTR_FIFO_SIZE))];
} cvm_ipfwd_fifo_t;

/* Per-core queue for freed flow_info buffers */
static cvm_ipfwd_fifo_t free_fifo[NR_CPUS] = {{0}};

   
/**
 *
 * IPFWD/NAT OFFLOAD INLINE FUNCTIONS
 *
 */

