/* ipt_BWPROF.c - bandwidth profile module */

/* Copyright (c) 2007, TeamF1, Inc. */

/*
modification history
--------------------
01f,10apr07,rnp removed unused code
01e,06Apr07,ayz fixed issue of invalidating class id on correct interface
01d,02Apr07,ayz fixed issue with large classid and classid wrapping, using stack
01c,30mar07,rnp fixed issue with large classid and classid wrapping
01b,20Mar07,ayz fixed FTP ALG issue
01a,28Feb07,rnp created
*/

/* This is a module which is used for setting the TC_INDEX field of an skb. */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/rtnetlink.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ipt_BWPROF.h>
#include <linux/netfilter_ipv4/ipt_LOG.h>

MODULE_LICENSE("Proprietary");
MODULE_DESCRIPTION("iptables BWPROF bandwidth profile module");

BW_MGMT bwMgmt;
EXPORT_SYMBOL(bwMgmt);

static u_int32_t debug_log = 0xFFFFFFF3; /* Enable ALL-type INFO-level logs */
static struct list_head classTbl;
static spinlock_t       classTblLock;

#define BWPROF_IFINDEX(x,y)  ((BW_PROF_INFO*)x->bwprof.bwprofinfo)->ifindex[y]

/* tc index stack variables */
#define BW_IS_STACK_EMPTY (topOfStack != BW_STACK_EMPTY)  

/* this is size of tc index stack in case we hit the starvation due to
 * unavailibilty of class id in stack we repopulate the stack and increase the
 * size of stack by BW_SIZEOF_STACK every time to make room availeble for
 * occupied class id's to pushed when free'ed */

#define BW_SIZEOF_STACK 0x000f

#define BW_STACK_EMPTY -1
#define BW_ERROR -1
#define BW_OK     0

/* class id 0 and 1  are invalid class id's so pick 0 to say it */
#define BW_INVALID_CLASSID 0x0

uint32_t *pClassIdStackPool = NULL ;
static int topOfStack = BW_STACK_EMPTY;
static int stackSize= 0;    

/* the maximum limit of tc index is 0xffff(16-bit) but as 0xffff is default we
 * use till 0xfffe */

#define BW_MAX_LIMIT_OF_STACK_SIZE  ( 0xfffe )

/* 16 values from 0 to 15 are available */
#define BWPROF_LOG_NONE    0x0
#define BWPROF_LOG_ERROR   0x1
#define BWPROF_LOG_WARN    0x2
#define BWPROF_LOG_INFO    0x3
#define BWPROF_LOG_DEBUG   0x4
#define BWPROF_LOG_DUMP    0x5
#define BWPROF_LOG_ALL     0xF

/* 28 values, one for each bit is available */
#define BWPROF_LOG_TYPE_NONE    0x0000000
#define BWPROF_LOG_TYPE_GENERIC 0x0000001
#define BWPROF_LOG_TYPE_REFCNT  0x0000002
#define BWPROF_LOG_TYPE_TCINDEX 0x0000004
#define BWPROF_LOG_TYPE_TC_STACK   0x0000008
#define BWPROF_LOG_TYPE_ALL     0xFFFFFFF

#define BWPROF_LOG_LEVEL (debug_log & 0xF)
#define BWPROF_LOG_TYPE ((debug_log & 0xFFFFFFF0) >> 4)

#define BWPROF_PRINTK(type, level, x...)    \
    do { \
        if (BWPROF_LOG_LEVEL > level && (BWPROF_LOG_TYPE & type)) \
            printk(x); \
    } while (0)

#define BWPROF_ERROR(x...)    \
    do { \
        BWPROF_PRINTK(BWPROF_LOG_TYPE_GENERIC, BWPROF_LOG_ERROR, \
                      KERN_ERR "BWPROF:" x); \
    } while (0)
#define BWPROF_WARN(x...)    \
    do { \
        BWPROF_PRINTK(BWPROF_LOG_TYPE_GENERIC, BWPROF_LOG_WARN, \
                      KERN_WARNING "BWPROF:" x); \
    } while (0)
#define BWPROF_INFO(x...)    \
    do { \
        BWPROF_PRINTK(BWPROF_LOG_TYPE_GENERIC, BWPROF_LOG_INFO, \
                      KERN_INFO "BWPROF:" x); \
    } while (0)
#define BWPROF_DEBUG(x...)    \
    do { \
        BWPROF_PRINTK(BWPROF_LOG_TYPE_GENERIC, BWPROF_LOG_DEBUG, \
                      "BWPROF:" x); \
    } while (0)
#define BWPROF_DUMP(x...)    \
    do { \
        BWPROF_PRINTK(BWPROF_LOG_TYPE_GENERIC, BWPROF_LOG_DUMP, \
                      "BWPROF:" x); \
    } while (0)

typedef struct classEntry
    {
    struct list_head  node;
    void *            pIdx;
    int               idxLen;
    uint32_t          classid;
    uint32_t          bwprofId;
    atomic_t          refcnt;
    int               status;
    } CLASS_ENTRY;

static unsigned char * bwprofIdx2Str
    (
    void * pIdx,
    int    idxLen
    )
    {
    static unsigned char buf[64];
    int i = 0;
    int offset = 0;
    unsigned char * data = pIdx;
    for(i = 0; i < idxLen; i++)
        offset += snprintf(buf+offset, sizeof(buf)-offset, "%02x", data[i]);
    return buf;
    }



static int bwPushStack(int classid)
    {
    /* 
     * we allow to push if stack is empty i.e top is -1 and disallow if 
     * we are crossing the limits 
     */
    
    if( topOfStack >= stackSize  )
        return BW_ERROR;

    pClassIdStackPool[++topOfStack]=classid;
    BWPROF_PRINTK(BWPROF_LOG_TYPE_TC_STACK, BWPROF_LOG_DEBUG, 
                  "Pushed classid: %x. Stack Top: %d\n",
                  classid, topOfStack);


    return BW_OK;
    }

static uint32_t bwPopStack (void)
    {
    uint32_t classid;
    
    if( BW_STACK_EMPTY == topOfStack )
        return BW_INVALID_CLASSID;
    
    classid=pClassIdStackPool[topOfStack--];
    BWPROF_PRINTK(BWPROF_LOG_TYPE_TC_STACK, BWPROF_LOG_DEBUG,
                  "Popped classid: %x. Stack Top: %d\n",
                  classid, topOfStack);

    return classid;
    }

  

static int bwPopulateStack (void)
     {
    static uint32_t populateClassid = 2;

    if(BW_IS_STACK_EMPTY)
        {
        BWPROF_ERROR("Not populating stack because it is not empty\n");
        return BW_ERROR ;
        }

    if(stackSize > BW_MAX_LIMIT_OF_STACK_SIZE)
        {
        BWPROF_WARN("Reached maximum stack size.\
                    Cannot populate stack anymore\n");
        return BW_ERROR;
        }

    if(pClassIdStackPool)
        kfree(pClassIdStackPool);
    
    /* increase stack size every time we populate stack by BW_SIZEOF_STACK */

    stackSize+=BW_SIZEOF_STACK;
    pClassIdStackPool=(uint32_t *)kmalloc(sizeof(uint32_t)*stackSize, 
                                          GFP_KERNEL);

    while(topOfStack  < (BW_SIZEOF_STACK - 1))
        pClassIdStackPool[++topOfStack]=populateClassid++;

    BWPROF_WARN("Repopulating Stack Size of Stack now is : \
                %d  total elements in stack :%d  \n", 
                stackSize, topOfStack );
    
    return BW_OK;
    }

static uint32_t bwprofGetFreeClassid(void)
    {
    uint32_t freeClassid = BW_INVALID_CLASSID ;

    /* we use stack to get free class id , here when  we  free class id we
     * push it to stack , and when we need it we pop it so in case if pop fails
     * and we need free class id  then we go for repopulating the stack in this 
     * scheme  we reuse the free'ed class id */

    freeClassid = bwPopStack();
    if( BW_INVALID_CLASSID == freeClassid )
        {
        if(  BW_OK == bwPopulateStack() )
            freeClassid = bwPopStack();
        }

    return freeClassid;

    }

static void * bwprofPkttoIndex
    (
    struct sk_buff * skb, 
    BW_PROF_INFO *   bwprofinfo, 
    int *            pIdxLen
    )
    {
    void * pIdx = NULL;
    switch(bwprofinfo->type)
        {
        case BWPROF_TYPE_ALL:
            {
            /* all packets that match this rule will have same index */
            /* 
             * we use the ID of bwprofinfo as index because it is unique
             * for each rule
             */ 
            pIdx = kmalloc(sizeof(bwprofinfo->id), GFP_KERNEL);
            memcpy(pIdx, &bwprofinfo->id, sizeof(bwprofinfo->id));
            *pIdxLen = sizeof(bwprofinfo->id);
            break;
            }
        case BWPROF_TYPE_PER_SRC_IP:
            {
            /* 
             * all packets that match this rule and have same source IP will
             * have same index 
             */
            /* 
             * we use the concatenation of packet source IP and ID of
             * bwprofinfo as the index
             */ 
            pIdx = kmalloc(sizeof(skb->nh.iph->saddr)+sizeof(bwprofinfo->id),
                           GFP_KERNEL);
            memcpy(pIdx, &skb->nh.iph->saddr, sizeof(skb->nh.iph->saddr));
            memcpy(pIdx+sizeof(skb->nh.iph->saddr), 
                   &bwprofinfo->id, sizeof(bwprofinfo->id));
            *pIdxLen = sizeof(skb->nh.iph->saddr)+sizeof(bwprofinfo->id);
            break;
            }
        case BWPROF_TYPE_PER_DST_IP:
            {
            break;
            }
        default:
            break;
        }

    BWPROF_DUMP("Packet index(%d bytes): 0x%s\n", *pIdxLen,
                bwprofIdx2Str(pIdx, *pIdxLen));

    return pIdx;
    }
    
int bwprofClassByIdGet
    (
    uint32_t       classId, 
    BW_PROF_INFO * pBwProf
    )
    {
    struct list_head * pNode = NULL;
    CLASS_ENTRY * pClass = NULL;

    if (bwMgmt.classCtl == NULL)
        return 0;

    /* input checks */
    if (pBwProf == NULL)
        {
        BWPROF_ERROR("No profile in connection which has classId: %x\n",
                     classId);
        return -1;
        }

    if(classId == BWPROF_NO_TC_INDEX)
       return 0;

    
    spin_lock_bh(&classTblLock);
    list_for_each(pNode, &classTbl)
        {
        pClass = (CLASS_ENTRY *)pNode;
        if (pClass->classid == classId)
            break;
        }
    spin_unlock_bh(&classTblLock);

    if (pNode == &classTbl)
        {
        BWPROF_ERROR("Class not found for related connection: %x\n", classId);
        return -1;
        }
    
    BWPROF_DUMP("Found class(%x) for Packet index(%d bytes): 0x%s\n", 
                classId, pClass->idxLen, 
                bwprofIdx2Str(pClass->pIdx, pClass->idxLen));

    /* increment refnct and return */
    atomic_inc(&pClass->refcnt);
    BWPROF_PRINTK(BWPROF_LOG_TYPE_REFCNT, BWPROF_LOG_DEBUG,
                  "Refcnt incremented for class(%x): 0x%x\n",
                  classId, atomic_read(&pClass->refcnt));
    return 0;
    }

/*
 * Get the number of class instances for the give bandwidth profile
 */
int bwprofNumClassInstancesGet
    (
    int bwprofId
    )
    {
    struct list_head * pNode = NULL;
    CLASS_ENTRY * pClass = NULL;
    int numInstances=0;

    spin_lock_bh(&classTblLock);
    list_for_each(pNode, &classTbl)
        {
        pClass = (CLASS_ENTRY *)pNode;

        if (pClass->bwprofId == bwprofId)
            numInstances++;
        }
    spin_unlock_bh(&classTblLock);

    return (numInstances);
    }

/*
 * Get the lease used class which has the bandwidth profile indicated by
 * bwprofId.
 */
CLASS_ENTRY * bwprofLeastUsedClassGet
    (
    int bwprofId
    )
    {
    int prev_refcnt=0;
    struct list_head * pNode = NULL;
    CLASS_ENTRY * pClass = NULL;
    CLASS_ENTRY * pLeastUsedClass = NULL;

    spin_lock_bh(&classTblLock);
    list_for_each(pNode, &classTbl)
        {
        pClass = (CLASS_ENTRY *)pNode;

        if (pClass->bwprofId == bwprofId)
            {
            /* first match */
            if (!prev_refcnt) 
                pLeastUsedClass = pClass;
            else if (prev_refcnt > atomic_read(&pClass->refcnt))
                pLeastUsedClass = pClass;
            }

        prev_refcnt = atomic_read(&pClass->refcnt);
        }
    spin_unlock_bh(&classTblLock);

    return (pLeastUsedClass);
    }

int bwprofClassByPktIndexGet
    (
    uint8_t * pIdx, 
    int idxLen, 
    uint32_t * pClassId, 
    BW_PROF_INFO * pBwProf
    )
    {
    int result = 0;
    uint32_t classid = 0;
    struct list_head * pNode = NULL;
    CLASS_ENTRY * pClass = NULL;
    int i = 0;
    int numOfInstances=0;

    if (bwMgmt.classCtl == NULL)
        return 0;

    /* input checks */
    if (pBwProf == NULL)
        return -1;
         
    if (pBwProf->type == BWPROF_TYPE_PER_SRC_IP)
        {
        /* For an individual type of profile, check the number of class
         * instances created by this profile.
         */
        numOfInstances = bwprofNumClassInstancesGet(pBwProf->id);

        /* if number of instances is more than the maximum, then re-use one
         * of the class created by the profile in a RR manner(kind of..)
         */
         if ((pBwProf->maxInstances > 0) && (numOfInstances > 0) &&
             (numOfInstances >= pBwProf->maxInstances))
             {
             /* find a class that has a profile Id (pBwProf->id) and
              * has the least refcount (less used).
              */
             pClass = bwprofLeastUsedClassGet(pBwProf->id);
             if (pClass) 
                 classid = pClass->classid;
             }
        }

    if (pClass == NULL)
        {
        /* lookup pIdx in table */
        spin_lock_bh(&classTblLock);
        list_for_each(pNode, &classTbl)
            {
            pClass = (CLASS_ENTRY *)pNode;
            if ((pClass->idxLen == idxLen) &&
                (memcmp(pClass->pIdx, pIdx, idxLen) == 0))
                break;
            }
        spin_unlock_bh(&classTblLock);

        if (pNode != &classTbl)
            {
            BWPROF_DUMP("Found class(%x) for Packet index(%d bytes): 0x%s\n", 
                        pClass->classid, idxLen, bwprofIdx2Str(pIdx, idxLen));
            classid = pClass->classid;
            }
        else
            {
            pClass = NULL;

            /* find a free classid */
            classid = bwprofGetFreeClassid();
            if( BW_INVALID_CLASSID == classid)
                {
                BWPROF_ERROR(" Unable to find free class id\n");
                return BW_ERROR;
                }
            }
        }

    /* 
     * We should attempt to create the class(in the qdisc layeer) on every class
     * lookup. This will ensure that the class is always present and thus
     * workaround situations where classes get deleted automatically(ex: when
     * ppp interfaces disappear). If class already exists, then we ignore that
     * error.
     */ 

    BWPROF_DEBUG("Creating %sa new class of classid: %u\n", 
                 pClass ? "again " : " ", classid);
    for (i = 0; i < BWPROF_MAX_IF; i ++)
        {
        if (pBwProf->ifindex[i] == BWPROF_INVALID_IFINDEX)
            continue;

        BWPROF_DEBUG("Creating %sa new class(%x) on if(%s)\n",
                     pClass ? "again " : " ", classid, pBwProf->ifName[i]);
        result = bwMgmt.classCtl(classid,
                                 RTM_NEWTCLASS,pBwProf->ifindex[i],
                                 pBwProf->minbw, pBwProf->maxbw);
        if (result < 0 && result != -EEXIST)
            {
            BWPROF_ERROR("Failed to add class(%x) to if(%s): %d\n", classid,
                         pBwProf->ifName[i], result);
            /* push back the classid, since we failed to create the class */
            bwPushStack(classid);
            return result;
            }
        else
            {
            if(bwMgmt.qdiscCtl)
                result = bwMgmt.qdiscCtl(RTM_NEWQDISC,classid,
                                         pBwProf->ifindex[i]);
            if(result < 0 && result != -EEXIST)
                {
                BWPROF_ERROR("Failed to add Qdisc for parent (%x) to if(%s): "
                             "%d\n", classid, pBwProf->ifName[i], result);
                return result;
                }
            }
        if (result == -EEXIST)
            {
            BWPROF_DEBUG("Class(%x) already exists for Packet index(%d bytes): "
                         "0x%s\n", classid, idxLen, 
                         bwprofIdx2Str(pIdx, idxLen));
            }
        else if (pClass != NULL)
            {
            BWPROF_DEBUG("Created missing class(%x) for Packet index(%d bytes):"
                         " 0x%s\n", classid, idxLen, 
                         bwprofIdx2Str(pIdx, idxLen));
            }
        }

    if (pClass == NULL)
        {
        /* add entry for class and entry into table */
        pClass = kmalloc(sizeof(*pClass), GFP_KERNEL);
        if (pClass != NULL)
            {
            memset(pClass, 0, sizeof(*pClass));
            pClass->idxLen = idxLen;
            pClass->pIdx = kmalloc(idxLen, GFP_KERNEL);
            if (pClass->pIdx != NULL)
                memcpy(pClass->pIdx, pIdx, idxLen);
            pClass->classid = classid;
            pClass->bwprofId = pBwProf->id;
            spin_lock_bh(&classTblLock);
            list_add(&pClass->node, &classTbl); 
            spin_unlock_bh(&classTblLock);
            BWPROF_DEBUG("Added class(%x) for Packet index(%d bytes): 0x%s\n", 
                         classid, idxLen, bwprofIdx2Str(pIdx, idxLen));
            }
        }

    /* increment refnct and return */
    atomic_inc(&pClass->refcnt);
    *pClassId = pClass->classid;
    BWPROF_PRINTK(BWPROF_LOG_TYPE_REFCNT, BWPROF_LOG_DEBUG,
                  "Refcnt incremented for class(%x): 0x%x\n",
                  pClass->classid, atomic_read(&pClass->refcnt));
    return 0;
    }

int bwprofClassPut
    (
    uint32_t classid,
    BW_PROF_INFO * pBwProf
    )
    {
    int result = 0;
    struct list_head * pNode = NULL;
    CLASS_ENTRY * pClass = NULL;
    int i = 0;

    if (bwMgmt.classCtl == NULL)
        return 0;

    /* 
     * We never use zero classid(NO_TCINDEX). Ignore silently. This can happen
     * when connection tracking entries which have no BWPROF associated with
     * them, get destroyed. We also ignore INVALID_TCINDEX
     */ 
    if (classid == BWPROF_NO_TC_INDEX || 
        classid == BWPROF_INVALID_TC_INDEX)
        return 0;

    /* lookup classid in table */
    spin_lock_bh(&classTblLock);
    list_for_each(pNode, &classTbl)
        {
        pClass = (CLASS_ENTRY *)pNode;
        if (pClass->classid == classid)
            break;
        }
    spin_unlock_bh(&classTblLock);

    /* if classid not found return */
    if (pClass == NULL)
        {
        BWPROF_ERROR("Failed to release class(%x): Not Found\n", classid);
        return -ENOENT;
        }

    if (pClass->bwprofId != pBwProf->id)
        {
        BWPROF_ERROR("Failed to release class(%x): Profile ID in class(%lu) "
                     "mismatched with Profile ID in connection\n",
                     pClass->bwprofId, pBwProf->id);
        return -EINVAL;
        }

    /* decrement refcnt for classid. If refcnt is zero destroy class */
    if (atomic_dec_and_test(&pClass->refcnt) == 0)
        {
        BWPROF_PRINTK(BWPROF_LOG_TYPE_REFCNT, BWPROF_LOG_DEBUG,
                      "Refcnt decremented for class(%x): 0x%x\n",
                      pClass->classid, atomic_read(&pClass->refcnt));
        return 0;
        }
    else
        {
        BWPROF_PRINTK(BWPROF_LOG_TYPE_REFCNT, BWPROF_LOG_DEBUG,
                      "Refcnt decremented for class(%x): 0x%x\n",
                      pClass->classid, atomic_read(&pClass->refcnt));
        }

    /* delete the class */
    for (i = 0; i < BWPROF_MAX_IF; i ++)
        {
        if (pBwProf->ifindex[i] == BWPROF_INVALID_IFINDEX)
            continue;
        /* when class get deleted it child sfq autometicaly gets deleted so 
         * we dont to add code for that */
        result = bwMgmt.classCtl(pClass->classid, RTM_DELTCLASS,
                                 pBwProf->ifindex[i], 0, 0);
        if (result < 0)
            {
            BWPROF_ERROR("Failed to delete class(%x) on if(%s): %d\n", classid,
                         pBwProf->ifName[i], result);
            /* this can happen if
             * - interface is removed(ex: ppp), in which case class gets deleted
             *   automatically
             * - the ifindex we are using now is stale(i.e. interface indices
             *   have been reshuffled since the class was created). This perhaps
             *   needs more thinking to see if its a grave issue.
             */   
            }
        }

    /* remove the class entry */
    BWPROF_DEBUG("Removing class(%x) for Packet index(%d bytes): 0x%s\n",
                 classid, pClass->idxLen,
                 bwprofIdx2Str(pClass->pIdx, pClass->idxLen));
    list_del(&pClass->node);
    kfree(pClass->pIdx);
    memset(pClass, 0, sizeof(*pClass));
    kfree(pClass);
    if(bwPushStack(classid) == BW_ERROR)
        {
        BWPROF_ERROR(" Push failed \n");
        return BW_ERROR;
        }
    return 0;
    }


int  bwprof_invalidate_tc_index(struct sk_buff *skb,char *pDevName)
    {
    int ifCounter =0;
    struct ip_conntrack * ct = NULL;
    enum ip_conntrack_info ctinfo;
    ct = ip_conntrack_get(skb, &ctinfo);

    while(ifCounter < BWPROF_MAX_IF && ct)
        {
        /* here we are invalidating the tc_index if
         * interface exists in our list */
        
        if( skb->dev->ifindex == BWPROF_IFINDEX(ct,ifCounter)) 
            {
            bwprofClassPut(ct->bwprof.tc_index[skb->ctdir],
                           ct->bwprof.bwprofinfo);
            ct->bwprof.tc_index[skb->ctdir] = BWPROF_INVALID_TC_INDEX;

            printk("Invalidating connection index: %x \n",
                   ct->bwprof.tc_index[skb->ctdir]);
            break;
            }
        ifCounter++;
        }

    if(ct)
        printk("No class found on %s for packet(%s) with index: %x.\n",
               pDevName,(skb->ctdir == BWPROF_DIR_ORIGINAL) ? 
               "ORIGINAL" : "REPLY",
               skb->tc_index);   
    return 0;
    }

static unsigned int bwprofInitCt
    (
    struct ip_conntrack * ct, 
    BW_PROF_INFO * bwprofinfo,
    int dir
    )
    {
    BW_PROF_INFO * ctbwinfo = ct->bwprof.bwprofinfo;

    /* 
     * if the connection already has bwprof information which is the same
     * current bwprofinfo, then we dont disturb the connection. This can happen
     * with UDP one-way traffic. We check for matching id in bwprofinfo before
     * clearing the connection. The id mismatch will occur if the connection
     * hits a different BWPROF compared to what is saved in the connection and
     * so the connection must be updated with params of the new BWPROF rule.
     */
    if (ctbwinfo != NULL)
        {
        BWPROF_PRINTK(BWPROF_LOG_TYPE_TCINDEX, BWPROF_LOG_DEBUG,
                      "initct: already initialized tc_index[ORIGINAL]:%x, "
                      "tc_index[REPLY]%x, id:%lu pktdir=%s\n",
                      ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL],
                      ct->bwprof.tc_index[BWPROF_DIR_REPLY], 
                      ctbwinfo->id,
                      (dir == BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : 
                      ((dir == BWPROF_DIR_REPLY) ? "REPLY" : "UNKNOWN"));
        /* id is same, so the connection doesnt need to be initialized */
        if (ctbwinfo->id == bwprofinfo->id)
            return 0;
        /* release reference to classes and re-initialize connection(below) */
        bwprofClassPut(ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL],
                       ct->bwprof.bwprofinfo);
        bwprofClassPut(ct->bwprof.tc_index[BWPROF_DIR_REPLY],
                       ct->bwprof.bwprofinfo);
        memset(ctbwinfo, 0, sizeof(*ctbwinfo));
        }

    /* create new ctbwinfo */
    if (ctbwinfo == NULL)
        {
        ctbwinfo = kmalloc(sizeof(*ctbwinfo), GFP_KERNEL);
        memset(ctbwinfo, 0, sizeof(*ctbwinfo));
        ct->bwprof.bwprofinfo = ctbwinfo;
        }

    /* 
     * When we see the first packet, we mark the ct->tc_index[dir] as
     * invalid. For bwprofinfo->dir is ORIGINAL, the ct->tc_index[ORIG] will
     * be updated further below. For bwprofinfo->dir is REPLY, the
     * ct->tc_index[REPLY] will be updated when reply packet hits the --dir
     * NONE rule. Likewise when bwprofinfo->dir is BOTH
     */
    *ctbwinfo = *bwprofinfo;
    if (ctbwinfo->dir == BWPROF_DIR_BOTH)
        {
        ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL] = BWPROF_INVALID_TC_INDEX;
        ct->bwprof.tc_index[BWPROF_DIR_REPLY] = BWPROF_INVALID_TC_INDEX;
        }
    else
        ct->bwprof.tc_index[bwprofinfo->dir] = BWPROF_INVALID_TC_INDEX;
    BWPROF_DEBUG("initct: tc_index[ORIGINAL]:%x, tc_index[REPLY]%x, "
                 "id:%lu pktdir=%s\n",
                 ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL],
                 ct->bwprof.tc_index[BWPROF_DIR_REPLY], 
                 ctbwinfo->id,
                 (dir == BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : 
                 ((dir == BWPROF_DIR_REPLY) ? "REPLY" : "UNKNOWN"));
    return 0;
    }

static unsigned int bwprofProcessInvalidTcIndex
    (
    struct sk_buff * skb, 
    struct ip_conntrack * ct, 
    int dir
    )
    {
    int result = 0;
    uint8_t * pIdx = NULL;
    int idxLen = 0;
    BW_PROF_INFO * bwprofinfo = ct->bwprof.bwprofinfo;

    if (bwprofinfo == NULL)
        {
        BWPROF_ERROR("Connection in %s dir has no bwprof, but has "
                     "INVALID_TC_INDEX\n",
                     (dir==BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : "REPLY");
        return 0;
        }
    /* 
     * use custom ct_tc_index if available i.e. no dynamic class creation. This
     * option is helpful for debugging
     */
    if (bwprofinfo->ct_tc_index != 0)
        {
        ct->bwprof.tc_index[dir] = bwprofinfo->ct_tc_index;
        return 0;
        }

    /* compute index from packet(i.e. pkt_index) */
    pIdx = bwprofPkttoIndex(skb, bwprofinfo, &idxLen);
    if (pIdx != NULL && idxLen != 0)
        {
        /* create/find a class for pkt_index */
        BWPROF_DEBUG("Create/Find class in %s direction for "
                     "Packet index(%d bytes): 0x%s\n",
                     (dir == BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : "REPLY",
                     idxLen, bwprofIdx2Str(pIdx, idxLen));
        result = bwprofClassByPktIndexGet(pIdx, idxLen,
                                          (uint32_t *)&ct->bwprof.tc_index[dir],
                                          (BW_PROF_INFO *)bwprofinfo);
        if (result < 0)
            BWPROF_ERROR("Failed to get class in %s direction for "
                         "Packet index(%d bytes): 0x%s\n",
                         (dir==BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : "REPLY",
                         idxLen, bwprofIdx2Str(pIdx, idxLen));
        }

    return 0;
    }

static unsigned int
bwprof_target(struct sk_buff **pskb,
       const struct net_device *in,
       const struct net_device *out,
       unsigned int hooknum,
       const struct xt_target *target,
       const void *targinfo)
    {
    BW_PROF_INFO *bwprofinfo = (BW_PROF_INFO *)targinfo;

    struct ip_conntrack *ct;
    enum ip_conntrack_info ctinfo;
    __u32 dir = (*pskb)->ctdir;

    ct = ip_conntrack_get(*pskb, &ctinfo);
    if (ct == NULL)
        {
        /* 
         * if there is no connection tracking this packet, we cant do any
         * limiting with it. We just accept the packet, without ado.
         */ 
        return NF_ACCEPT;
        }

    /* 
     * Initial connection with bwprof infomation if it its proper rule(i.e. dir
     * is not NONE). This should happen only once for each connection
     */ 
    if (unlikely(bwprofinfo->dir != BWPROF_DIR_NONE))
        bwprofInitCt(ct, bwprofinfo, dir);

    /* 
     * if the index for stream is INVALID, then lookup/create a class for this
     * stream. If class is found/created, tc_index[dir] will contain that
     * classid. tc_index[dir] is invalid in following cases:
     * 1. dir is ORIGINAL and bwprofinfo->dir is ORIGINAL/BOTH and *pskb is
     *    first packet of ORIGINAL stream
     * 2. dir is REPLY and bwprofinfo->dir is REPLY/BOTH and *pskb is first
     *    packet of REPLY stream
     * 3. *pskb is any random packet in an established connection and "tc_index
     *    filter" invalidates the index, when doesnt find a class. In this case,
     *    subsequent packets in that direction will have tc_index[dir] as
     *    INVALID
     */ 
    if (unlikely(ct->bwprof.tc_index[dir] == BWPROF_INVALID_TC_INDEX))
        bwprofProcessInvalidTcIndex(*pskb, ct, dir);

    /* update index in packet */
    (*pskb)->tc_index = ct->bwprof.tc_index[dir];
    BWPROF_PRINTK(BWPROF_LOG_TYPE_TCINDEX, BWPROF_LOG_DEBUG,
                  "tc_index[%s]: %x\n",
                   (dir==BWPROF_DIR_ORIGINAL) ? "ORIGINAL" : "REPLY",
                   ct->bwprof.tc_index[dir]);

    /* --log-prifix --log-level --log-format these option are given to log
     * target to take for logging the packets */

   if(unlikely((bwprofinfo->flags &  FLAG(IPT_BWPROF_OPT_LOG)) ==  FLAG(IPT_BWPROF_OPT_LOG)))
       {
       struct ipt_log_info loginfo ;
       memset(&loginfo,0x0,sizeof(loginfo));

       loginfo.level=bwprofinfo->logLevel;
       if((bwprofinfo->flags &  FLAG(IPT_BWPROF_OPT_LOG_PREFIX)) ==  FLAG(IPT_BWPROF_OPT_LOG_PREFIX)) 
           strncpy(loginfo.prefix,bwprofinfo->logPrefix,
                   BWPROF_MAX_LOG_PREFIX_SZ -1);

       if((bwprofinfo->flags &  FLAG(IPT_BWPROF_OPT_LOG_FORMAT)) ==  FLAG(IPT_BWPROF_OPT_LOG_FORMAT))
           strncpy(loginfo.format,bwprofinfo->logFormat,
                   BWPROF_MAX_LOG_FORMAT_SZ -1);

       ipt_log_target_ex(pskb,in,out,hooknum,&loginfo,targinfo);

       }
           
    /* we will accept the packets */
    return NF_ACCEPT;
    }

/* cleanup a connection tracking entry's bwprof information */
void bwprof_ct_cleanup(struct ip_conntrack * ct)
    {
    /* release classes used by this ct */
    bwprofClassPut(ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL],
                   ct->bwprof.bwprofinfo);
    bwprofClassPut(ct->bwprof.tc_index[BWPROF_DIR_REPLY],
                   ct->bwprof.bwprofinfo);
    ct->bwprof.tc_index[BWPROF_DIR_ORIGINAL] = BWPROF_NO_TC_INDEX;
    ct->bwprof.tc_index[BWPROF_DIR_REPLY] = BWPROF_NO_TC_INDEX;
    if (ct->bwprof.bwprofinfo)
        kfree(ct->bwprof.bwprofinfo);
    ct->bwprof.bwprofinfo = NULL;
    return;
    }

void bwprof_ct_copy(struct ip_conntrack * new,struct ip_conntrack * old)    
    {
    int result;
    if( old->bwprof.bwprofinfo)
        {
        /* 
         * copy the old conntrack entry data(control connection to new one(data
         * connection) so that tracffic of new conn track(data connection) 
         * should get bandwidth limited 
         */
        new->bwprof.bwprofinfo = kmalloc(sizeof(BW_PROF_INFO), GFP_KERNEL);
        memcpy(new->bwprof.bwprofinfo, old->bwprof.bwprofinfo,
               sizeof(BW_PROF_INFO));
        /* 
         * in case of ACTIVE FTP flag is set in FTP ALG when it detects that
         * data transfer is happening using ACTIVE mode. Accordingly we need to
         * reverse the direction. The flag is set in ip_conntrack_ftp.c 
         */

        if(old->bwprof.flag)
            {
            BWPROF_DEBUG("Active FTP transfer detected, "
                         "reversing the direction\n");
            new->bwprof.tc_index[IP_CT_DIR_ORIGINAL] = 
                    old->bwprof.tc_index[IP_CT_DIR_REPLY];
            new->bwprof.tc_index[IP_CT_DIR_REPLY] = 
                    old->bwprof.tc_index[IP_CT_DIR_ORIGINAL];
            }
        else
            {
            new->bwprof.tc_index[IP_CT_DIR_ORIGINAL] = 
                    old->bwprof.tc_index[IP_CT_DIR_ORIGINAL];
            new->bwprof.tc_index[IP_CT_DIR_REPLY] = 
                    old->bwprof.tc_index[IP_CT_DIR_REPLY];
            }

        /* 
         * lets increase the ref. count of class as now related connection is
         * sharing the same class as original connection.(ex: data conn. which
         * is part of FTP conn.)
         */

        result = bwprofClassByIdGet(new->bwprof.tc_index[IP_CT_DIR_ORIGINAL], 
                                    (BW_PROF_INFO *)new->bwprof.bwprofinfo);
        result = bwprofClassByIdGet(new->bwprof.tc_index[IP_CT_DIR_REPLY], 
                                    (BW_PROF_INFO *)new->bwprof.bwprofinfo);
        BWPROF_DEBUG("Copied BWPROF info from old ct to new one: %x,%x\n",
                     new->bwprof.tc_index[BWPROF_DIR_ORIGINAL],
                     new->bwprof.tc_index[BWPROF_DIR_REPLY]);
        }
    }

static int
bwprof_checkentry(const char *tablename,
		  const void *e_void,
		  const struct xt_target *target,
		  void *targinfo,
		  unsigned int hook_mask)
    {
    int i = 0;
    struct ipt_bwprof_target_info *bwprofinfo = targinfo;
    struct net_device * dev = NULL;
    static __u32 lastBwprofId = 0;

    /* 
     * if bwprofinfo->id is non-zero, then netfilter is internally deleting and
     * adding the rule again, in which case we should not re-init the variables
     */ 
    if (bwprofinfo->id == 0)
        {
        bwprofinfo->id = ++lastBwprofId;

        /* initialize to -1 */
        memset(bwprofinfo->ifindex, 0xff, sizeof(bwprofinfo->ifindex));

        for(i = 0; i < BWPROF_MAX_IF; i++)
            {
            if (bwprofinfo->ifName[i][0] == '\0')
                break;
            dev = dev_get_by_name(bwprofinfo->ifName[i]);
            if (dev == NULL)
                {
                BWPROF_ERROR("unknown if '%s'\n", bwprofinfo->ifName[i]);
                continue;
                }
            bwprofinfo->ifindex[i] = dev->ifindex;
            dev_put(dev);
            }
        BWPROF_DEBUG("Added BWPROF rule with id: %lu\n", bwprofinfo->id);
        }

    return 1;
    }

static void bwprof_destroy(const struct xt_target *target, void *targinfo)
    {
    return;
    }

static struct ipt_target ipt_bwprof_reg = {
	.name		= "BWPROF",
	.family		= AF_INET,
	.target		= bwprof_target,
	.targetsize	= sizeof(struct ipt_bwprof_target_info),
	.checkentry	= bwprof_checkentry,
	.destroy	= bwprof_destroy,
	.me		= THIS_MODULE,
};

#ifdef CONFIG_SYSCTL
/* must be unique */
#define NET_BWPROF 20
#define BWPROFCTL_DEBUG_LOG 1
static ctl_table bwprof_sysctl_table[] = {
	{
		.ctl_name	= BWPROFCTL_DEBUG_LOG,
		.procname	= "debug_log",
		.data		= &debug_log,
		.maxlen		= sizeof(u_int32_t),
		.mode		= 0644,
		.proc_handler	= &proc_dointvec,
	},
	{ .ctl_name = 0 }
};

static ctl_table bwprof_net_table[] = {
    {NET_BWPROF, "bwprof", NULL, 0, 0555, bwprof_sysctl_table},
    {0}
};

static ctl_table bwprof_root_table[] = {
    {CTL_NET, "net", NULL, 0, 0555, bwprof_net_table},
    {0}
};
static struct ctl_table_header * bwprof_sysctl_header;
#endif /* CONFIG_SYSCTL */

static int __init ipt_bwprof_init(void)
{
    spin_lock_init(&classTblLock);
    INIT_LIST_HEAD(&classTbl);
#ifdef CONFIG_SYSCTL
    bwprof_sysctl_header = register_sysctl_table(bwprof_root_table);
#endif
	return xt_register_target(&ipt_bwprof_reg);
}

static void __exit ipt_bwprof_fini(void)
{
#ifdef CONFIG_SYSCTL
    unregister_sysctl_table(bwprof_sysctl_header);
#endif
	xt_unregister_target(&ipt_bwprof_reg);
}

module_init(ipt_bwprof_init);
module_exit(ipt_bwprof_fini);
