/*
 * This is a module which is used for logging packets.
 */

/* (C) 1999-2001 Paul `Rusty' Russell
 * (C) 2002-2004 Netfilter Core Team <coreteam@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/icmp.h>
#include <net/udp.h>
#include <net/tcp.h>
#include <net/route.h>

#include <linux/netfilter.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_ipv4/ipt_LOG.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Netfilter Core Team <coreteam@netfilter.org>");
MODULE_DESCRIPTION("iptables syslog logging module");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#ifdef TEAMF1_CHANGES
/* 
 * Flag for enabling the logging of packet. It can have the following values:
 * 0 - Nothing logged in INPUT and OUTPUT chains
 * 1 - Log only UNICAST Packets in INPUT and OUTPUT chains
 * 2 - Log only BROADCAST and MULTICAST packets in INPUT and OUTPUT chains
 * 3 - Log UNICAST, BROADCAST and MULTICAST packets in INPUT and OUTPUT chains 
 */
int tf1_log_flag = 0;
#define FW_TPROXY_MARK_MASK 0x30000
#endif /* TEAMF1_CHANGES */

/* Use lock to serialize, so printks don't overlap */
static DEFINE_SPINLOCK(log_lock);

#ifdef TEAMF1_CHANGES
#define WAN     "WAN"
#define LAN     "LAN"
#define DMZ     "DMZ"
#define SELF    "SELF"    
#define FW_LAN_ZONE 1        
#define FW_DMZ_ZONE 2        
#define FW_WAN_ZONE 4               
#define LAN_IFACE "bdg1"


/*
 * This function returns the network device the packet would use to exit, if
 * it were not re-routed to the input by the tproxy rules 
 */
static struct rtable * tproxy_out_dev(const struct sk_buff *skb)
{
    const struct iphdr *iph = skb->nh.iph;
    struct dst_entry *odst;
    struct rtable *rt;
    struct sk_buff *m_skb = (struct sk_buff *)skb;
    int ret;
    __u32 oldmark;

    odst = m_skb->dst;
    m_skb->dst=NULL;

    oldmark = m_skb->mark;

    m_skb->mark = m_skb->mark & (~FW_TPROXY_MARK_MASK);

    if (0 != ip_route_input(m_skb, iph->daddr, iph->saddr, iph->tos,
                m_skb->dev)) {
        m_skb->dst = odst;
        m_skb->mark =  oldmark;
        return NULL;
    }

    rt = (struct rtable *)m_skb->dst;
    /* recover to the original dst */
    m_skb->dst = odst;

    /* recover the TPROXY MARK */
    m_skb->mark =  oldmark;

    if (!rt) return NULL;

    if (rt->u.dst.error || !rt->u.dst.dev) {
        dst_release(&rt->u.dst);
        return NULL;
    }

    return rt;
}

int printInterface (struct sk_buff *skb, int hooknum,
                    const struct net_device *dev, int direction,
                    char * logBuff)
{
	int i = -1;
    struct net_device *iface=NULL;
    struct rtable *rt=NULL;

    /* a) rule is hooked in the input path
     * b) packet has be re-routed by the tproxy rules in prerouting.
     * c) direction = 0 is output interface
     */
    if ((hooknum == NF_IP_LOCAL_IN) && !direction && 
                    (skb->mark & FW_TPROXY_MARK_MASK)) {
        rt = tproxy_out_dev(skb);
        if (rt) {
            dev = rt->u.dst.dev;
        }
    }
    
    iface = dev;

	if (!iface) {
        if (direction && (hooknum == NF_IP_LOCAL_OUT))
		    i = snprintf(logBuff,16," %s=%s ", direction ? "IN":"OUT", SELF);
        else
            i = 0;

        if (rt) dst_release(&rt->u.dst);    
        return i;
        }

	if (strncmp(iface->name,LAN_IFACE, sizeof(iface->name)) == 0 )
	    i = snprintf(logBuff,16," %s=%s ", direction ? "IN":"OUT", LAN);
    	else 
	    i = snprintf(logBuff,16," %s=%s ", direction ? "IN":"OUT", WAN);

    if (rt) dst_release(&rt->u.dst);    

	return i;
}
#endif /* TEAMF1_CHANGES */

/* One level of recursion won't kill us */
#ifdef TEAMF1_CHANGES
static int dump_packet(const struct nf_loginfo *info,
			const struct sk_buff *skb,
			unsigned int iphoff, char * logBuff)
#else /* TEAMF1_CHANGES */
static void dump_packet(const struct nf_loginfo *info,
			const struct sk_buff *skb,
			unsigned int iphoff)
#endif /* TEAMF1_CHANGES */
{
	struct iphdr _iph, *ih;
	unsigned int logflags;
#ifdef TEAMF1_CHANGES
	int len = -1;
	int totLen = 0;
#endif /* TEAMF1_CHANGES */

	if (info->type == NF_LOG_TYPE_LOG)
		logflags = info->u.log.logflags;
	else
		logflags = NF_LOG_MASK;

	ih = skb_header_pointer(skb, iphoff, sizeof(_iph), &_iph);
	if (ih == NULL) {
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,10,"TRUNCATED");
		totLen += len;
		return totLen;
#else /* TEAMF1_CHANGES */
		printk("TRUNCATED");
#endif /* TEAMF1_CHANGES */
	}

	/* Important fields:
	 * TOS, len, DF/MF, fragment offset, TTL, src, dst, options. */
	/* Max length: 40 "SRC=255.255.255.255 DST=255.255.255.255 " */
#ifdef TEAMF1_CHANGES
	len = snprintf (logBuff,60,"SRC=%u.%u.%u.%u DST=%u.%u.%u.%u ",
	       NIPQUAD(ih->saddr), NIPQUAD(ih->daddr));
	logBuff = logBuff + len;
	totLen += len;
#else /* TEAMF1_CHANGES */
	printk("SRC=%u.%u.%u.%u DST=%u.%u.%u.%u ",
	       NIPQUAD(ih->saddr), NIPQUAD(ih->daddr));

	/* Max length: 46 "LEN=65535 TOS=0xFF PREC=0xFF TTL=255 ID=65535 " */
	printk("LEN=%u TOS=0x%02X PREC=0x%02X TTL=%u ID=%u ",
	       ntohs(ih->tot_len), ih->tos & IPTOS_TOS_MASK,
	       ih->tos & IPTOS_PREC_MASK, ih->ttl, ntohs(ih->id));

	/* Max length: 6 "CE DF MF " */
	if (ntohs(ih->frag_off) & IP_CE)
		printk("CE ");
	if (ntohs(ih->frag_off) & IP_DF)
		printk("DF ");
	if (ntohs(ih->frag_off) & IP_MF)
		printk("MF ");

	/* Max length: 11 "FRAG:65535 " */
	if (ntohs(ih->frag_off) & IP_OFFSET)
		printk("FRAG:%u ", ntohs(ih->frag_off) & IP_OFFSET);

	if ((logflags & IPT_LOG_IPOPT)
	    && ih->ihl * 4 > sizeof(struct iphdr)) {
		unsigned char _opt[4 * 15 - sizeof(struct iphdr)], *op;
		unsigned int i, optsize;

		optsize = ih->ihl * 4 - sizeof(struct iphdr);
		op = skb_header_pointer(skb, iphoff+sizeof(_iph),
					optsize, _opt);
		if (op == NULL) {
			printk("TRUNCATED");
			return;
		}

		/* Max length: 127 "OPT (" 15*4*2chars ") " */
		printk("OPT (");
		for (i = 0; i < optsize; i++)
			printk("%02X", op[i]);
		printk(") ");
	}
#endif /* TEAMF1_CHANGES */

	switch (ih->protocol) {
	case IPPROTO_TCP: {
		struct tcphdr _tcph, *th;

		/* Max length: 10 "PROTO=TCP " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,15,"PROTO=TCP ");
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("PROTO=TCP ");
#endif /* TEAMF1_CHANGES */

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		th = skb_header_pointer(skb, iphoff + ih->ihl * 4,
					sizeof(_tcph), &_tcph);
		if (th == NULL) {
#ifdef TEAMF1_CHANGES
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
#else /* TEAMF1_CHANGES */
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
#endif /* TEAMF1_CHANGES */
			break;
		}

		/* Max length: 20 "SPT=65535 DPT=65535 " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,50,"SPT=%u DPT=%u ",
		       ntohs(th->source), ntohs(th->dest));
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("SPT=%u DPT=%u ",
		       ntohs(th->source), ntohs(th->dest));
		/* Max length: 30 "SEQ=4294967295 ACK=4294967295 " */
		if (logflags & IPT_LOG_TCPSEQ)
			printk("SEQ=%u ACK=%u ",
			       ntohl(th->seq), ntohl(th->ack_seq));
		/* Max length: 13 "WINDOW=65535 " */
		printk("WINDOW=%u ", ntohs(th->window));
		/* Max length: 9 "RES=0x3F " */
		printk("RES=0x%02x ", (u8)(ntohl(tcp_flag_word(th) & TCP_RESERVED_BITS) >> 22));
		/* Max length: 32 "CWR ECE URG ACK PSH RST SYN FIN " */
		if (th->cwr)
			printk("CWR ");
		if (th->ece)
			printk("ECE ");
		if (th->urg)
			printk("URG ");
		if (th->ack)
			printk("ACK ");
		if (th->psh)
			printk("PSH ");
		if (th->rst)
			printk("RST ");
		if (th->syn)
			printk("SYN ");
		if (th->fin)
			printk("FIN ");
		/* Max length: 11 "URGP=65535 " */
		printk("URGP=%u ", ntohs(th->urg_ptr));

		if ((logflags & IPT_LOG_TCPOPT)
		    && th->doff * 4 > sizeof(struct tcphdr)) {
			unsigned char _opt[4 * 15 - sizeof(struct tcphdr)];
			unsigned char *op;
			unsigned int i, optsize;

			optsize = th->doff * 4 - sizeof(struct tcphdr);
			op = skb_header_pointer(skb,
						iphoff+ih->ihl*4+sizeof(_tcph),
						optsize, _opt);
			if (op == NULL) {
				printk("TRUNCATED");
				return;
			}

			/* Max length: 127 "OPT (" 15*4*2chars ") " */
			printk("OPT (");
			for (i = 0; i < optsize; i++)
				printk("%02X", op[i]);
			printk(") ");
		}
#endif /* TEAMF1_CHANGES */
		break;
	}
	case IPPROTO_UDP:
	case IPPROTO_UDPLITE: {
		struct udphdr _udph, *uh;

#ifdef TEAMF1_CHANGES
		if (ih->protocol == IPPROTO_UDP)
			/* Max length: 10 "PROTO=UDP "     */
			len = snprintf (logBuff,15,"PROTO=UDP ");
		else	/* Max length: 14 "PROTO=UDPLITE " */
			len = snprintf (logBuff,15,"PROTO=UDPLITE ");
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		if (ih->protocol == IPPROTO_UDP)
			/* Max length: 10 "PROTO=UDP "     */
			printk("PROTO=UDP " );
		else	/* Max length: 14 "PROTO=UDPLITE " */
			printk("PROTO=UDPLITE ");
#endif /* TEAMF1_CHANGES */

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		uh = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_udph), &_udph);
		if (uh == NULL) {
#ifdef TEAMF1_CHANGES
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
#else /* TEAMF1_CHANGES */
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
#endif /* TEAMF1_CHANGES */
			break;
		}
		/* Max length: 20 "SPT=65535 DPT=65535 " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,50,"SPT=%u DPT=%u",
				ntohs(uh->source), ntohs(uh->dest));
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("SPT=%u DPT=%u LEN=%u ",
		       ntohs(uh->source), ntohs(uh->dest),
		       ntohs(uh->len));
#endif /* TEAMF1_CHANGES */
		break;
	}
	case IPPROTO_ICMP: {
		struct icmphdr _icmph, *ich;
#ifndef TEAMF1_CHANGES
		static const size_t required_len[NR_ICMP_TYPES+1]
			= { [ICMP_ECHOREPLY] = 4,
			    [ICMP_DEST_UNREACH]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_SOURCE_QUENCH]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_REDIRECT]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_ECHO] = 4,
			    [ICMP_TIME_EXCEEDED]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_PARAMETERPROB]
			    = 8 + sizeof(struct iphdr),
			    [ICMP_TIMESTAMP] = 20,
			    [ICMP_TIMESTAMPREPLY] = 20,
			    [ICMP_ADDRESS] = 12,
			    [ICMP_ADDRESSREPLY] = 12 };
#endif /* TEAMF1_CHANGES */
		/* Max length: 11 "PROTO=ICMP " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,15,"PROTO=ICMP ");
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("PROTO=ICMP ");
#endif /* TEAMF1_CHANGES */

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		ich = skb_header_pointer(skb, iphoff + ih->ihl * 4,
					 sizeof(_icmph), &_icmph);
		if (ich == NULL) {
#ifdef TEAMF1_CHANGES
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
#else /* TEAMF1_CHANGES */
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
#endif /* TEAMF1_CHANGES */
			break;
		}

		/* Max length: 18 "TYPE=255 CODE=255 " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,60,"TYPE=%u CODE=%u ",
				ich->type, ich->code);
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("TYPE=%u CODE=%u ", ich->type, ich->code);
#endif /* TEAMF1_CHANGES */

#ifndef TEAMF1_CHANGES
		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		if (ich->type <= NR_ICMP_TYPES
		    && required_len[ich->type]
		    && skb->len-iphoff-ih->ihl*4 < required_len[ich->type]) {
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			break;
		}

		switch (ich->type) {
		case ICMP_ECHOREPLY:
		case ICMP_ECHO:
			/* Max length: 19 "ID=65535 SEQ=65535 " */
			printk("ID=%u SEQ=%u ",
			       ntohs(ich->un.echo.id),
			       ntohs(ich->un.echo.sequence));
			break;

		case ICMP_PARAMETERPROB:
			/* Max length: 14 "PARAMETER=255 " */
			printk("PARAMETER=%u ",
			       ntohl(ich->un.gateway) >> 24);
			break;
		case ICMP_REDIRECT:
			/* Max length: 24 "GATEWAY=255.255.255.255 " */
			printk("GATEWAY=%u.%u.%u.%u ",
			       NIPQUAD(ich->un.gateway));
			/* Fall through */
		case ICMP_DEST_UNREACH:
		case ICMP_SOURCE_QUENCH:
		case ICMP_TIME_EXCEEDED:
			/* Max length: 3+maxlen */
			if (!iphoff) { /* Only recurse once. */
				printk("[");
				dump_packet(info, skb,
					    iphoff + ih->ihl*4+sizeof(_icmph));
				printk("] ");
			}

			/* Max length: 10 "MTU=65535 " */
			if (ich->type == ICMP_DEST_UNREACH
			    && ich->code == ICMP_FRAG_NEEDED)
				printk("MTU=%u ", ntohs(ich->un.frag.mtu));
		}
#endif /* TEAMF1_CHANGES */
		break;
	}
	/* Max Length */
	case IPPROTO_AH: {
		struct ip_auth_hdr _ahdr, *ah;

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 9 "PROTO=AH " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,10,"PROTO=AH ");
		logBuff = logBuff + len;
		totLen += len; 
#else /* TEAMF1_CHANGES */
		printk("PROTO=AH ");
#endif /* TEAMF1_CHANGES */

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		ah = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_ahdr), &_ahdr);
		if (ah == NULL) {
#ifdef TEAMF1_CHANGES
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
#else /* TEAMF1_CHANGES */
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
#endif /* TEAMF1_CHANGES */
			break;
		}

		/* Length: 15 "SPI=0xF1234567 " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,30,"SPI=0x%x ", ntohl(ah->spi));
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("SPI=0x%x ", ntohl(ah->spi));
#endif /* TEAMF1_CHANGES */
		break;
	}
	case IPPROTO_ESP: {
		struct ip_esp_hdr _esph, *eh;

		/* Max length: 10 "PROTO=ESP " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,15,"PROTO=ESP ");
		logBuff = logBuff + len;
		totLen += len; 
#else /* TEAMF1_CHANGES */
		printk("PROTO=ESP ");
#endif /* TEAMF1_CHANGES */

		if (ntohs(ih->frag_off) & IP_OFFSET)
			break;

		/* Max length: 25 "INCOMPLETE [65535 bytes] " */
		eh = skb_header_pointer(skb, iphoff+ih->ihl*4,
					sizeof(_esph), &_esph);
		if (eh == NULL) {
#ifdef TEAMF1_CHANGES
			len = snprintf (logBuff,40,"INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
			logBuff = logBuff + len;
			totLen += len;
#else /* TEAMF1_CHANGES */
			printk("INCOMPLETE [%u bytes] ",
			       skb->len - iphoff - ih->ihl*4);
#endif /* TEAMF1_CHANGES */
			break;
		}

		/* Length: 15 "SPI=0xF1234567 " */
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,30,"SPI=0x%x ", ntohl(eh->spi));
		logBuff = logBuff + len;
		totLen += len;
#else /* TEAMF1_CHANGES */
		printk("SPI=0x%x ", ntohl(eh->spi));
#endif /* TEAMF1_CHANGES */
		break;
	}
	/* Max length: 10 "PROTO 255 " */
	default:
#ifdef TEAMF1_CHANGES
		len = snprintf (logBuff,15,"PROTO=%u ", ih->protocol);
		logBuff = logBuff + len;
		totLen += len; 
#else /* TEAMF1_CHANGES */
		printk("PROTO=%u ", ih->protocol);
#endif /* TEAMF1_CHANGES */
	}

	/* Max length: 15 "UID=4294967295 " */
	if ((logflags & IPT_LOG_UID) && !iphoff && skb->sk) {
		read_lock_bh(&skb->sk->sk_callback_lock);
		if (skb->sk->sk_socket && skb->sk->sk_socket->file)
			printk("UID=%u ", skb->sk->sk_socket->file->f_uid);
		read_unlock_bh(&skb->sk->sk_callback_lock);
	}

#ifdef TEAMF1_CHANGES
	return totLen;
#endif /* TEAMF1_CHANGES */
	/* Proto    Max log string length */
	/* IP:      40+46+6+11+127 = 230 */
	/* TCP:     10+max(25,20+30+13+9+32+11+127) = 252 */
	/* UDP:     10+max(25,20) = 35 */
	/* UDPLITE: 14+max(25,20) = 39 */
	/* ICMP:    11+max(25, 18+25+max(19,14,24+3+n+10,3+n+10)) = 91+n */
	/* ESP:     10+max(25)+15 = 50 */
	/* AH:      9+max(25)+15 = 49 */
	/* unknown: 10 */

	/* (ICMP allows recursion one level deep) */
	/* maxlen =  IP + ICMP +  IP + max(TCP,UDP,ICMP,unknown) */
	/* maxlen = 230+   91  + 230 + 252 = 803 */
}

#ifndef TEAMF1_CHANGES
static struct nf_loginfo default_loginfo = {
	.type	= NF_LOG_TYPE_LOG,
	.u = {
		.log = {
			.level    = 0,
			.logflags = NF_LOG_MASK,
		},
	},
};
#endif /* TEAMF1_CHANGES */

#ifdef TEAMF1_CHANGES
#define IS_PKT_UNICAST(x,y)                         \
	((((inet_addr_type(x) == RTN_UNICAST)           \
	   || (inet_addr_type(x) == RTN_LOCAL)))        \
	  && (((inet_addr_type(y) == RTN_UNICAST))      \
	       || ((inet_addr_type(y) == RTN_LOCAL))))


void dump_packet_wrapper(const struct ipt_log_info *info,
			 const struct sk_buff *skb,
			 unsigned int iphoff)
{
	char * pTmp = NULL;    
	char logBuff[1024];
	int len = 0;
	struct nf_loginfo nfloginfo;

	nfloginfo.u.log.level = info->level;
	nfloginfo.u.log.logflags = info->logflags;
	pTmp = logBuff;
	len = snprintf (pTmp,40,"%s ",info->prefix);
	pTmp = pTmp + len;    
	dump_packet(&nfloginfo, skb, iphoff,pTmp);
	printk("%s\n",logBuff);
	return;
}
#endif /* TEAMF1_CHANGES */

#ifdef TEAMF1_CHANGES
static int
ipt_log_packet(unsigned int pf,
	       unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const struct ipt_log_info *loginfo,
	       const char *prefix,
	       char * logBuff)
#else /* TEAMF1_CHANGES */
static void
ipt_log_packet(unsigned int pf,
	       unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const struct nf_loginfo *loginfo,
	       const char *prefix)
#endif /* TEAMF1_CHANGES */
{
#ifdef TEAMF1_CHANGES
	struct iphdr _iph,*ih;
	int len = -1;
	int totLen = 0;
	struct nf_loginfo nfloginfo;

	spin_lock_bh(&log_lock);
	ih = skb_header_pointer(skb, 0, sizeof(_iph), &_iph);
	if (ih == NULL) {
		spin_unlock_bh(&log_lock);
		return 1;
	}

	/* Check whether the packet is for INPUT, OUTPUT or FORWARD chain */
	switch (hooknum) {
	case NF_IP_LOCAL_IN:
	case NF_IP_LOCAL_OUT:
	{
		/* Check the value of proc entry tf1_log_flag */
		switch ( tf1_log_flag ) {
		case IPT_LOG_UNICAST:
		{
			/* Log only UNICAST packets in INPUT and OUTPUT chain */

			/* 
			 * Checking whether the packets source or destination
			 * address is UNICAST or LOCAL. Here LOCAL means ip
			 * address of the box
			 */
			if (!(IS_PKT_UNICAST(ih->saddr,ih->daddr))) {
				spin_unlock_bh(&log_lock);
				return 1;
			}
            
			/*
			 * Checking local generated packets.
			 * Example:127.0.0.1<->127.0.0.1
			 */
			if (LOOPBACK(ih->saddr) || LOOPBACK(ih->daddr) ) {
				spin_unlock_bh(&log_lock);
				return 1;
			}
			break;
		}
		case IPT_LOG_BCAST_MCAST:
		{
			/* Checking whether the packet is UNICAST or LOCAL */
			if (IS_PKT_UNICAST(ih->saddr, ih->daddr)) {
				spin_unlock_bh(&log_lock);
				return 1;
			}
			break;
		}
		case IPT_LOG_UCAST_BCAST_MCAST:
		{
			/* 
			 * Checking local generated packets.
			 * Example:127.0.0.1<->127.0.0.1
			 */
			if (LOOPBACK(ih->saddr) || LOOPBACK(ih->daddr))	{
				spin_unlock_bh(&log_lock);
				return 1;
			}

			/* 
			 * Log UNICAST,MULTICAST and BROADCAST packets in INPUT
			 * and OUTPUT chains
			 */
			break;
		}
		case IPT_LOG_NO_PKT:
		default: 
		{
			if (strcmp (loginfo->prefix, "WAN2LAN[DROP]") == 0)
				break;

            /* LOG all packets that are been re-directed to the TPROXY */
            if (skb->mark & FW_TPROXY_MARK_MASK)
                break;
			/* 
			 * Nothing to be logged, this is the scenario when
			 * tf1_log_flag value is other than 0,1,2,3. 
			 */

			spin_unlock_bh(&log_lock);
			return 1;
		}
		} /* End of Inner switch*/ 
		break ;
	} /* End of outer case */
        
	default:
	{
		/* 
		 * All the packets should be logged except coming from
		 * INPUT and OUTPUT chains which have rule configured
		 * for them.
		 */
        
		/* Checking the broadcast packets */
		if ((inet_addr_type(ih->saddr) == RTN_BROADCAST)
		     || ((inet_addr_type(ih->daddr) == RTN_BROADCAST))) {
			spin_unlock_bh(&log_lock);
			return 1;
		}
	}
	} /* End of outer switch */
    
	len = printInterface((struct sk_buff *)skb, hooknum, in, 1, logBuff);
    if (len > 0) {
        logBuff = logBuff + len;
        totLen += len;
    }
    
	len = printInterface((struct sk_buff *)skb, hooknum, out, 0, logBuff);
    if (len > 0) {
        logBuff = logBuff + len;
        totLen += len;
    }
#else /* TEAMF1_CHANGES */
	if (!loginfo)
		loginfo = &default_loginfo;

	spin_lock_bh(&log_lock);
	printk("<%d>%sIN=%s OUT=%s ", loginfo->u.log.level,
	       prefix,
	       in ? in->name : "",
	       out ? out->name : "");
#endif /* TEAMF1_CHANGES */

#ifdef CONFIG_BRIDGE_NETFILTER
	if (skb->nf_bridge) {
		struct net_device *physindev = skb->nf_bridge->physindev;
		struct net_device *physoutdev = skb->nf_bridge->physoutdev;

		if (physindev && in != physindev)
			printk("PHYSIN=%s ", physindev->name);
		if (physoutdev && out != physoutdev)
			printk("PHYSOUT=%s ", physoutdev->name);
	}
#endif

#ifdef TEAMF1_CHANGES
	nfloginfo.u.log.level = loginfo->level;
	nfloginfo.u.log.logflags = loginfo->logflags;

	len = dump_packet(&nfloginfo, skb, 0,logBuff);
	logBuff = logBuff + len;
	totLen += len;

	spin_unlock_bh(&log_lock);
	return 0;
#else /* TEAMF1_CHANGES */
	if (in && !out) {
		/* MAC logging for input chain only. */
		printk("MAC=");
		if (skb->dev && skb->dev->hard_header_len
		    && skb->mac.raw != (void*)skb->nh.iph) {
			int i;
			unsigned char *p = skb->mac.raw;
			for (i = 0; i < skb->dev->hard_header_len; i++,p++)
				printk("%02x%c", *p,
				       i==skb->dev->hard_header_len - 1
				       ? ' ':':');
		} else
			printk(" ");
	}

	dump_packet(loginfo, skb, 0);
	printk("\n");
	spin_unlock_bh(&log_lock);
#endif /* TEAMF1_CHANGES */
}

static unsigned int
ipt_log_target(struct sk_buff **pskb,
	       const struct net_device *in,
	       const struct net_device *out,
	       unsigned int hooknum,
	       const struct xt_target *target,
	       const void *targinfo)
{
	const struct ipt_log_info *loginfo = targinfo;
	struct nf_loginfo li;
#ifdef TEAMF1_CHANGES
	char logBuff[1024];
	char * pTmp = logBuff;
	char * pFormat = NULL;
	int len = 0;
	int tokenFlag = 0;
	int formatFlag = 0;
	unsigned char *p = (*pskb)->mac.raw;
	int retVal = 0;

	pFormat = (char *) loginfo->format;
//	level_string[1] = '0' + (loginfo->level % 8);

	len = snprintf (pTmp,40,"%s ",loginfo->prefix);
	pTmp = pTmp + len;

	while((*pFormat) != '\0')
	{
		tokenFlag = 0;
		formatFlag = 0;
		if(*pFormat == '%') {   
			switch(*(++pFormat)) {
			case '.': 
				formatFlag = 1;     
				++pFormat;
				break;    
			}    

			switch(*pFormat) {
			case 's':
				tokenFlag = 1;   
				++pFormat;
				break;
			case 'd':
				tokenFlag = 2;
				++pFormat;
				break;    
			}

			switch(*(pFormat)) {
			case 'i':
				break;     
			case 'm':
				if(tokenFlag != 2) {
					len = snprintf (pTmp, 100, "SRC MAC = "
						"%02x:%02x:%02x:%02x:%02x:%02x ",
						*(p+6), *(p+7), *(p+8), *(p+9),
						*(p+10), *(p+11));
					pTmp = pTmp + len;
				}
				else {
					len = snprintf (pTmp, 100, "DST MAC = "
						"%02x:%02x:%02x:%02x:%02x:%02x ",
						*p, *(p+1), *(p+2), *(p+3),
						*(p+4), *(p+5));
					pTmp = pTmp + len;
				}
				break;     
			case 'p':
				break;     
			case 'u':
				break;     
			case 'I':
				break;     
			case 'T':
				break;     
			case 'U':
				break;     
			case 'C':
				break;     
			case 'S':
				break;     
			case 'L':
				len = snprintf (pTmp, 250, "ETHER[%02x:%02x:"
					"%02x:%02x:%02x:%02x -> %02x:%02x:%02x:"
					"%02x:%02x:%02x ]", *(p+6), *(p+7),
					*(p+8), *(p+9), *(p+10), *(p+11), *p,
					*(p+1), *(p+2), *(p+3), *(p+4), *(p+5));
				pTmp = pTmp + len;
				break;    
			case 'P':
				retVal = ipt_log_packet(PF_INET, hooknum, *pskb,
						in, out, loginfo, NULL,pTmp);
				break;
			default :
				if (tokenFlag != 0)
					len = snprintf (pTmp, 5, "%%%c%c",
							*(pFormat-1), *pFormat);
				else
					len = snprintf (pTmp, 5, "%%%c",
							*pFormat);
				pTmp = pTmp + len;
			}    
		}
		else {
			len = snprintf (pTmp,5,"%c",*(pFormat));
			pTmp = pTmp + len;
		}
		pFormat++;
	}
	if (strlen(loginfo->format) == 0) {
		if (loginfo->logflags & IPT_LOG_NFLOG) {
			li.type = NF_LOG_TYPE_LOG;
			li.u.log.level = loginfo->level;
			li.u.log.logflags = loginfo->logflags;
			nf_log_packet(PF_INET, hooknum, *pskb, in, out, &li,
			loginfo->prefix);
			retVal = 0;
		}
		else
			retVal = ipt_log_packet(PF_INET, hooknum, *pskb, in, out, loginfo,
					loginfo->prefix, pTmp);
	}
	if(retVal == 0)
		printk("%s\n",logBuff);
#else /* TEAMF1_CHANGES */
	li.type = NF_LOG_TYPE_LOG;
	li.u.log.level = loginfo->level;
	li.u.log.logflags = loginfo->logflags;

	ipt_log_packet(PF_INET, hooknum, *pskb, in, out, &li,
		       loginfo->prefix);
#endif /* TEAMF1_CHANGES */
	return XT_CONTINUE;
}

#ifdef TEAMF1_CHANGES
int ipt_log_target_ex(struct sk_buff **pskb,
	       const struct net_device *in,
	       const struct net_device *out,
	       unsigned int hooknum,
	       const void *targinfo,
	       const void *userinfo)
    {
    return  ipt_log_target(pskb,in,out,hooknum,targinfo,userinfo);
    }

void ipt_logfn(unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const char *prefix)
{
	struct ipt_log_info loginfo = { 
		.level = 4, /* WARNING */ 
		.logflags = IPT_LOG_MASK, 
		.prefix = "", 
		.format = "" 
	};
	if (prefix != NULL)
		{
		strncpy (loginfo.prefix, prefix, sizeof(loginfo.prefix));
		loginfo.prefix[sizeof(loginfo.prefix) - 1] = '\0';
		}
//	ipt_log_packet(PF_INET, hooknum, skb, in, out, loginfo,
//			loginfo->prefix, logBuff);
	/* 
	 * the prefix is sent as userinfo argument, because it includes the prefix
	 * and the url information 
	 */ 
	ipt_log_target((struct sk_buff **)&skb, in, out, hooknum, NULL, &loginfo);
}		

static void nf_log_packet_wrapper (unsigned int pf,
	       unsigned int hooknum,
	       const struct sk_buff *skb,
	       const struct net_device *in,
	       const struct net_device *out,
	       const struct nf_loginfo *nfloginfo,
	       const char *prefix)
{
	char logBuff [1024];
	struct ipt_log_info info;

	info.level = nfloginfo->u.log.level;
	info.logflags = nfloginfo->u.log.logflags;
	ipt_log_packet (pf, hooknum, skb, in, out, &info, prefix, logBuff);
	
}
#endif /* TEAMF1_CHANGES */

static int ipt_log_checkentry(const char *tablename,
			      const void *e,
			      const struct xt_target *target,
			      void *targinfo,
			      unsigned int hook_mask)
{
	const struct ipt_log_info *loginfo = targinfo;

	if (loginfo->level >= 8) {
		DEBUGP("LOG: level %u >= 8\n", loginfo->level);
		return 0;
	}
	if (loginfo->prefix[sizeof(loginfo->prefix)-1] != '\0') {
		DEBUGP("LOG: prefix term %i\n",
		       loginfo->prefix[sizeof(loginfo->prefix)-1]);
		return 0;
	}
#ifdef TEAMF1_CHANGES
	if (loginfo->format[sizeof(loginfo->format)-1] != '\0') {
		DEBUGP("LOG: format term %i\n",
		       loginfo->format[sizeof(loginfo->format)-1]);
		return 0;
	}
#endif /* TEAMF1_CHANGES */

	return 1;
}

static struct xt_target ipt_log_reg = {
	.name		= "LOG",
	.family		= AF_INET,
	.target		= ipt_log_target,
	.targetsize	= sizeof(struct ipt_log_info),
	.checkentry	= ipt_log_checkentry,
	.me		= THIS_MODULE,
};

static struct nf_logger ipt_log_logger ={
	.name		= "ipt_LOG",
#ifdef TEAMF1_CHANGES
	.logfn		= &nf_log_packet_wrapper,
#else /* TEAMF1_CHANGES */
	.logfn		= &ipt_log_packet,
#endif /* TEAMF1_CHANGES */
	.me		= THIS_MODULE,
};

static int __init ipt_log_init(void)
{
	int ret;

	ret = xt_register_target(&ipt_log_reg);
	if (ret < 0)
		return ret;
	if (nf_log_register(PF_INET, &ipt_log_logger) < 0) {
		printk(KERN_WARNING "ipt_LOG: not logging via system console "
		       "since somebody else already registered for PF_INET\n");
		/* we cannot make module load fail here, since otherwise
		 * iptables userspace would abort */
	}

	return 0;
}

static void __exit ipt_log_fini(void)
{
	nf_log_unregister(&ipt_log_logger);
	xt_unregister_target(&ipt_log_reg);
}

module_init(ipt_log_init);
module_exit(ipt_log_fini);
