/* Kernel module to match the port-ranges, trigger related port-ranges,
 * and alters the destination to a local IP address.
 *
 * Copyright (C) 2003, CyberTAN Corporation
 * All Rights Reserved.
 *
 * Description:
 *   This is kernel module for port-triggering.
 *
 *   The module follows the Netfilter framework, called extended packet 
 *   matching modules. 
 */

#include <linux/types.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netdevice.h>
#include <linux/if.h>
#include <linux/inetdevice.h>
#include <net/protocol.h>
#include <net/checksum.h>

#ifdef TEAMF1_CHANGES
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#endif /* TEAMF1_CHANGES */

#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_tuple.h>
#include <linux/netfilter_ipv4/ip_autofw.h>
#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ipt_TRIGGER.h>

/* This rwlock protects the main hash table, protocol/helper/expected
 *    registrations, conntrack timers*/

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)

#include <linux/netfilter_ipv4/listhelp.h>

#ifdef TEAMF1_CHANGES
static struct proc_dir_entry *proc_net_ipt_trigger = NULL;
static struct proc_dir_entry *proc_net_ipt_trigger_status = NULL;
int triggerNodeIdentifier=0;
#endif /* TEAMF1_CHANGES */

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

struct ipt_trigger {
	struct list_head list;		/* Trigger list */
	struct timer_list timeout;	/* Timer for list destroying */
	u_int32_t srcip;		/* Outgoing source address */
	u_int32_t dstip;		/* Outgoing destination address */
	u_int16_t mproto;		/* Trigger protocol */
	u_int16_t rproto;		/* Related protocol */
	struct ipt_trigger_ports ports;	/* Trigger and related ports */
	u_int8_t reply;			/* Confirm a reply connection */
#ifdef TEAMF1_CHANGES
	u_int8_t nodeId;		/* node identifier - VPP*/
#endif    
};

LIST_HEAD(trigger_list);
//DECLARE_LOCK(ip_trigger_lock);

static void trigger_refresh(struct ipt_trigger *trig, unsigned long extra_jiffies)
{
    DEBUGP("%s: \n", __FUNCTION__);
    IP_NF_ASSERT(trig);
    WRITE_LOCK(&ip_conntrack_lock);

    /* Need del_timer for race avoidance (may already be dying). */
    if (del_timer(&trig->timeout)) {
	trig->timeout.expires = jiffies + extra_jiffies;
	add_timer(&trig->timeout);
    }

    WRITE_UNLOCK(&ip_conntrack_lock);
}

static void __del_trigger(struct ipt_trigger *trig)
{
    DEBUGP("%s: \n", __FUNCTION__);
    IP_NF_ASSERT(trig);
    MUST_BE_WRITE_LOCKED(&ip_conntrack_lock);

     /* delete from 'trigger_list' */
    list_del(&trig->list);
    kfree(trig);
}

static void trigger_timeout(unsigned long ul_trig)
{
    struct ipt_trigger *trig= (void *) ul_trig;

    DEBUGP("trigger list %p timed out\n", trig);
    WRITE_LOCK(&ip_conntrack_lock);
    __del_trigger(trig);
    WRITE_UNLOCK(&ip_conntrack_lock);
}

static unsigned int
add_new_trigger(struct ipt_trigger *trig)
{
    struct ipt_trigger *new;

    DEBUGP("!!!!!!!!!!!! %s !!!!!!!!!!!\n", __FUNCTION__);
    WRITE_LOCK(&ip_conntrack_lock);
    new = (struct ipt_trigger *)
	kmalloc(sizeof(struct ipt_trigger), GFP_ATOMIC);

    if (!new) {
	WRITE_UNLOCK(&ip_conntrack_lock);
	DEBUGP("%s: OOM allocating trigger list\n", __FUNCTION__);
	return -ENOMEM;
    }

    memset(new, 0, sizeof(*trig));
    INIT_LIST_HEAD(&new->list);
    memcpy(new, trig, sizeof(*trig));

    /* add to global table of trigger */
    list_prepend(&trigger_list, &new->list);
    /* add and start timer if required */
    init_timer(&new->timeout);
    new->timeout.data = (unsigned long)new;
    new->timeout.function = trigger_timeout;
    new->timeout.expires = jiffies + (TRIGGER_TIMEOUT * HZ);
#ifdef TEAMF1_CHANGES    
    new->nodeId=triggerNodeIdentifier++;
#endif /* TEAMF1_CHANGES */    
    add_timer(&new->timeout);
	    
    WRITE_UNLOCK(&ip_conntrack_lock);

    return 0;
}

static inline int trigger_out_matched(const struct ipt_trigger *i,
	const u_int16_t proto, const u_int16_t dport)
{
    /* DEBUGP("%s: i=%p, proto= %d, dport=%d.\n", __FUNCTION__, i, proto, dport);
    DEBUGP("%s: Got one, mproto= %d, mport[0..1]=%d, %d.\n", __FUNCTION__, 
	    i->mproto, i->ports.mport[0], i->ports.mport[1]); */

    return ((i->mproto == proto) && (i->ports.mport[0] <= dport) 
	    && (i->ports.mport[1] >= dport));
}

static unsigned int
trigger_out(struct sk_buff **pskb,
		unsigned int hooknum,
		const struct net_device *in,
		const struct net_device *out,
		const void *targinfo)
{
    const struct ipt_trigger_info *info = targinfo;
    struct ipt_trigger trig, *found;
    const struct iphdr *iph = (*pskb)->nh.iph;
    struct tcphdr *tcph = (void *)iph + iph->ihl*4;	/* Might be TCP, UDP */

    DEBUGP("############# %s ############\n", __FUNCTION__);
    /* Check if the trigger range has already existed in 'trigger_list'. */
    found = LIST_FIND(&trigger_list, trigger_out_matched,
	    struct ipt_trigger *, iph->protocol, ntohs(tcph->dest));

    if (found) {
	/* Yeah, it exists. We need to update(delay) the destroying timer. */
	trigger_refresh(found, TRIGGER_TIMEOUT * HZ);
	/* In order to allow multiple hosts use the same port range, we update
	   the 'saddr' after previous trigger has a reply connection. */
	if (found->reply)
        {
	    found->srcip = ntohl(iph->saddr);
#ifdef TEAMF1_CHANGES
        found->dstip = ntohl(iph->daddr);
#endif /* TEAMF1_CHANGES */
        }
    }
    else {
	/* Create new trigger */
	memset(&trig, 0, sizeof(trig));
	trig.srcip = ntohl(iph->saddr);
#ifdef TEAMF1_CHANGES
    trig.dstip = ntohl(iph->daddr);
#endif /* TEAMF1_CHANGES */
	trig.mproto = iph->protocol;
	trig.rproto = info->proto;
	memcpy(&trig.ports, &info->ports, sizeof(struct ipt_trigger_ports));
	add_new_trigger(&trig);	/* Add the new 'trig' to list 'trigger_list'. */
    }

#ifdef TEAMF1_CHANGES    
    return NF_ACCEPT;	/* We don't block any packet. */
#else   /* TEAMF1_CHANGES */
    return IPT_CONTINUE;	/* We don't block any packet. */
#endif /* TEAMF1_CHANGES */    
}

static inline int trigger_in_matched(const struct ipt_trigger *i,
	const u_int16_t proto, const u_int16_t dport)
{
    /* DEBUGP("%s: i=%p, proto= %d, dport=%d.\n", __FUNCTION__, i, proto, dport);
    DEBUGP("%s: Got one, rproto= %d, rport[0..1]=%d, %d.\n", __FUNCTION__, 
	    i->rproto, i->ports.rport[0], i->ports.rport[1]); */
    u_int16_t rproto = i->rproto;

    if (!rproto)
	rproto = proto;

    return ((rproto == proto) && (i->ports.rport[0] <= dport) 
	    && (i->ports.rport[1] >= dport));
}

static unsigned int
trigger_in(struct sk_buff **pskb,
		unsigned int hooknum,
		const struct net_device *in,
		const struct net_device *out,
		const void *targinfo)
{
    struct ipt_trigger *found;
    const struct iphdr *iph = (*pskb)->nh.iph;
    struct tcphdr *tcph = (void *)iph + iph->ihl*4;	/* Might be TCP, UDP */

    /* Check if the trigger-ed range has already existed in 'trigger_list'. */
    found = LIST_FIND(&trigger_list, trigger_in_matched,
	    struct ipt_trigger *, iph->protocol, ntohs(tcph->dest));
    if (found) {
	DEBUGP("############# %s ############\n", __FUNCTION__);
	/* Yeah, it exists. We need to update(delay) the destroying timer. */
	trigger_refresh(found, TRIGGER_TIMEOUT * HZ);
	return NF_ACCEPT;	/* Accept it, or the imcoming packet could be 
				   dropped in the FORWARD chain */
    }
 
    return IPT_CONTINUE;	/* Our job is the interception. */
}

static unsigned int
trigger_dnat(struct sk_buff **pskb,
		unsigned int hooknum,
		const struct net_device *in,
		const struct net_device *out,
		const void *targinfo)
{
    struct ipt_trigger *found;
    const struct iphdr *iph = (*pskb)->nh.iph;
    struct tcphdr *tcph = (void *)iph + iph->ihl*4;	/* Might be TCP, UDP */
    struct ip_conntrack *ct;
    enum ip_conntrack_info ctinfo;
    struct ip_nat_multi_range_compat newrange;

    IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING);
    /* Check if the trigger-ed range has already existed in 'trigger_list'. */
    found = LIST_FIND(&trigger_list, trigger_in_matched,
	    struct ipt_trigger *, iph->protocol, ntohs(tcph->dest));

    if (!found || !(htonl(found->srcip)) 
#ifdef TEAMF1_CHANGES
        || found->dstip != htonl(iph->saddr)
#endif /* TEAMF1_CHANGES */
        )        
	return IPT_CONTINUE;	/* We don't block any packet. */

    DEBUGP("############# %s ############\n", __FUNCTION__);
    found->reply = 1;	/* Confirm there has been a reply connection. */
    ct = ip_conntrack_get(*pskb, &ctinfo);
    IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW));

    DEBUGP("%s: got ", __FUNCTION__);
    DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);

    /* Alter the destination of imcoming packet. */
    newrange = ((struct ip_nat_multi_range_compat)
	    { 1, { { IP_NAT_RANGE_MAP_IPS,
	             found->srcip, found->srcip,
	             { 0 }, { 0 }
	           } } });

#ifdef TEAMF1_CHANGES
    ct->mark = 0x2222;
#endif /* TEAMF1_CHANGES */

    /* Hand modified range to generic setup. */
    return ip_nat_setup_info(ct, &newrange.range[0], hooknum);
}

static unsigned int
trigger_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		unsigned int hooknum,
		const struct xt_target *target,
		const void *targinfo)
{
    const struct ipt_trigger_info *info = targinfo;
    const struct iphdr *iph = (*pskb)->nh.iph;

    /* DEBUGP("%s: type = %s\n", __FUNCTION__, 
	    (info->type == IPT_TRIGGER_DNAT) ? "dnat" :
	    (info->type == IPT_TRIGGER_IN) ? "in" : "out"); */

    /* The Port-trigger only supports TCP and UDP. */
    if ((iph->protocol != IPPROTO_TCP) && (iph->protocol != IPPROTO_UDP))
	return IPT_CONTINUE;

    if (info->type == IPT_TRIGGER_OUT)
	return trigger_out(pskb, hooknum, in, out, targinfo);
    else if (info->type == IPT_TRIGGER_IN)
	return trigger_in(pskb, hooknum, in, out, targinfo);
    else if (info->type == IPT_TRIGGER_DNAT)
    	return trigger_dnat(pskb, hooknum, in, out, targinfo);

    return IPT_CONTINUE;
}

static int
trigger_check(const char *tablename,
	      const void *e_void,
	      const struct xt_target *target,
	      void *targinfo,
	      unsigned int hook_mask)
{
	const struct ipt_trigger_info *info = targinfo;
	struct list_head *cur_item, *tmp_item;

	if ((strcmp(tablename, "mangle") == 0)) {
		DEBUGP("trigger_check: bad table `%s'.\n", tablename);
		return 0;
	}
	if (info->proto) {
	    if (info->proto != IPPROTO_TCP && info->proto != IPPROTO_UDP) {
		DEBUGP("trigger_check: bad proto %d.\n", info->proto);
		return 0;
	    }
	}
	if (info->type == IPT_TRIGGER_OUT) {
	    if (!info->ports.mport[0] || !info->ports.rport[0]) {
		DEBUGP("trigger_check: Try 'iptbles -j TRIGGER -h' for help.\n");
		return 0;
	    }
	}

	/* Empty the 'trigger_list' */
	list_for_each_safe(cur_item, tmp_item, &trigger_list) {
	    struct ipt_trigger *trig = (void *)cur_item;

	    DEBUGP("%s: list_for_each_safe(): %p.\n", __FUNCTION__, trig);
	    del_timer(&trig->timeout);
	    __del_trigger(trig);
	}

	return 1;
}

#ifdef TEAMF1_CHANGES
static int
proc_trigger_status_entry_show (struct seq_file *s, void *v)
    {
    struct list_head *cur_item, *tmp_item;
    struct ipt_trigger *trig=NULL;
    int index=0;
    
    list_for_each_safe(cur_item, tmp_item, &trigger_list)
        {
        trig = (void *)cur_item;
        seq_printf(s, "LanIP=%u.%u.%u.%u relatedPortStart=%x relatedPortEnd=%x "
                   "matchPortStart=%x matchPortEnd=%x TimeRemaining=%ld \n",
                   HIPQUAD(trig->srcip), trig->ports.rport[0],
                   trig->ports.rport[1], trig->ports.mport[0],
                   trig->ports.mport[1], (trig->timeout.expires - jiffies));            
        index++;
        }

    return 0;    
    }

static int
proc_trigger_status_open (struct inode *inode, struct file *file)
    {
    int  ret=0;
    ret  = single_open(file, proc_trigger_status_entry_show, NULL);
    if (!ret) seq_lseek(file, 0, 0);    
    return (ret);
    }

static int
proc_trigger_status_release (struct inode *inode, struct file *file)
    {
    return (single_release(inode, file));
    }

static struct file_operations proc_trigger_status_file_ops = {
  .owner = THIS_MODULE,
  .open = proc_trigger_status_open,
  .read = seq_read,
  .llseek = seq_lseek,
  .release = proc_trigger_status_release
};
#endif /* TEAMF1_CHANGES */


static struct ipt_target redirect_reg = { 
	.name 		= "TRIGGER",
	.family		= AF_INET,
	.target 	= trigger_target, 
	.targetsize	= sizeof(struct ipt_trigger_info),
	.hooks		= (1 << NF_IP_PRE_ROUTING) | (1 << NF_IP_FORWARD),
	.checkentry 	= trigger_check,
	.me 		= THIS_MODULE,
};

static int __init ipt_trigger_init(void)
{
#ifdef TEAMF1_CHANGES
    /* create /proc/net/ipt_trigger directory */
    proc_net_ipt_trigger = proc_mkdir("ipt_trigger", proc_net);
    if (!proc_net_ipt_trigger) {		
        printk(KERN_INFO "failed to create procfs entry .\n");
        return (-ENOMEM);
    }
    proc_net_ipt_trigger->owner = THIS_MODULE;

    /* create /proc/net/ipt_trigger/ipt_trigger_status */
    proc_net_ipt_trigger_status =
        create_proc_entry ("proc_net_ipt_trigger_status", 0644,
                           proc_net_ipt_trigger);

    if (proc_net_ipt_trigger_status) {
        proc_net_ipt_trigger_status->proc_fops = &proc_trigger_status_file_ops;
        proc_net_ipt_trigger_status->owner = THIS_MODULE;
    }
#endif /* TEAMF1_CHANGES */
	return xt_register_target(&redirect_reg);
}

static void __exit ipt_trigger_fini(void)
{
	xt_unregister_target(&redirect_reg);

#ifdef TEAMF1_CHANGES
	/* remove /proc/net/ipt_trigger/ directory */
	remove_proc_entry("ipt_trigger", proc_net);
	remove_proc_entry("ipt_trigger_status", proc_net_ipt_trigger);
#endif /* TEAMF1_CHANGES */    
}

module_init(ipt_trigger_init);
module_exit(ipt_trigger_fini);
