/*
 * iptables module to match IP6 address ranges
 *
 * Based on net/ipv4/netfilter/ipt_iprange.c
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ipv6.h>
#include <linux/types.h>
#include <net/checksum.h>
#include <net/ipv6.h>

#include <linux/netfilter_ipv6/ip6_tables.h>
#include <linux/netfilter_ipv6/ip6t_ip6range.h>

MODULE_DESCRIPTION("iptables arbitrary IP6 range match module");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif


/* function return 1 if a1 < a2, returns 0 if a1 = a2 or a1 > a2 */

static int __ipv6_addr_range_cmp(const struct in6_addr *a1, const struct in6_addr *a2) 
{
	unsigned int i;

	for (i = 0; i < 4; i++) {
        if ((ntohl(a1->s6_addr32[i]) == ntohl(a2->s6_addr32[i])))
            continue;

        if ((ntohl(a1->s6_addr32[i]) < ntohl(a2->s6_addr32[i])))
            return 1;
    }

    return 0;
}

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset, unsigned int protoff, int *hotdrop)
{
	const struct ip6t_ip6range_info *info = matchinfo;
    const struct ipv6hdr *ip6h = skb->nh.ipv6h;

	if (info->flags & IP6RANGE_SRC) {
        if ((__ipv6_addr_range_cmp(&ip6h->saddr, &info->src.min_addr) ||
            __ipv6_addr_range_cmp(&info->src.max_addr, &ip6h->saddr))
			 ^ !!(info->flags & IP6RANGE_SRC_INV)) {
			DEBUGP("src IP %04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x "
                   "NOT in range %s"
			       "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x-"
                   "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
				    NIP6(ip6h->saddr),
			        info->flags & IP6RANGE_SRC_INV ? "(INV) " : "",
				    NIP6(info->src.min_addr),
				    NIP6(info->src.max_addr));
			return 0;
		}
	}

	if (info->flags & IP6RANGE_DST) {
        if ((__ipv6_addr_range_cmp(&ip6h->daddr, &info->dst.min_addr) ||
            __ipv6_addr_range_cmp(&info->dst.max_addr, &ip6h->daddr))
			 ^ !!(info->flags & IP6RANGE_DST_INV)) {
			DEBUGP("dst IP %04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x "
                   "NOT in range %s"
			       "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x-"
                   "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n",
				    NIP6(ip6h->daddr),
			        info->flags & IP6RANGE_DST_INV ? "(INV) " : "",
				    NIP6(info->dst.min_addr),
				    NIP6(info->dst.max_addr));
			return 0;
		}
	}

	return 1;
}

static struct ip6t_match ip6range_match = 
{ 
	.list = { NULL, NULL }, 
	.name = "ip6range", 
	.family = AF_INET6,
	.match = &match, 
	.matchsize    = sizeof(struct ip6t_ip6range_info),
	.checkentry = NULL, 
	.destroy = NULL, 
	.me = THIS_MODULE
};

static int __init ip6t_ip6range_init(void)
{
	return xt_register_match(&ip6range_match);
}

static void __exit ip6t_ip6range_fini(void)
{
	xt_unregister_match(&ip6range_match);
}

module_init(ip6t_ip6range_init);
module_exit(ip6t_ip6range_fini);


