/*
 * This is a module which is used for setting the skb->priority field
 * of an skb for qdisc classification.
 */

/* (C) 2001-2002 Patrick McHardy <kaber@trash.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/checksum.h>

#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_CLASSIFY.h>

MODULE_AUTHOR("Patrick McHardy <kaber@trash.net>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("iptables qdisc classification target module");
MODULE_ALIAS("ipt_CLASSIFY");

/* urgent */
#define IPT_DSCP_EF      0xb8
#define IPT_DSCP_CS5     0xa0
#define IPT_DSCP_CS6     0xc0
#define IPT_DSCP_CS7     0xe0   

/* high */
#define IPT_DSCP_AF41    0x88
#define IPT_DSCP_AF42    0x90
#define IPT_DSCP_AF43    0x98
#define IPT_DSCP_CS4     0x80   

/* medium high */
#define IPT_DSCP_AF31    0x68
#define IPT_DSCP_AF32    0x70
#define IPT_DSCP_AF33    0x78
#define IPT_DSCP_CS3     0x60   

/* medium */
#define IPT_DSCP_AF21    0x48
#define IPT_DSCP_AF22    0x50
#define IPT_DSCP_AF23    0x58
#define IPT_DSCP_CS2     0x40  

#define IPT_DSCP_MASK	0xfc	/* 11111100 */
#define TC_PRIO_QUEUE_MASK  0xFFFF0000

#define IPT_URGENT_PRIO_QUEUE           0x1
#define IPT_HIGH_PRIO_QUEUE             0x2
#define IPT_MEDIUM_HIGH_PRIO_QUEUE      0x3
#define IPT_MEDIUM_PRIO_QUEUE           0x4
#define IPT_LOW_PRIO_QUEUE              0x5


static unsigned int
target(struct sk_buff **pskb,
       const struct net_device *in,
       const struct net_device *out,
       unsigned int hooknum,
       const struct xt_target *target,
       const void *targinfo)
{
	const struct xt_classify_target_info *clinfo = targinfo;
    u_int8_t sh_dscp = ((*pskb)->nh.iph->tos & IPT_DSCP_MASK);

    /* do default classification if clinfo->priority is 0 */
    if (!clinfo->priority) {

        /* return if packet has already been classified  */
        if ((*pskb)->priority != 0)
            return XT_CONTINUE;

        /* 
         * Default DSCP/IP Prec to Priority Queue Mapping
         */
        if ((sh_dscp == IPT_DSCP_EF) || (sh_dscp == IPT_DSCP_CS5)||
            (sh_dscp == IPT_DSCP_CS6)|| (sh_dscp == IPT_DSCP_CS7))
            (*pskb)->priority = TC_PRIO_QUEUE_MASK|IPT_URGENT_PRIO_QUEUE;

        else if ((sh_dscp == IPT_DSCP_AF41) || (sh_dscp == IPT_DSCP_AF42)||
                 (sh_dscp == IPT_DSCP_AF43) || (sh_dscp == IPT_DSCP_CS4))
            (*pskb)->priority = TC_PRIO_QUEUE_MASK|IPT_HIGH_PRIO_QUEUE;

        else if ((sh_dscp == IPT_DSCP_AF31) || (sh_dscp == IPT_DSCP_AF32)||
                 (sh_dscp == IPT_DSCP_AF33) || (sh_dscp == IPT_DSCP_CS3))
            (*pskb)->priority = TC_PRIO_QUEUE_MASK|IPT_MEDIUM_HIGH_PRIO_QUEUE;

        else if ((sh_dscp == IPT_DSCP_AF21) || (sh_dscp == IPT_DSCP_AF22) ||
                 (sh_dscp == IPT_DSCP_AF23) || (sh_dscp == IPT_DSCP_CS2))
            (*pskb)->priority = TC_PRIO_QUEUE_MASK|IPT_MEDIUM_PRIO_QUEUE;

        else
            (*pskb)->priority = TC_PRIO_QUEUE_MASK|IPT_LOW_PRIO_QUEUE;

        return XT_CONTINUE;
    }

	(*pskb)->priority = clinfo->priority;
	return XT_CONTINUE;
}

static struct xt_target xt_classify_target[] = {
	{
		.family		= AF_INET,
		.name 		= "CLASSIFY",
		.target 	= target,
		.targetsize	= sizeof(struct xt_classify_target_info),
		.table		= "mangle",
		.hooks		= (1 << NF_IP_LOCAL_OUT) |
				  (1 << NF_IP_FORWARD) |
				  (1 << NF_IP_POST_ROUTING),
		.me 		= THIS_MODULE,
	},
	{
		.name 		= "CLASSIFY",
		.family		= AF_INET6,
		.target 	= target,
		.targetsize	= sizeof(struct xt_classify_target_info),
		.table		= "mangle",
		.hooks		= (1 << NF_IP6_LOCAL_OUT) |
				  (1 << NF_IP6_FORWARD) |
				  (1 << NF_IP6_POST_ROUTING),
		.me 		= THIS_MODULE,
	},
};

static int __init xt_classify_init(void)
{
	return xt_register_targets(xt_classify_target,
				   ARRAY_SIZE(xt_classify_target));
}

static void __exit xt_classify_fini(void)
{
	xt_unregister_targets(xt_classify_target,
			      ARRAY_SIZE(xt_classify_target));
}

module_init(xt_classify_init);
module_exit(xt_classify_fini);
