/*
 * An x_tables match module to match interface groups
 *
 * (C) 2006,2007 Balazs Scheidler <[EMAIL PROTECTED]>,
 *   Laszlo Attila Toth <[EMAIL PROTECTED]>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/skbuff.h>

#include <linux/netfilter/xt_ifgroup.h>
#include <linux/netfilter/x_tables.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Laszlo Attila Toth <[EMAIL PROTECTED]>");
MODULE_DESCRIPTION("Xtables interface group matching module");
MODULE_ALIAS("ipt_ifgroup");
MODULE_ALIAS("ip6t_ifgroup");


static inline int
ifgroup_match_in(const struct net_device *in,
                const struct xt_ifgroup_info *info)
{
       return ((in->ifgroup & info->in_mask) == info->in_group) ^
               ((info->flags & XT_IFGROUP_INVERT_IN) == XT_IFGROUP_INVERT_IN);
}

static inline int
ifgroup_match_out(const struct net_device *out,
                const struct xt_ifgroup_info *info)
{
       return ((out->ifgroup & info->out_mask) == info->out_group) ^
               ((info->flags & XT_IFGROUP_INVERT_OUT) == XT_IFGROUP_INVERT_OUT);
}

static int
ifgroup_match(const struct sk_buff *skb,
            const struct net_device *in,
            const struct net_device *out,
            const struct xt_match *match,
            const void *matchinfo,
            int offset,
            unsigned int protoff,
            int *hotdrop)
{
       const struct xt_ifgroup_info *info = matchinfo;
       
       if (info->flags & XT_IFGROUP_MATCH_IN && !ifgroup_match_in(in, info))
               return false;
       if (info->flags & XT_IFGROUP_MATCH_OUT && !ifgroup_match_out(out, info))
               return false;
       
       return true;
}

static int ifgroup_checkentry(const char *tablename, const void *ip_void,
                              const struct xt_match *match,
                              void *matchinfo, unsigned int hook_mask)
{
       struct xt_ifgroup_info *info = matchinfo;

       if (!(info->flags & (XT_IFGROUP_MATCH_IN|XT_IFGROUP_MATCH_OUT))) {
               printk(KERN_ERR "xt_ifgroup: neither incoming nor "
                               "outgoing device selected\n");
               return false;
       }
       if (hook_mask & (1 << NF_IP_PRE_ROUTING | 1 << NF_IP_LOCAL_IN)
           && info->flags & XT_IFGROUP_MATCH_OUT) {
               printk(KERN_ERR "xt_ifgroup: output device not valid in "
                               "PRE_ROUTING and INPUT\n");
               return false;
       }
       if (hook_mask & (1 << NF_IP_POST_ROUTING | 1 << NF_IP_LOCAL_OUT)
           && info->flags & XT_IFGROUP_MATCH_IN) {
               printk(KERN_ERR "xt_ifgroup: input device not valid in "
                               "POST_ROUTING and OUTPUT\n");
               return false;
       }
       return true;
}

static struct xt_match xt_ifgroup_match[] __read_mostly  = {
       {
               .name           = "ifgroup",
               .match          = ifgroup_match,
               .checkentry     = ifgroup_checkentry,
               .matchsize      = sizeof(struct xt_ifgroup_info),
               .family         = AF_INET,
               .me             = THIS_MODULE,
       
       },
       {
               .name           = "ifgroup",
               .match          = ifgroup_match,
               .checkentry     = ifgroup_checkentry,
               .matchsize      = sizeof(struct xt_ifgroup_info),
               .family         = AF_INET6,
               .me             = THIS_MODULE,
       },
};

static int __init xt_ifgroup_init(void)
{
       return xt_register_matches(xt_ifgroup_match,
                       ARRAY_SIZE(xt_ifgroup_match));
}

static void __exit xt_ifgroup_fini(void)
{
       xt_unregister_matches(xt_ifgroup_match,
                       ARRAY_SIZE(xt_ifgroup_match));
}

module_init(xt_ifgroup_init);
module_exit(xt_ifgroup_fini);
