/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                        Copyright 2006 - 2010                             */
/*                    One Convergence Devices Pvt Ltd                       */
/*                         ALL RIGHTS RESERVED                              */
/*                                                                          */
/* Permission is hereby granted to licensees of One Convergence Devices Pvt */
/* Ltd' products to use or abstract this computer program for the sole      */
/* purpose of implementing a product based on One Convergence Devices Pvt   */
/* Ltd products. No other rights to reproduce, use, or disseminate this     */
/* computer program, whether in part or in whole, are granted.              */
/****************************************************************************/

/****************************************************************************
*  File: sip_ct_states.c
*
*  Description: File contains sipalg module functions for Linux Netfilter
*
*  Author: sriphaneesh.n@oneconvergence.com
*
*  Revision History:
*  01a,03Feb09,sg  Added support for SIP over VPN
*
*****************************************************************************/


#include <linux/in.h>
#include <linux/netfilter.h>
#include <linux/interrupt.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

//#define DEBUG
#include "sip_ct_helpers.h"


/* Module's global data */

extern spinlock_t sip_ct_lock;
extern u_int8_t   rmmod;
extern atomic_t sip_session_count;

/*---------------------------------------------------------------------------*/

/* Only for debugging; Order must be same as defined enums in header file */
#ifdef DEBUG
static char *sip_ct_states[] =
{
    "NONE",
    "REGISTERING",
    "INVITING",
    "REGISTERED",
    "INCALL",
    "CLOSING",
    "CLOSED",
    "HOLDING",
    "HELD"
};


/* Only for debugging; Order must be same as defined enums in header file */

static char *sip_ct_events[] =
{
    "IGNORE",
    "REGISTER",
    "UNREGISTER",
    "INVITE",
    "INCALL",
    "OK",
    "ACK",
    "HOLD",
    "CLOSE",
    "REDIRECT",
    "FAILURE",
    "REQUEST"
};
#endif


/* This function deletes call related info from the call list. While deleting
 * callinfo, it will remove the expects owned by this call.
 */

void sip_ct_delete_call(int callid, int unexpect, sip_ct_info_t *alginfo)
{
    int    index;

    sip_exp_info_t    *expinfo;
    sip_call_info_t   *pcinfo;


    DEBUGP("SIPALG: Deleting call %d\n", callid);

    pcinfo = &alginfo->calllist[callid-1];

    pcinfo->method    = 0;
    pcinfo->state     = SIP_CT_STATE_NONE;
    pcinfo->cidlen    = 0;
    pcinfo->itimeout  = 0;
    pcinfo->keepalive = 0;
    pcinfo->rtpactive = FALSE;
    pcinfo->lan2lan   = FALSE;

    index   = 0;
    expinfo = alginfo->explist;

    while (index < SIP_CT_MAX_EXPS)
    {
        CLEAR_BIT(callid, expinfo->usedby);

        /* We need not remove expects which is used by some other */

        if ((!expinfo->usedby) && (expinfo->exp1))
        {
            if (unexpect == FALSE)
            {
                write_lock_bh(&ip_conntrack_lock);
                expinfo->exp1->flags = 0;
//              expinfo->exp1->timeout.expires = jiffies + (10 * HZ);
                write_unlock_bh(&ip_conntrack_lock);
                ip_conntrack_expect_put(expinfo->exp1);
                expinfo->exp1 = NULL;
            }
            else
            {
                ip_conntrack_unexpect_related(expinfo->exp1);
                ip_conntrack_expect_put(expinfo->exp1);
                expinfo->exp1 = NULL;
            }
        }

        index   ++;
        expinfo ++;
    }
}


/* This function is a timer handler which is used to cleanup inactive, closed
 * calls and to release session info if there are no active calls. It also
 * updates timeout values for expects which are still used by active calls.
 */

void sip_ct_idletimer(unsigned long ct)
{
    int    index, active;

    sip_ct_info_t    *alginfo;
//  sip_exp_info_t    *expinfo;
    sip_call_info_t    *pcinfo;

    struct ip_conntrack    *master = (void *) ct;


    DEBUGP("SIPALG: Timer deleting inactive calls\n");

    spin_lock_bh(&sip_ct_lock);

    alginfo   = (sip_ct_info_t*) master->help.ct_alg_info;

    if (!alginfo)
    {
        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    /* If rtp media is going through this session, keep active call alive.
       All other calls keepalive time can be reduced */

    index  = 0;
    active = FALSE;
    pcinfo = alginfo->calllist;

    while (index < SIP_CT_MAX_CALLS)
    {
        if ((pcinfo->rtpactive) ||
            ((pcinfo->lan2lan) && (pcinfo->state == SIP_CT_STATE_INCALL)))
        {
            DEBUGP("SIPALG: Active call or Lan2Lan %d\n", index+1);
            pcinfo->keepalive = pcinfo->itimeout;
            pcinfo->rtpactive = FALSE;
            active = TRUE;
        }
        else if (pcinfo->state != SIP_CT_STATE_NONE)
        {
            pcinfo->keepalive -= SIP_CT_TIMER_TIMEOUT;

            if (pcinfo->keepalive <= 0)
            {
                DEBUGP("SIPALG: Deleting call %d\n", index+1);
                sip_ct_delete_call(index+1, FALSE, alginfo);
            }
            else
            {
                DEBUGP("SIPALG: Active call %d\n", index+1);
                active = TRUE;
            }
        }

        index  ++;
        pcinfo ++;
    }

    /* If there are no active calls, and if we are not wrong, we can safely
       delete this session */

    if ((active == FALSE) || (rmmod == TRUE))
    {
        kfree(alginfo);

        ip_conntrack_put(master);
        master->help.ct_alg_info = 0;

        DEBUGP("SIPALG: Session terminated\n");
        atomic_dec(&sip_session_count);

        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    /* If master timeout is proxy negotiated we should decrement it to
       indicate idletime after register */

    if (alginfo->initmethod == SIP_REGISTER)
        alginfo->keepalive  -= SIP_CT_TIMER_TIMEOUT;

    spin_unlock_bh(&sip_ct_lock);

    /* Idletimer has to run until there are no active calls, so restaring */

    init_timer(&alginfo->timeout);

    alginfo->timeout.data     = (unsigned long) ct;
    alginfo->timeout.function = sip_ct_idletimer;
    alginfo->timeout.expires  = jiffies + SIP_CT_TIMER_TIMEOUT * HZ;

    add_timer(&alginfo->timeout);

    return;
}


/* In this function I'm trying to infer events within a call based on the
 * parsed information from the traversing message. It is not complete sipfsm,
 * but I tried to extract events required for sipalg.
 */

static inline int sip_ct_get_callevent(sip_msg_info_t *msginfo)
{
    int    revent;


    DEBUGP("SIPALG: Extracting event\n");

    if (msginfo->msgtype == SIP_REQUEST)
    {
        switch (msginfo->method)
        {
        case SIP_INVITE:

            /* Event value is set to HOLD if it is holding invite */

            if (msginfo->revent == SIP_CT_EVENT_HOLD)
                revent = SIP_CT_EVENT_HOLD;
            else
                revent = SIP_CT_EVENT_INVITE;

        break;

        case SIP_BYE:
        case SIP_CANCEL:

            revent = SIP_CT_EVENT_CLOSE;

        break;

        case SIP_REGISTER:

            /* While parsing unregister we are setting expires to 1 */

            if (msginfo->expires == 1)
                revent = SIP_CT_EVENT_UNREGISTER;
            else
                revent = SIP_CT_EVENT_REGISTER;

        break;

        case SIP_ACK:

            revent = SIP_CT_EVENT_ACK;

        break;
        case SIP_OPTIONS:
        case SIP_NOTIFY:
        case SIP_SUBSCRIBE:

            revent = SIP_CT_EVENT_REQUEST;

        break;

        default:

            revent = SIP_CT_EVENT_IGNORE;
        }
    }
    else
    {
        /* I hope all 2XX messages are indication of success */

        if ((msginfo->status > 299) && (msginfo->status < 400))
            revent = SIP_CT_EVENT_REDIRECT;
        /* Fix for issues 10489 and 10426 based on RFC3261 sections 22.2 and
         * 22.3 : Ignoring 401 and 407 msgs. */
        else if ((msginfo->status == 407) || (msginfo->status == 401))
            revent = SIP_CT_EVENT_IGNORE;
        else if ((msginfo->status > 399) && (msginfo->status < 500))
            revent = SIP_CT_EVENT_FAILURE;
        else if ((msginfo->status > 199) && (msginfo->status < 300))
            revent = SIP_CT_EVENT_OK;
        else
            revent = SIP_CT_EVENT_IGNORE;
    }

    DEBUGP("SIPALG: Received event %s\n", sip_ct_events[revent]);
    return(revent);
}


/* This function upates the state and timeout of a call based on the messages
 * exchanged between communicating peers. This information is helpful to remove
 * call related data from the session and in turn it will be helpful to remove
 * entire session data (Ex. alginfo).
 */

void sip_ct_update_callstate(sip_msg_info_t *msginfo, sip_ct_info_t *alginfo)
{
    int    revent;

    sip_call_info_t    *pcinfo;


    DEBUGP("SIPALG: Updating call state\n");

    if (msginfo->msgtype == SIP_KEEPALIVE)
    {
        DEBUGP("SIPALG: Keep alive message\n");
        goto keepalive;
    }

    if (!msginfo->callid)
        return;

    /* This is to know the method initiating this session, it is useful for
       for session timeout */

    if (!alginfo->initmethod)
        alginfo->initmethod = msginfo->method;

    /* Try to extract event out of the message information */

    revent = sip_ct_get_callevent(msginfo);

    pcinfo = &alginfo->calllist[msginfo->callid-1];

    DEBUGP("SIPALG: Old state %s, timeout %d\n", sip_ct_states[pcinfo->state],
                                                            pcinfo->itimeout);

    switch (pcinfo->state)
    {
    case SIP_CT_STATE_NONE:

        if (revent == SIP_CT_EVENT_REGISTER)
            pcinfo->state = SIP_CT_STATE_REGISTERING;
        else if (revent == SIP_CT_EVENT_INVITE)
            pcinfo->state = SIP_CT_STATE_INVITING;
        else if (revent == SIP_CT_EVENT_REQUEST )
            pcinfo->state = SIP_CT_STATE_CLOSING;

        pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;

    break;

    case SIP_CT_STATE_REGISTERING:

        if (revent == SIP_CT_EVENT_OK)
        {
            pcinfo->state      = SIP_CT_STATE_REGISTERED;
            pcinfo->itimeout   = msginfo->expires + SIP_CT_DESTROY_DELAY;
            alginfo->keepalive = msginfo->expires + SIP_CT_DESTROY_DELAY;
        }

    break;

    case SIP_CT_STATE_INVITING:

        if (revent == SIP_CT_EVENT_ACK)
        {
            pcinfo->state    = SIP_CT_STATE_INCALL;
            pcinfo->itimeout = SIP_CT_INCALL_TIMEOUT;
        }
        else if (revent == SIP_CT_EVENT_HOLD)
        {
            pcinfo->state    = SIP_CT_STATE_HOLDING;
            pcinfo->itimeout = SIP_CT_INCALL_TIMEOUT;
        }
        else if ((revent == SIP_CT_EVENT_CLOSE) ||
                 (revent == SIP_CT_EVENT_FAILURE) ||
                 (revent == SIP_CT_EVENT_REDIRECT))
        {
            pcinfo->state    = SIP_CT_STATE_CLOSING;
            pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;
        }

    break;

    case SIP_CT_STATE_REGISTERED:

        if (revent == SIP_CT_EVENT_UNREGISTER)
        {
            pcinfo->state      = SIP_CT_STATE_CLOSING;
            pcinfo->itimeout   = msginfo->expires + SIP_CT_DESTROY_DELAY;
            alginfo->keepalive = msginfo->expires + SIP_CT_DESTROY_DELAY;
        }
        else if (revent == SIP_CT_EVENT_REGISTER)
        {
            pcinfo->state = SIP_CT_STATE_REGISTERING;
//          pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;
        }

    break;

    case SIP_CT_STATE_INCALL:

        if (revent == SIP_CT_EVENT_HOLD)
        {
            pcinfo->state    = SIP_CT_STATE_HOLDING;
            pcinfo->itimeout = SIP_CT_INCALL_TIMEOUT;
        }
        else if (revent == SIP_CT_EVENT_CLOSE)
        {
            pcinfo->state    = SIP_CT_STATE_CLOSING;
            pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;
        }

    break;

    case SIP_CT_STATE_CLOSING:

        if ((revent == SIP_CT_EVENT_OK) || (revent == SIP_CT_EVENT_ACK))
        {
            pcinfo->state    = SIP_CT_STATE_CLOSED;
            pcinfo->itimeout = SIP_CT_DESTROY_DELAY;
        }

    break;

    case SIP_CT_STATE_HOLDING:

        if (revent == SIP_CT_EVENT_ACK)
        {
            pcinfo->state    = SIP_CT_STATE_HELD;
            pcinfo->itimeout = SIP_CT_HELD_TIMEOUT;
        }

    break;

    case SIP_CT_STATE_HELD:

        if (revent == SIP_CT_EVENT_INVITE)
        {
            pcinfo->state    = SIP_CT_STATE_INVITING;
            pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;
        }
        else if (revent == SIP_CT_EVENT_CLOSE)
        {
            pcinfo->state    = SIP_CT_STATE_CLOSING;
            pcinfo->itimeout = SIP_CT_DEFAULT_TIMEOUT;
        }

    break;

    default:

        DEBUGP("SIPALG: Unknown state\n");
    }

    DEBUGP("SIPALG: New state %s, timeout %d\n", sip_ct_states[pcinfo->state],
                                                 pcinfo->itimeout);

    /* Some activity is going on this call, reinit keepalive value with call
       specific idletime to keep it alive */

    pcinfo->keepalive = pcinfo->itimeout;

    /* Refresh timeout values to keep connections alive. At max we can refresh
       master and this ct, but what about other related connections? */
keepalive:
    if (alginfo->proto != IPPROTO_TCP)
        ip_ct_refresh(msginfo->master, *msginfo->pskb, alginfo->keepalive*HZ);

    if ((msginfo->proto != IPPROTO_TCP) && (msginfo->master != msginfo->ct))
        ip_ct_refresh(msginfo->ct, *msginfo->pskb, SIP_CT_DEFAULT_TIMEOUT*HZ);

    return;
}


/* In this alg implementation call delete happens only in timer handler. This 
 * is requierd because I expect some message passing through alg after call
 * termination event. Our call list is limited and a new call may come, at
 * that time instead of waiting for timer to delete callinfo, this function
 * deletes callinfo of a closed call.
 */

int sip_ct_clean_calllist(int flags, sip_ct_info_t *alginfo)
{
    int    index;

    sip_call_info_t    *pcinfo;


    DEBUGP("SIPALG: Cleaning calllist\n");

    index   = 0;
    pcinfo  = alginfo->calllist;

    while (index < SIP_CT_MAX_CALLS)
    {
        if ((pcinfo->state == SIP_CT_STATE_CLOSED) ||
            (pcinfo->state == SIP_CT_STATE_CLOSING))
        {
            sip_ct_delete_call(index+1, flags, alginfo);
            return(index+1);
        }

        index  ++;
        pcinfo ++;
    }

    return(0);
}


/* In a proxy setup, a single alg session allows multiple calls through it.
 * This function performs few checks on message, like, whether this messages is
 * with in the existing calls or a new call, can we allow this call or not, and
 * if it is allowed records callid and etc in calllist for future validations.
 */

int sip_ct_validate_call(sip_msg_info_t *msginfo, sip_ct_info_t *alginfo)
{
    int    index, freeid;

    char    *pstr, *pcid;

    sip_call_info_t    *pclist;


    DEBUGP("SIPALG: Validating callid\n");

    /* Get the callid pointer from the message */

    if (sip_ct_parse_callid(msginfo, &pstr) == FAILURE)
        return(FAILURE);

    /* Search if this call already exists in our calllist or not */

    index  = 0;
    freeid = 0;
    pclist = alginfo->calllist;

    while (index < SIP_CT_MAX_CALLS)
    {
        if (pclist->method)
        {
            if (!strncmp(pstr, pclist->callid, pclist->cidlen))
            {
                DEBUGP("SIPALG: Message for existing calls\n");
                msginfo->callid = index+1;
                return(SUCCESS);
            }
        }
        else if (!freeid)
            freeid = index+1;

        index ++;
        pclist ++;
    }

 if(msginfo->ct->mark == FW_IPSEC_MARK ) /* IPSEC marked */
	DEBUGP("SIPALG: Found SIP packet with IPSEC Mark\n");
    else
    {
    /* A new call can be initiated only with the following requests */

    /* TF1: Adding support for SUBSCRIBE,NOTIFY & OPTIONS methods */

    if ((msginfo->msgtype != SIP_REQUEST) ||
        ((msginfo->method != SIP_INVITE) && (msginfo->method != SIP_REGISTER) 
         && (msginfo->method != SIP_NOTIFY) && 
             (msginfo->method != SIP_SUBSCRIBE)  && 
                 (msginfo->method != SIP_OPTIONS)))
    {
        DEBUGP("SIPALG: Cant accept new call\n");
        return(FAILURE);
    }
    }
    /* If this is a new call and we dont have room for another call, try to
       remove an already closed/inactive call. Even then no room? Deny it! */

    if ((!msginfo->callid) && (!freeid))
    {
        freeid = sip_ct_clean_calllist(TRUE, alginfo);

        if (!freeid)
        {
            DEBUGP("SIPALG: Call list is full\n");
            return(FAILURE);
        }
    }

    /* New call, record in our call list and allow */

    pclist = &alginfo->calllist[freeid-1];
    pcid   = pclist->callid;
    index  = 0;

    /* Copy callid into callinfo, truncate if it is bigger than space avail.
       It should not be a problem as we are not changing in message.
       Even after truncate, I assumed id will be unique within this session. */

    while ((*pstr != '\r') && (index < SIP_CT_CALLID_LEN))
    {
        pcid[index] = *pstr;
        index ++;
        pstr  ++;
    }

    if (*pstr == '\r')
    {
        DEBUGP("SIPALG: Callid is trimmed\n");
    }

    pcid[index]     = '\0';
    pclist->cidlen  = index;
    pclist->lan2lan = FALSE;
    pclist->method  = msginfo->method;
    msginfo->callid = freeid;

    DEBUGP("SIPALG: New call initiated %s\n", pcid);
    return(SUCCESS);
}
