/* Copyright (c) 2008, TeamF1, Inc. */

/*
modification history
--------------------
01b,23nov09,grj fixed lan side server issue
01a,24sep08,sat Added Hardware offload support
*/

/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                        Copyright 2006 - 2010                             */
/*                    One Convergence Devices Pvt Ltd                       */
/*                         ALL RIGHTS RESERVED                              */
/*                                                                          */
/* Permission is hereby granted to licensees of One Convergence Devices Pvt */
/* Ltd' products to use or abstract this computer program for the sole      */
/* purpose of implementing a product based on One Convergence Devices Pvt   */
/* Ltd products. No other rights to reproduce, use, or disseminate this     */
/* computer program, whether in part or in whole, are granted.              */
/****************************************************************************/

/****************************************************************************
*  File: sip_ct_expects.c
*
*  Description: File contains sipalg module functions for Linux Netfilter
*
*  Author: sriphaneesh.n@oneconvergence.com
*
*  Revision History:
*
*****************************************************************************/

#include <linux/in.h>
#include <linux/netfilter.h>
#include <linux/interrupt.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/version.h>

//#define DEBUG
#include "sip_ct_helpers.h"


/* Module's global data */

extern struct ip_conntrack_helper *psip_ct_tcp_helpers;
extern struct ip_conntrack_helper *psip_ct_udp_helpers;
extern struct ip_conntrack_helper sip_ct_rtp_helpers;
extern spinlock_t sip_ct_lock;

/*---------------------------------------------------------------------------*/

/* This function is going to be called by nefilter when singalling expectation
 * is initiating new connection.
 */

static void sip_ct_expectfn(struct ip_conntrack *new,
                            struct ip_conntrack_expect *exp1)
{

    DEBUGP("SIPALG: Singalling expectfn\n");

    /* It is a singalling connection, assign respective signalling helper */

    write_lock_bh(&ip_conntrack_lock);

    if (exp1->tuple.dst.protonum == IPPROTO_TCP)
        new->helper = psip_ct_tcp_helpers;
    else
        new->helper = psip_ct_udp_helpers;

    write_unlock_bh(&ip_conntrack_lock);

    /* If this expectation is not perminent, remove expect from our list.
       It is required because Netfilter anyway deletes this expect. Right now
       all sipalg expects are permanent, so it will never be true */

    if ((exp1->flags & IP_CT_EXPECT_PERMANENT) != IP_CT_EXPECT_PERMANENT)
    {
        sip_ct_info_t    *alginfo;
        sip_exp_info_t    *expinfo;

        spin_lock_bh(&sip_ct_lock);

        alginfo = (sip_ct_info_t*) exp1->master->help.ct_alg_info;

        if (alginfo == NULL)
        {
            spin_unlock_bh(&sip_ct_lock);
            return;
        }

        expinfo = sip_ct_get_expinfo(exp1, alginfo);

        if (expinfo == NULL)
        {
            spin_unlock_bh(&sip_ct_lock);
            return;
        }

        write_lock_bh(&ip_conntrack_lock);
        expinfo->exp1->flags = 0;
        write_unlock_bh(&ip_conntrack_lock);

        expinfo->exp1 = NULL;
        ip_conntrack_expect_put(exp1);

        spin_unlock_bh(&sip_ct_lock);
    }

    return;
}


/* This function is going to be called by nefilter when media expectation
 * is initiating new connection.
 */

static void sip_ct_rtp_expectfn(struct ip_conntrack *new,
                                struct ip_conntrack_expect *exp1)
{
    int    usedby, callid;

    sip_ct_info_t    *alginfo;
    sip_exp_info_t    *expinfo;


    DEBUGP("SIPALG: Expectfn for rtp media\n");

    /* It is a media connection, assign media helper for this connection */

    write_lock_bh(&ip_conntrack_lock);

    new->helper = &sip_ct_rtp_helpers;

    write_unlock_bh(&ip_conntrack_lock);

    /* Trying to associate this media connection with the call. Association
       is useful to update media activity in session's call list */

    spin_lock_bh(&sip_ct_lock);

    alginfo = (sip_ct_info_t*) exp1->master->help.ct_alg_info;

    if (alginfo == NULL)
    {
        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    expinfo = sip_ct_get_expinfo(exp1, alginfo);

    if (!expinfo)
    {
        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    /* If this expectation is not perminent, remove expect from our list.
       Right now all sipalg expects are permanent, so it will never be true */

    if ((exp1->flags & IP_CT_EXPECT_PERMANENT) != IP_CT_EXPECT_PERMANENT)
    {
        write_lock_bh(&ip_conntrack_lock);
        expinfo->exp1->flags = 0;
        write_unlock_bh(&ip_conntrack_lock);

        expinfo->exp1 = NULL;
        ip_conntrack_expect_put(exp1);
    }

    /* Find who is expecting this media in session's calllist */

    usedby = expinfo->usedby;

    callid = 1;

    while (callid <= SIP_CT_MAX_CALLS)
    {
        if (CHECK_BIT(callid, usedby))
            break;
        callid ++;
    }

    /* Assign callid in this ct to update media activity in that callinfo */

    if (callid <= SIP_CT_MAX_CALLS)
    {
        DEBUGP("SIPALG: Media for callid %d\n", callid);

        new->help.ct_alg_info = (unsigned long) callid;
    }
    else
    {
        DEBUGP("SIPALG: Don't know callid of media\n");
    }

    spin_unlock_bh(&sip_ct_lock);

    return;
}


/* This function searches for the given expectation tuple in alginfo expect
 * list. Returns pointer to the matching expinfo entry if found or null if not
 * found.
 */

sip_exp_info_t *sip_ct_lookup_explist(struct ip_conntrack_tuple *tuple1,
                                      sip_ct_info_t *alginfo)
{
    int    index;

    sip_exp_info_t    *expinfo;

    struct ip_conntrack_tuple    *tuple2;


    DEBUGP("SIPALG: Checking for matching exectation\n");

    index   = 0;
    expinfo = alginfo->explist;

    while (index < SIP_CT_MAX_EXPS)
    {
        if (expinfo->usedby)
        {
            tuple2 = &expinfo->origtuple;

            if (sip_ct_tuple_cmp(tuple1, tuple2))
            {
                DEBUGP("SIPALG: Expectation already exist\n");
                return(expinfo);
            }
        }

        index   ++;
        expinfo ++;
    }

    DEBUGP("SIPALG: Expectation is not available\n");
    return(NULL);
}


/* This function searches for the unused expinfo entry in expect list. Returns
 * pointer to the unused expinfo entry if available or null if not available.
 */

sip_exp_info_t *sip_ct_get_free_expinfo(sip_ct_info_t *alginfo)
{
    int    index;

    sip_exp_info_t    *expinfo;


    DEBUGP("SIPALG: Checking for unused expinfo\n");

    index   = 0;
    expinfo = alginfo->explist;

    while (index < SIP_CT_MAX_EXPS)
    {
        if (!expinfo->usedby)
            return expinfo;

        index   ++;
        expinfo ++;
    }

    DEBUGP("SIPALG: Expect list is full\n");
    return(NULL);
}


/* This function searches for the expinfo by the specified expect. Returns
 * pointer to the expinfo entry if found or null if not available.
 */

sip_exp_info_t *sip_ct_get_expinfo(struct ip_conntrack_expect *exp1,
                                   sip_ct_info_t *alginfo)
{
    int    index;

    sip_exp_info_t    *expinfo;


    DEBUGP("SIPALG: Searching expinfo by expect\n");

    index   = 0;
    expinfo = alginfo->explist;

    while (index < SIP_CT_MAX_EXPS)
    {
        if ((expinfo->usedby) && (expinfo->exp1 == exp1))
            return(expinfo);

        index   ++;
        expinfo ++;
    }

    DEBUGP("SIPALG: Could not find expinfo by exp\n");
    return(NULL);
}


/* This function create expects and update session's expect list. It will not
 * try to create expect if the same is already created within this session,
 * Updates usedby bit to allow more than one call within same session make
 * use of already created expects, and taked care of NAT related issues.
 */

sip_exp_info_t *sip_ct_create_expect(int exptype, int flags,
                                     struct ip_conntrack_tuple *tuple1,
                                     struct ip_conntrack_tuple *mask1,
                                     sip_msg_info_t *msginfo,
                                     sip_ct_info_t *alginfo)
{
    sip_exp_info_t    *expinfo;

    struct ip_conntrack_expect    *exp1;
    int ret = FAILURE;

    DEBUGP("SIPALG: Creating expect entry\n");

    /* Check if our expectation already exists in session's expect list */

    expinfo = sip_ct_lookup_explist(tuple1, alginfo);
    
    
    
    /* Update usedby flag if expect was already existing in this session */

    if (expinfo)
    {
        if (expinfo->exp1)
        {
            SET_BIT(msginfo->callid, expinfo->usedby);
            return(expinfo);
        }
        else /* This should not happen in this sipalg */
        {
            DEBUGP("SIPALG: BUG! Expect was deleted\n");
            return(NULL);
        }
    }

    /* Create new expect if it does not exist in the session explist */

    expinfo = sip_ct_get_free_expinfo(alginfo);

    if (expinfo == NULL)
        return(NULL);

    exp1 = ip_conntrack_expect_alloc(msginfo->master);

    if (exp1 == NULL)
    {
        DEBUGP("SIPALG: Expect alloc failed\n");
        return(NULL);
    }

    memcpy(&exp1->tuple, tuple1, sizeof(exp1->tuple));
    memcpy(&exp1->mask, mask1, sizeof(exp1->mask));

    exp1->flags    = 0;
    exp1->expectfn = 0;
    exp1->master   = msginfo->master;

    /* Setting PERMANENT to expect more than one connection using same expect.
       FOREVER makes an expect live for ever or until we clear this flag */

    if (flags == SIP_CT_EXP_PERM)
        exp1->flags |= (IP_CT_EXPECT_PERMANENT|IP_CT_EXPECT_FOREVER);

    /* Creating expect is bit different if nat is enabled on this session */

    if (!alginfo->natip)
    {
        if (exptype == SIP_CT_EXP_SIP)
            exp1->expectfn = sip_ct_expectfn;
        else
            exp1->expectfn = sip_ct_rtp_expectfn;
        
       
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
     if (( ret = ip_conntrack_expect_related(exp1) != SUCCESS) &&
            (!ip_conntrack_expect_find_get(&exp1->tuple)))                 
#else
      if (( ret = ip_conntrack_expect_related(exp1) != SUCCESS) &&
            (!ip_conntrack_expect_find(&exp1->tuple)))   
#endif        
        {
            ip_conntrack_expect_put(exp1);
            DEBUGP("SIPALG: Create expect failed\n");
            return(NULL);
        }

#ifdef CONFIG_TF1_HW_OFFLOAD
        if ((exp1->tuple.dst.protonum == IPPROTO_TCP))
            {
            DEBUGP("Dynamic tcp port registration %d %d\n",exptype,exp1->tuple.dst.u.tcp.port);
            hwOffloadRegister(exp1->tuple.dst.u.tcp.port,1);
            }
        if ((exp1->tuple.dst.protonum == IPPROTO_UDP))
            {
            DEBUGP("Dynamic udp port registration %d %d\n",exptype,exp1->tuple.dst.u.udp.port);
            hwOffloadRegister(exp1->tuple.dst.u.udp.port,1);
            }
#endif
    }
    else
    {
        if (ret = sip_nat_create_expect(exptype, exp1, expinfo,
                                  msginfo, alginfo) != SUCCESS)
        {
            ip_conntrack_expect_put(exp1);
            return(NULL);
        }
    }

    /* Since we are using exp pointer, increment reference count to ensure
       it will not be deleted by others */

    if (ret ==  SUCCESS)
        atomic_inc(&exp1->use);
    expinfo->exp1 = exp1;
    SET_BIT(msginfo->callid, expinfo->usedby);
    memcpy(&expinfo->origtuple, tuple1, sizeof(struct ip_conntrack_tuple));

    DEBUGP("SIPALG: New expect created\n");
    return(expinfo);
}
