/* Copyright (c) 2008, TeamF1, Inc. */

/*
modification history
--------------------
01a,24sep08,sat Added Hardware offload support
*/

/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                        Copyright 2006 - 2010                             */
/*                    One Convergence Devices Pvt Ltd                       */
/*                         ALL RIGHTS RESERVED                              */
/*                                                                          */
/* Permission is hereby granted to licensees of One Convergence Devices Pvt */
/* Ltd' products to use or abstract this computer program for the sole      */
/* purpose of implementing a product based on One Convergence Devices Pvt   */
/* Ltd products. No other rights to reproduce, use, or disseminate this     */
/* computer program, whether in part or in whole, are granted.              */
/****************************************************************************/

/****************************************************************************
*  File: sip_ct_helpers.c
*
*  Description: File contains sipalg module functions for Linux Netfilter
*
*  Author: sriphaneesh.n@oneconvergence.com
*
*  Revision History:
*
*****************************************************************************/
/* 
 * modification history
 * --------------------
 * 01b,12oct09,grj Added sip support to VLANs 
 * 01a,24sep08,vpp Added marking ct->mark to 0x00 in Load balancing mode.  
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/version.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ipt_DSCP.h>

//#define DEBUG
#include "sip_ct_helpers.h"
#include "sipalg_cfg.h"

//MODULE_LICENSE("One Convergence");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("<sriphaneesh.n@oneconvergence.com>");
MODULE_DESCRIPTION("Netfilter connection tracking helper module for sipalg");


/* Module's global data */

u_int8_t   rmmod;

DEFINE_SPINLOCK(sip_ct_lock);
atomic_t sip_session_count = ATOMIC_INIT(0);

struct ip_conntrack_helper *psip_ct_tcp_helpers;
struct ip_conntrack_helper *psip_ct_udp_helpers;
extern SIP_NODE_LIST * sipSubnetEntryFind (int ipaddr, int subnet);


/*---------------------------------------------------------------------------*/

/* Processing sip messages in a better way within sipalg require the direction
 * of the traversing messages. Inbound=Internet to LAN;Outbound=LAN to Internet.
 */

/* The following is one way to find the message direction. Here lan_netaddr
 * and lan_netmask are configurable parameters. But this method fails, when
 * router hosts multiple LANs/different subnets. Depending on the deployment
 * we may have to change the implementation.
 */

unsigned int    lan_netaddr = 0xc0a80100, lan_netmask = 0xffffff00;
//unsigned short int    sig_dscp = 0x00, rtp_dscp = 0x00;
/* For supporting TeamF1's Diffserve implementation we will mark sip packet's 
 * DSCP byte with 0x28 
 */
unsigned short int    sig_dscp = 0x28, rtp_dscp = 0x28;

module_param(lan_netaddr, uint, 0000);
MODULE_PARM_DESC(lan_netaddr, "LAN Network address in hex format");
module_param(lan_netmask, uint, 0000);
MODULE_PARM_DESC(lan_netmask, "LAN Netmask in hex format");
module_param(sig_dscp, ushort, 0000);
MODULE_PARM_DESC(sig_dscp, "DSCP value for signalling packets");
module_param(rtp_dscp, ushort, 0000);
MODULE_PARM_DESC(rtp_dscp, "DSCP value for RTP packets");


static inline int sip_ct_get_msgdir(struct ip_conntrack *ct,
                                    enum ip_conntrack_info ctinfo)
{
    int    dir;


    dir = CTINFO2DIR(ctinfo);
    
    if(sipSubnetEntryFind (ct->tuplehash[dir].tuple.src.ip,
                        htonl(lan_netmask)))
       {
       DEBUGP("outbound entry exists for %u.%u.%u.%u\n",
              NIPQUAD(ct->tuplehash[dir].tuple.src.ip));
        return(SIP_CT_OUTBOUND);
       }
    else 
        {
         return(SIP_CT_INBOUND);
        }
}


/* This is a tricky function to check if both communicating peers are within
 * our own lan. If both are lan clients, force them to establish media session
 * directly between them, instead of going through sipalg.
 */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
extern struct ip_conntrack_expect *
ip_conntrack_expect_find_get(const struct ip_conntrack_tuple *);
#else
extern struct ip_conntrack_expect *
ip_conntrack_expect_find(const struct ip_conntrack_tuple *);
#endif

struct ip_conntrack_expect *sip_ct_is_lan2lan(sip_hdr_info_t *hdrinfo,
                                              struct ip_conntrack_tuple *tuple1)
{
    DEBUGP("SIPALG: Checking for Lan2Lan call\n");

    if(sipSubnetEntryFind (hdrinfo->ipaddr2, htonl(lan_netmask)))
    {
        struct ip_conntrack_expect    *exp1;

        DEBUGP("SIPALG: Looks like Lan2Lan call\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
        exp1 = ip_conntrack_expect_find_get(tuple1);        
#else
        exp1 = ip_conntrack_expect_find(tuple1); 
#endif
       

        if (exp1)
        {
            DEBUGP("SIPALG: Confirmed Lan2Lan call\n");
        }

        return(exp1);
    }

    return(NULL);
}


/* This function sets specified dscp value in ip header and calculate header
 * check-sum.
 */

static inline void sip_ct_set_dscp(struct sk_buff **pskb, u_int8_t dscp)
{
    u_int8_t sh_dscp = ((dscp << IPT_DSCP_SHIFT) & IPT_DSCP_MASK);

  DEBUGP("pdscp %d, cdscp %d\n", sh_dscp, dscp);

    if (((*pskb)->nh.iph->tos & IPT_DSCP_MASK) != sh_dscp)
    {
        u_int16_t diffs[2];

        if (!skb_make_writable(pskb, sizeof(struct iphdr)))
            return;

        diffs[0] = htons((*pskb)->nh.iph->tos) ^ 0xFFFF;
        (*pskb)->nh.iph->tos = ((*pskb)->nh.iph->tos & ~IPT_DSCP_MASK)|sh_dscp;
        diffs[1] = htons((*pskb)->nh.iph->tos);
        (*pskb)->nh.iph->check = csum_fold(csum_partial((char *)diffs,
                                           sizeof(diffs),
                                           (*pskb)->nh.iph->check ^ 0xFFFF));
    }
}


/* This function allocates and initializes conntrack alg state info structure
 * for each session passing through sipalg. This will be called by the helper
 * function only at the time of session initialization.
 */

static sip_ct_info_t *sip_ct_init_alginfo(sip_msg_info_t *msginfo)
{
    sip_ct_info_t    *alginfo;

    struct ip_conntrack    *ct;


    DEBUGP("SIPALG: Initializing session info\n");

    /* Allocate memory and link alginfo to master ct */

    alginfo = kmalloc(sizeof(sip_ct_info_t), GFP_ATOMIC);

    if (alginfo)
    {
        memset((void*)alginfo, 0, sizeof(*alginfo));

        ct = msginfo->master;
        ct->help.ct_alg_info = (unsigned long) alginfo;

        /* In P2P case (without proxy), session can be initated from internet.
           So, I would like to keep the init direction for future use */

        alginfo->initdir = msginfo->msgdir;

        /* Get the nat ip address for this session, and the same ip is used
           for all connections in this session. Either nat is enabled or not
           keep lan client ip for future use */

        if (alginfo->initdir == SIP_CT_INBOUND)
        {
            if (ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip !=
                                  ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip)
                alginfo->natip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;

            alginfo->origip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;

            DEBUGP("SiPAlG: Natip %x, Origip %x\n",
                                ntohl(alginfo->natip), ntohl(alginfo->origip));
        }

        if (alginfo->initdir == SIP_CT_OUTBOUND)
        {
            if (ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip !=
                                  ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip)
                alginfo->natip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;

            alginfo->origip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;

            DEBUGP("SiPALg: Natip %x, Origip %x\n",
                                ntohl(alginfo->natip), ntohl(alginfo->origip));
        }

        /* Keep session initiator's transport protocol to create expects */

        alginfo->proto = msginfo->proto;

        /* Till there is a time negotiation, this is master's idle time */

        alginfo->keepalive = SIP_CT_DEFAULT_TIMEOUT;

        /* Register a timer handler to cleanup inactive calls, closed calls,
           to close session and to update expect timeouts */

        init_timer(&alginfo->timeout);

        alginfo->timeout.data     = (unsigned long) ct;
        alginfo->timeout.function = sip_ct_idletimer;
        alginfo->timeout.expires  = jiffies + SIP_CT_TIMER_TIMEOUT * HZ;

        add_timer(&alginfo->timeout);

        atomic_inc(&sip_session_count);

        /* Ensuring master not to be freed by netfilter before our timer */

        atomic_inc(&ct->ct_general.use);
    }

    return(alginfo);
}


/* Function to handle outbound messages (Lan-Internet). Here is the place we
 * process required headers in each message to create netfilter expects for
 * related connections. Handling nat related stuff like mangling,if applicable,
 * is also taken care here.
 */

static int sip_ct_handle_outbound(sip_msg_info_t *msginfo,
                                  sip_ct_info_t *alginfo)
{
    int    ret, index, hdrtype;

    char   *hdrptr;

    sip_hdr_info_t    hdrinfo;


    DEBUGP("SIPALG: Outbound message\n");

    index  = msginfo->hdrcount; ret = SUCCESS;

    while ((ret != FAILURE) && ((index--) >= 0))
    {
        hdrtype = msginfo->hdrlist[index].hdrtype;
        hdrptr  = msginfo->hdrlist[index].hdrptr;

        switch (hdrtype)
        {
        case SIP_VIA_HDR:

            /* Outbound request; We have to create an expect to recv inbound
               asymetric responses */

            if (msginfo->msgtype == SIP_REQUEST)
            {
                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_via_hdr(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Via -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                        ntohs(hdrinfo.port1),hdrinfo.proto1);

                if (ret != FAILURE)
                {
                    struct ip_conntrack_tuple tuple1, mask1;

                    memset((void*)&tuple1, 0, sizeof(tuple1));
                    memset((void*)&mask1, 0, sizeof(mask1));

                    if (hdrinfo.ipaddr1)
                        tuple1.dst.ip = hdrinfo.ipaddr1;
                    else if (hdrinfo.ipaddr2)
                        tuple1.dst.ip = hdrinfo.ipaddr2;
                    else
                        tuple1.dst.ip = alginfo->origip;

                    mask1.dst.ip = 0xffffffff;

                    tuple1.dst.protonum = hdrinfo.proto1;
                    mask1.dst.protonum  = 0xff;

                    if (tuple1.dst.protonum == IPPROTO_TCP)
                    {
                        tuple1.dst.u.tcp.port = hdrinfo.port1;
                        mask1.dst.u.tcp.port  = 0xffff;
                    }
                    else
                    {
                        tuple1.dst.u.udp.port = hdrinfo.port1;
                        mask1.dst.u.udp.port  = 0xffff;
                    }

                    if (!sip_ct_create_expect(SIP_CT_EXP_SIP, SIP_CT_EXP_PERM,
                                              &tuple1,&mask1,msginfo,alginfo))
                        ret = FAILURE;
                }

                /* Mangling via is required, if this is going through nat */

                if ((ret != FAILURE) && (alginfo->natip))
                    ret = sip_nat_mangle_hdr(hdrtype,&hdrinfo,msginfo,alginfo);
            }

        break;

        case SIP_CONTACT_HDR:

            if (msginfo->status == 302)
                break;

            /* Outbound request/response; We have to create an expect to recv
               inbound requests */

            memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

            ret = sip_ct_parse_contact_hdr(hdrptr, &hdrinfo);

            DEBUGP("SIPALG: Con -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                    ntohs(hdrinfo.port1),hdrinfo.proto1);

            /* Contact in register request may contain expires. We use that
               for session timeout */

            if (!msginfo->expires)
                msginfo->expires = hdrinfo.expires;

            if (ret != FAILURE)
            {
                struct ip_conntrack_tuple tuple1, mask1;

                memset((void*)&tuple1, 0, sizeof(tuple1));
                memset((void*)&mask1, 0, sizeof(mask1));

                if (hdrinfo.ipaddr1)
                    tuple1.dst.ip = hdrinfo.ipaddr1;
                else if (hdrinfo.ipaddr2)
                    tuple1.dst.ip = hdrinfo.ipaddr2;
                else
                    tuple1.dst.ip = alginfo->origip;

                mask1.dst.ip = 0xffffffff;

                /* If transport is not mentioned in this header, we are
                   forcing it to the session creators transport protocol */

                if (hdrinfo.proto1)
                    tuple1.dst.protonum = hdrinfo.proto1;
                else
                    tuple1.dst.protonum = alginfo->proto;

                mask1.dst.protonum = 0xff;

                if (tuple1.dst.protonum == IPPROTO_TCP)
                {
                    tuple1.dst.u.tcp.port = hdrinfo.port1;
                    mask1.dst.u.tcp.port  = 0xffff;
                }
                else
                {
                    tuple1.dst.u.udp.port = hdrinfo.port1;
                    mask1.dst.u.udp.port  = 0xffff;
                }

                if (!sip_ct_create_expect(SIP_CT_EXP_SIP, SIP_CT_EXP_PERM,
                                           &tuple1, &mask1, msginfo, alginfo))
                    ret = FAILURE;
            }

            /* Mangling contact is required, if this it is going through nat */

            if ((ret != FAILURE) && (alginfo->natip))
                ret = sip_nat_mangle_hdr(hdrtype, &hdrinfo, msginfo, alginfo);

        break;

        case SIP_SDP_CONTENT:

            memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

            ret = sip_ct_parse_sdp(msginfo, &hdrinfo);

            DEBUGP("SIPALG: Sdp ->  %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                    ntohs(hdrinfo.port1), hdrinfo.proto1);

            if ((ret != FAILURE) && (hdrinfo.ipaddr1))
            {
                struct ip_conntrack_tuple tuple1, mask1;

                sip_exp_info_t    *expinfo;

                u_int16_t    rtcpport;

                memset((void*)&tuple1, 0, sizeof(tuple1));
                memset((void*)&mask1, 0, sizeof(mask1));

                tuple1.dst.ip = hdrinfo.ipaddr1;
                mask1.dst.ip  = 0xffffffff;

                tuple1.dst.protonum = hdrinfo.proto1;
                mask1.dst.protonum  = 0xff;

                if (tuple1.dst.protonum == IPPROTO_TCP)
                {
                    tuple1.dst.u.tcp.port = hdrinfo.port1;
                    mask1.dst.u.tcp.port  = 0xffff;
                }
                else
                {
                    tuple1.dst.u.udp.port = hdrinfo.port1;
                    mask1.dst.u.udp.port  = 0xffff;
                }

                do
                {
                    expinfo = sip_ct_create_expect(SIP_CT_EXP_RTP,
                                                   SIP_CT_EXP_PERM, &tuple1,
                                                   &mask1, msginfo, alginfo);

                    if (!expinfo)
                    {
                        ret = FAILURE;
                        break;
                    }

                    rtcpport = ntohs(hdrinfo.port1) + 1;

                    if (tuple1.dst.protonum == IPPROTO_TCP)
                        tuple1.dst.u.tcp.port = htons(rtcpport);
                    else
                        tuple1.dst.u.udp.port = htons(rtcpport);

                    rtcpport = ntohs(expinfo->natport) + 1;
                    msginfo->rtcpnatport = htons(rtcpport);

                    if (!sip_ct_create_expect(SIP_CT_EXP_RTCP, 
                                              SIP_CT_EXP_PERM, &tuple1,
                                              &mask1, msginfo, alginfo))
                    {
                        ip_conntrack_unexpect_related(expinfo->exp1);
                        ip_conntrack_expect_put(expinfo->exp1);
                        expinfo->exp1   = NULL;
                        expinfo->usedby = 0;
                    }
                    else
                        break;

                } while (TRUE);
            }

            /* 'c' field will contain null ip in call hold message */

            if ((ret != FAILURE) && (!hdrinfo.ipaddr1))
                msginfo->revent = SIP_CT_EVENT_HOLD;

            /* Mangling sdp is required, if this session is going through nat */

            if ((ret != FAILURE) && (alginfo->natip))
                ret = sip_nat_mangle_sdp(&hdrinfo, msginfo, alginfo, NULL);

        break;

        case SIP_CONTENT_LEN:

            /* After mangling sdp fields, we may have to change content length
               in message header */

            if ((alginfo->natip) && (msginfo->sdplendiff))
                ret = sip_nat_mangle_content_len(msginfo, msginfo->sdplendiff);

        break;
        }
    }

    if ((ret!=FAILURE)&&(msginfo->proto==IPPROTO_TCP)&&(msginfo->msglendiff))
        ip_nat_mangle_tcp_sequence(msginfo->pskb, msginfo->ct, msginfo->ctinfo,
                                                           msginfo->msglendiff);

    return(ret);
}


/* Function to handle inbound messages (Internet-Lan). Here is the place we
 * process required headers in each message to create netfilter expects for
 * related connections. Handling nat related stuff like mangling,if applicable,
 * is also taken care here. Not everhthing is same here as outbound.
 */

static int sip_ct_handle_inbound(sip_msg_info_t *msginfo,
                                 sip_ct_info_t *alginfo)
{
    int    ret, index, hdrtype;

    char   *hdrptr;

    sip_hdr_info_t    hdrinfo;


    DEBUGP("SIPALG: Inbound message\n");

    index  = msginfo->hdrcount; ret = SUCCESS;

    while ((ret != FAILURE) && ((index--) >= 0))
    {
        hdrtype = msginfo->hdrlist[index].hdrtype;
        hdrptr  = msginfo->hdrlist[index].hdrptr;

        switch (hdrtype)
        {
        case SIP_REQUEST_STATUS:

            /* Inbound request line some times may contain lanip address.
               Requires mangling, if it is natted session */

            if ((alginfo->natip) && (msginfo->msgtype == SIP_REQUEST))
            {
                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_request_line(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Req -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                        ntohs(hdrinfo.port1),hdrinfo.proto1);

                if (ret != FAILURE)
                    ret = sip_nat_mangle_hdr(hdrtype,&hdrinfo,msginfo,alginfo);
            }

        break;

        case SIP_VIA_HDR:

            /* Inbound request; We have to create an expect to send outbound
               asymetric responses */
 
            if (msginfo->msgtype == SIP_REQUEST)
            {
                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_via_hdr(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Via -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                       ntohs(hdrinfo.port1), hdrinfo.proto1);

                if ((ret != FAILURE) && (hdrinfo.ipaddr1))
                {
                    struct ip_conntrack_tuple tuple1, mask1;

                    memset((void*)&tuple1, 0, sizeof(tuple1));
                    memset((void*)&mask1, 0, sizeof(mask1));

                    tuple1.dst.ip = hdrinfo.ipaddr1;
                    mask1.dst.ip  = 0xffffffff;

                    tuple1.src.ip = alginfo->origip;
                    mask1.src.ip  = 0xffffffff;

                    tuple1.dst.protonum = hdrinfo.proto1;
                    mask1.dst.protonum  = 0xff;

                    if (tuple1.dst.protonum == IPPROTO_TCP)
                    {
                        tuple1.dst.u.tcp.port = hdrinfo.port1;
                        mask1.dst.u.tcp.port  = 0xffff;
                    }
                    else
                    {
                        tuple1.dst.u.udp.port = hdrinfo.port1;
                        mask1.dst.u.udp.port  = 0xffff;
                    }

                    sip_ct_create_expect(SIP_CT_EXP_SIP, SIP_CT_EXP_PERM,
                                         &tuple1, &mask1, msginfo, alginfo);
                }
            }
            else if (alginfo->natip)
            {
                /* Via header in responses contain lan addresses, mangling is
                   required if this is natted session */

                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_via_hdr(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Via -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                       ntohs(hdrinfo.port1), hdrinfo.proto1);

                if (ret != FAILURE)
                    ret = sip_nat_mangle_hdr(hdrtype,&hdrinfo,msginfo,alginfo);
            }

        break;

        case SIP_CONTACT_HDR:

            /* Message could be an inabound request/response; Create an expect
               entry to send out requests */

            /* Outbound requests follow record route, if record route is
               presend in the msg. No need to create expect */

            if ((msginfo->method != SIP_REGISTER) && (!msginfo->rrflag))
            {
                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_contact_hdr(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Con -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                       ntohs(hdrinfo.port1), hdrinfo.proto1);

                if (ret != FAILURE)
                {
                    struct ip_conntrack_tuple tuple1, mask1;

                    memset((void*)&tuple1, 0, sizeof(tuple1));
                    memset((void*)&mask1, 0, sizeof(mask1));

                    if (hdrinfo.ipaddr1)
                        tuple1.dst.ip = hdrinfo.ipaddr1;
                    else if (hdrinfo.ipaddr2)
                        tuple1.dst.ip = hdrinfo.ipaddr2;
                    else
                        break;

                    mask1.dst.ip  = 0xffffffff;

                    tuple1.src.ip = alginfo->origip;
                    mask1.src.ip  = 0xffffffff;

                    /* If transport is not mentioned in this header, we are
                       forcing it to the protocol that created session */

                    if (hdrinfo.proto1)
                        tuple1.dst.protonum = hdrinfo.proto1;
                    else
                        tuple1.dst.protonum = alginfo->proto;

                    mask1.dst.protonum = 0xff;

                    if (tuple1.dst.protonum == IPPROTO_TCP)
                    {
                        tuple1.dst.u.tcp.port = hdrinfo.port1;
                        mask1.dst.u.tcp.port  = 0xffff;
                    }
                    else
                    {
                        tuple1.dst.u.udp.port = hdrinfo.port1;
                        mask1.dst.u.udp.port  = 0xffff;
                    }

                    sip_ct_create_expect(SIP_CT_EXP_SIP, SIP_CT_EXP_PERM,
                                             &tuple1,&mask1,msginfo,alginfo);
                }
            }

            /* Responses to register message contain lan client contact.
               Mangling is required, if this is natted session */

            if ((msginfo->method == SIP_REGISTER) && (alginfo->natip))
            {
                memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

                ret = sip_ct_parse_contact_hdr(hdrptr, &hdrinfo);

                DEBUGP("SIPALG: Con -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                       ntohs(hdrinfo.port1), hdrinfo.proto1);

                /* Contact in register msg may come with expires. It is used
                   for session timeout */

                if (!msginfo->expires)
                    msginfo->expires = hdrinfo.expires;

                if (ret != FAILURE)
                    ret = sip_nat_mangle_hdr(hdrtype,&hdrinfo,msginfo,alginfo);
            }

        break;

        case SIP_RROUTE_HDR:

            /* Message could be an inbound request/response; Create an expect
               entry to send requests/responses out */

            memset((void*)&hdrinfo, 0, sizeof(hdrinfo));

            ret = sip_ct_parse_rroute_hdr(hdrptr, &hdrinfo);

            DEBUGP("SIPALG: Rrt -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                    ntohs(hdrinfo.port1),hdrinfo.proto1);

            if ((ret != FAILURE) && (hdrinfo.ipaddr1))
            {
                struct ip_conntrack_tuple tuple1, mask1;

                memset((void*)&tuple1, 0, sizeof(tuple1));
                memset((void*)&mask1, 0, sizeof(mask1));

                tuple1.dst.ip = hdrinfo.ipaddr1;
                mask1.dst.ip  = 0xffffffff;

                tuple1.src.ip = alginfo->origip;
                mask1.src.ip  = 0xffffffff;

                /* If transport is not mentioned in this header, we are forcing
                   it to the protocol that created session */

                if (hdrinfo.proto1)
                    tuple1.dst.protonum = hdrinfo.proto1;
                else
                    tuple1.dst.protonum = alginfo->proto;

                mask1.dst.protonum = 0xff;

                if (tuple1.dst.protonum == IPPROTO_TCP)
                {
                    tuple1.dst.u.tcp.port = hdrinfo.port1;
                    mask1.dst.u.tcp.port  = 0xffff;
                }
                else
                {
                    tuple1.dst.u.udp.port = hdrinfo.port1;
                    mask1.dst.u.udp.port  = 0xffff;
                }

                sip_ct_create_expect(SIP_CT_EXP_SIP, SIP_CT_EXP_PERM,
                                     &tuple1, &mask1, msginfo, alginfo);
            }

        break;

        case SIP_SDP_CONTENT:

            /* Sdp content contain rtp session details, we have to create an
               expect entry to send media out */

            memset(&hdrinfo, 0, sizeof(hdrinfo));

            ret = sip_ct_parse_sdp(msginfo, &hdrinfo);

            DEBUGP("SIPALG: Sdp -> %x, %d, %d\n", ntohl(hdrinfo.ipaddr1),
                                   ntohs(hdrinfo.port1), hdrinfo.proto1);

            /* Mangling sdp is required, if this session is going through nat */

            if ((ret != FAILURE) && (hdrinfo.ipaddr1))
            {
                struct ip_conntrack_tuple    tuple1, mask1;
                struct ip_conntrack_expect    *exp1;
                sip_exp_info_t    *expinfo;
                u_int16_t    rtcpport;

                memset(&tuple1, 0, sizeof(tuple1));
                memset((void*)&mask1, 0, sizeof(mask1));

                tuple1.dst.ip         = hdrinfo.ipaddr1;
                mask1.dst.ip          = 0xffffffff;

                tuple1.dst.u.udp.port = hdrinfo.port1;
                mask1.dst.u.udp.port  = 0xffff;

                tuple1.dst.protonum   = hdrinfo.proto1;
                mask1.dst.protonum    = 0xff;

                if ((exp1 = sip_ct_is_lan2lan(&hdrinfo, &tuple1)))
                {
                    alginfo->calllist[msginfo->callid-1].lan2lan = TRUE;

                    if (hdrinfo.ipaddr1 != hdrinfo.ipaddr2)
                        ret = sip_nat_mangle_sdp(&hdrinfo,msginfo,alginfo,exp1);

                    ip_conntrack_expect_put(exp1);

                    break;
                }

                tuple1.src.ip = alginfo->origip;
                mask1.src.ip  = 0xffffffff;

                do
                {
                    expinfo = sip_ct_create_expect(SIP_CT_EXP_RTP, 
                                                   SIP_CT_EXP_PERM, &tuple1,
                                                   &mask1, msginfo, alginfo);

                    if (!expinfo)
                    {
//                      ret = FAILURE;
                        break;
                    }

                    rtcpport = ntohs(hdrinfo.port1) + 1;

                    if (tuple1.dst.protonum == IPPROTO_TCP)
                        tuple1.dst.u.tcp.port = htons(rtcpport);
                    else
                        tuple1.dst.u.udp.port = htons(rtcpport);

                    rtcpport = ntohs(expinfo->natport) + 1;
                    msginfo->rtcpnatport = htons(rtcpport);

                    if (!sip_ct_create_expect(SIP_CT_EXP_RTCP, 
                                              SIP_CT_EXP_PERM, &tuple1,
                                              &mask1, msginfo, alginfo))
                    {
                        ip_conntrack_unexpect_related(expinfo->exp1);
                        ip_conntrack_expect_put(expinfo->exp1);
                        expinfo->exp1   = NULL;
                        expinfo->usedby = 0;
                    }
                    else
                        break;

                } while (TRUE);
            }

            /* 'c' field will contain null ip in call hold message */

            if ((ret != FAILURE) && (!hdrinfo.ipaddr1))
                msginfo->revent = SIP_CT_EVENT_HOLD;

        break;

        case SIP_CONTENT_LEN:

            /* After mangling sdp fields, we may have to change content length
               in message header */

            if (msginfo->sdplendiff)
                ret = sip_nat_mangle_content_len(msginfo, msginfo->sdplendiff);

        break;
        }
    }

    if ((ret!=FAILURE)&&(msginfo->proto==IPPROTO_TCP)&&(msginfo->msglendiff))
        ip_nat_mangle_tcp_sequence(msginfo->pskb, msginfo->ct, msginfo->ctinfo, 
                                                           msginfo->msglendiff);

    return(ret);
}


/* This function is main function for sipalg. It processes messages independent
 * of the transport protocol. After basic checks on packet TCP and UDP helpers
 * call this functions. 
 */

static int sip_ct_handle_msg(sip_msg_info_t *msginfo)
{
    int    ret;

    sip_ct_info_t     *alginfo;

    struct ip_conntrack    *ct;


    DEBUGP("SIPALG: Processing sip message\n");

    /* First of all, knowing directon of the message is very important */

    ct = msginfo->ct;

    msginfo->msgdir = sip_ct_get_msgdir(ct, msginfo->ctinfo);

    /* State info for each session is stored in session's master ct alginfo.
       All other related connections use the state info avilable in master */

    /* Check whether there is a master. if none, this ct is the master */

    spin_lock_bh(&sip_ct_lock);

    if (!ct->master)
    {
        msginfo->master = ct;

        /* Session state initialization is done only by the first msg */

        if (ct->help.ct_alg_info == 0)
            alginfo = sip_ct_init_alginfo(msginfo);
        else
            alginfo = (sip_ct_info_t*) ct->help.ct_alg_info;
    }
    else
    {
        msginfo->master = ct->master;
        alginfo         = (sip_ct_info_t*) ct->master->help.ct_alg_info;
    }

    /* We cant go ahead without session alginfo */

    if (!alginfo)
    {
        spin_unlock_bh(&sip_ct_lock);
        DEBUGP("SIPALG: Alginfo is not available\n");
        return(NF_DROP);
    }

    /* Everything looks fine, take-off and the story starts */

    ret = sip_ct_parse_msg(msginfo->msgptr, msginfo->msglen, msginfo);

    if ((ret != FAILURE) && (msginfo->msgtype != SIP_KEEPALIVE))
    {
        ret = sip_ct_validate_call(msginfo, alginfo);

        if (ret != FAILURE)
        {
 
            if (msginfo->msgdir == SIP_CT_OUTBOUND)
                ret = sip_ct_handle_outbound(msginfo, alginfo);
            else
                ret = sip_ct_handle_inbound(msginfo, alginfo);
        }
    }

    if (ret != FAILURE)
        sip_ct_update_callstate(msginfo, alginfo);

    spin_unlock_bh(&sip_ct_lock);

    if (ret != FAILURE)
{
     DEBUGP("msgdir = %d, sig_dscp = %d \n",msginfo->msgdir,sig_dscp);
        if ((sig_dscp) && (msginfo->msgdir == SIP_CT_OUTBOUND)){
      DEBUGP("Going to set DSCP\n");
            sip_ct_set_dscp(msginfo->pskb, sig_dscp);
        }
      DEBUGP("Not setting DSCP\n");
        return(NF_ACCEPT);
 }

    DEBUGP("SIPALG: **** DROP ****\n");
    return(NF_DROP);
}


/* This function is the entry point to the sipalg for tcp signalling and is
 * called by the netfilter when packet matches with the defined helper tuple.
 */

int sip_ct_tcp_helper(struct sk_buff **pskb,
                      struct ip_conntrack *ct,
                      enum ip_conntrack_info ctinfo)
{
    int    iphlen, tcphlen, ret;

    sip_msg_info_t    *msginfo;


    DEBUGP("\nSIPALG: TCP Helper\n");

    /* Until tcp handshake is complete, we cant do any processing */

    if ((ctinfo != IP_CT_ESTABLISHED) && 
        (ctinfo != IP_CT_ESTABLISHED+IP_CT_IS_REPLY))
        return(NF_ACCEPT);

    /* Check whether the complete packet is in a single skb or spread across
       multiple buffers; We have to pull every thing into a single skb */
    //TODO

    /* Allocate buffer to keep message info while processing each msg */
    /* kmalloc is bit slow, instead we can think of using memory pool */
    //TODO

    msginfo = kmalloc(sizeof(sip_msg_info_t), GFP_ATOMIC);

    if (!msginfo)
    {
        DEBUGP("SIPALG: Memeory allocation failed\n");
        return(NF_DROP);
    }

    memset(msginfo, 0, sizeof(*msginfo));

    msginfo->ct     = ct;
    msginfo->ctinfo = ctinfo;
    msginfo->pskb   = pskb;
    msginfo->iph    = (*pskb)->nh.iph;
    iphlen          = msginfo->iph->ihl * 4;
    msginfo->tcph   = (struct tcphdr *)((char *)msginfo->iph + iphlen);
    tcphlen         = msginfo->tcph->doff * 4;
    msginfo->msgptr = (char *) ((char*) msginfo->tcph + tcphlen);
    msginfo->msglen = ntohs(msginfo->iph->tot_len) - (iphlen + tcphlen);
    msginfo->proto  = IPPROTO_TCP;

    /* We cant process if there is no message (Ex. ack, fin, rst, keepalive) */
    //TODO: Is there any other way to handle incomplete/junk/keepalive msgs

//  if (msginfo->msglen < 5)
    if (!msginfo->msglen)
    {
        kfree(msginfo);
        return(NF_ACCEPT);
    }

    /* Process sip messages */

    ret = sip_ct_handle_msg(msginfo);

    kfree(msginfo);

    DEBUGP("SIPALG: Leaving tcp helper module\n");

    return(ret);
}


/* This function is the entry point to the sipalg for udp signalling and is
 * called by the netfilter when packet matches with the defined helper tuple.
 */

int sip_ct_udp_helper(struct sk_buff **pskb,
                      struct ip_conntrack *ct,
                      enum ip_conntrack_info ctinfo)
{
    int    ret;

    sip_msg_info_t    *msginfo;


    DEBUGP("\nSIPALG: UDP Helper\n");

       

    /* Check whether the complete packet is in a single skb or spread across
       multiple buffers; We have to pull every thing into a single skb */
    //TODO

    
//  printk(KERN_WARNING "** skb len %d, dlen %d\n",(*pskb)->len,
//                                                (*pskb)->data_len);
    if (skb_is_nonlinear(*pskb))
    {
//      printk(KERN_WARNING "********** Non linear skb\n");
    }

#if 0
    if ((skb_is_nonlinear(skb) || skb_cloned(skb)) &&
            skb_linearize(skb, GFP_ATOMIC) != 0)
    {
                err = -ENOMEM;
                goto out;
    }
#endif

    /* Allocate buffer to keep message info while processing each msg */
    /* kmalloc is bit slow, instead we can think of using memory pool */
    //TODO

    msginfo = kmalloc(sizeof(sip_msg_info_t), GFP_ATOMIC);

    if (!msginfo)
    {
        DEBUGP("SIPALG: Memeory allocation failed\n");
        return(NF_DROP);
    }

    memset(msginfo, 0, sizeof(*msginfo));

    msginfo->ct     = ct;
    msginfo->ctinfo = ctinfo;
    msginfo->pskb   = pskb;
    msginfo->iph    = (*pskb)->nh.iph;
    msginfo->udph   = (struct udphdr *)((u_int32_t *)msginfo->iph +
                                                     msginfo->iph->ihl);
    msginfo->msglen = ntohs(msginfo->udph->len) - sizeof(struct udphdr);
    msginfo->msgptr = (char *) ((char*) msginfo->udph + sizeof(struct udphdr));
    msginfo->proto  = IPPROTO_UDP;

    /* Copying the mark from the packet to the conntrack
     * This is required for SIP over VPN */     
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
    msginfo->ct->mark = (*pskb)->mark;
#else
    msginfo->ct->mark = (*pskb)->nfmark;
#endif     

    /* We cant process if there is no message (Ex. keepalive) */
    //TODO: Is there any other way to handle incomplete/junk/keepalive msgs

//  if (msginfo->msglen < 5)
    if (!msginfo->msglen)
    {
        kfree(msginfo);
        return(NF_DROP);
    }

    /* Process sip messages */

    ret = sip_ct_handle_msg(msginfo);

    kfree(msginfo);

    DEBUGP("SIPALG: Leaving udp helper module\n");
 
    return(ret);
}


/* This helper is useful to indicate master that there is an active media within
 * a call, so that idletimer should not delete the call. This update is required
 * because every client may not enable keep alives within an active call.
 */

static int sip_ct_rtp_helper(struct sk_buff **pskb,
                             struct ip_conntrack *ct,
                             enum ip_conntrack_info ctinfo)
{
    int    callid, msgdir;

    sip_ct_info_t    *alginfo;
    sip_call_info_t    *pcinfo;


    /* This ct has callid that tells the call for which this media belongs.
       Raise media active flag on that call in master call list */

    spin_lock_bh(&sip_ct_lock);

    alginfo = (sip_ct_info_t*) ct->master->help.ct_alg_info;

    // Teamf1 for LB
    ct->mark = 0x00;

    callid  = (int) ct->help.ct_alg_info;

    /* I dont think these checks are required. But wanted to be on safe side */

    if ((alginfo == NULL) || (callid < 1) || (callid > SIP_CT_MAX_CALLS))
    {
        DEBUGP("SIPALG: Media is not associated to any call (palg %p,call %d\n",
                                                               alginfo, callid);
        spin_unlock_bh(&sip_ct_lock);
        return(NF_DROP);
    }

    /* Update activity when call is established or in held state */

    pcinfo = &alginfo->calllist[callid-1];

    if ((pcinfo->state == SIP_CT_STATE_INCALL) ||
        (pcinfo->state == SIP_CT_STATE_HELD))
    {
        pcinfo->rtpactive = TRUE;
        spin_unlock_bh(&sip_ct_lock);

        msgdir = sip_ct_get_msgdir(ct, ctinfo);

        if ((rtp_dscp) && (msgdir == SIP_CT_OUTBOUND)){
            DEBUGP("SIPALG:Going to set RTP DSCP\n");
            sip_ct_set_dscp(pskb, rtp_dscp);
        }
            DEBUGP("SIPALG:Not setting RTPRTP  DSCP\n");
        return(NF_ACCEPT);
    }

//  DEBUGP("SIPALG: Forcing media connection to death\n");
//  ip_ct_refresh(ct, *pskb, SIP_CT_FORCED_DEATH*HZ);

#if 1
    /* When clients are tend to use same media port (user configured port).
       It is possible that the same media ct/connection is used if calls are
       established more than once between same clients (establish call -
       disconnect - and before media ct expires, establish call again) */

    /* Here is the work around! May fail in some cases! Trying to reassociate
       media with new call, otherwise activity is updated on the old call! */

    callid = 1;
    pcinfo = alginfo->calllist;

//  ct->help.ct_alg_info = (unsigned long) 0;

    while (callid <= SIP_CT_MAX_CALLS)
    {
        if (pcinfo->state == SIP_CT_STATE_INCALL)
        {
            ct->help.ct_alg_info = (unsigned long) callid;
            break;
        }

        callid ++;
        pcinfo ++;
    }
#endif

    spin_unlock_bh(&sip_ct_lock);

    return(NF_ACCEPT);
}


/* RTP Helper updates media activity in master's call list to keep calls alive.
 * This structure is a dummy and is attached to the rtp ct, but not useful for
 * helper registraton. We are only interested at the helper function in the
 * entire structure.
 */

struct ip_conntrack_helper sip_ct_rtp_helpers =
{
    .list         = { NULL, NULL },
    .name         = "SIPALG",
    .me           = THIS_MODULE,
    .max_expected = 0,
    .timeout      = SIP_CT_DEFAULT_TIMEOUT,
    .help         = sip_ct_rtp_helper
};


#ifndef SIPALG_CFG

/* Port and transport protocol fields are configurable. For that we can use
 * module_param method just like lan_addr, lan_netmask. It is an easy and
 * extendable method.
 */

/* TCP Helper structure */

struct ip_conntrack_helper sip_ct_tcp_helpers =
{
    .list         = { NULL, NULL },
    .name         = "SIPALG",
    .me           = THIS_MODULE,
    .max_expected = SIP_CT_MAX_EXPS,
    .timeout      = SIP_CT_DEFAULT_TIMEOUT,

    .tuple.src.u.tcp.port = __constant_htons(SIP_CT_PORT),
    .tuple.dst.protonum   = IPPROTO_TCP,
    .mask.src.u.tcp.port  = 0xffff,
    .mask.dst.protonum    = 0xff,

    .help = sip_ct_tcp_helper
};


/* UDP Helper structure */

struct ip_conntrack_helper sip_ct_udp_helpers =
{
    .list         = { NULL, NULL },
    .name         = "SIPALG",
    .me           = THIS_MODULE,
    .max_expected = SIP_CT_MAX_EXPS,
    .timeout      = SIP_CT_DEFAULT_TIMEOUT,

    .tuple.src.u.udp.port = __constant_htons(SIP_CT_PORT),
    .tuple.dst.protonum   = IPPROTO_UDP,
    .mask.src.u.udp.port  = 0xffff,
    .mask.dst.protonum    = 0xff,

    .help = sip_ct_udp_helper
};


/* This function will be called at the time of loading sipalg module into the
 * kernel using insmod/modprobe command.
 */

static int __init init(void)
{
    int    ret;


    DEBUGP("SIPALG: Helper init\n");

    DEBUGP("SIPALG: LAN Network Address: 0x%x\n", lan_netaddr);
    DEBUGP("SIPALG: LAN Network Mask: 0x%x\n", lan_netmask);

    rmmod = FALSE;

    /* TCP Helper registration */

    ret = ip_conntrack_helper_register(&sip_ct_tcp_helpers);

    /* UDP Helper registration */

    if (!ret)
        ret = ip_conntrack_helper_register(&sip_ct_udp_helpers);

    if (!ret)
    {
        psip_ct_tcp_helpers = &sip_ct_tcp_helpers; 
        psip_ct_udp_helpers = &sip_ct_udp_helpers; 
    }

    return(ret);
}


/* This function will be called at the time of removing sipalg module from the
 * kernel using rmmod command.
 */

static void __exit fini(void)
{

    DEBUGP("SIPALG: Helper uninit... Wait.\n");

    rmmod = TRUE;

    do
    {
        schedule_timeout(SIP_CT_DESTROY_DELAY*HZ);
    } while (atomic_read(&sip_session_count));

    /* Unregistering TCP and UDP Helpers */

    ip_conntrack_helper_unregister(&sip_ct_tcp_helpers);
    ip_conntrack_helper_unregister(&sip_ct_udp_helpers);

    /* RTP helper is not registered but media connections are pointing to
       this helper. Trying to remove RTP helper pointers from media cts */

    if (!ip_conntrack_helper_register(&sip_ct_rtp_helpers))
        ip_conntrack_helper_unregister(&sip_ct_rtp_helpers);

    DEBUGP("SIPALG: Helper uninit... Done.\n");

    return;
}


module_init(init);
module_exit(fini);

#endif
