/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                        Copyright 2006 - 2010                             */
/*                    One Convergence Devices Pvt Ltd                       */
/*                         ALL RIGHTS RESERVED                              */
/*                                                                          */
/* Permission is hereby granted to licensees of One Convergence Devices Pvt */
/* Ltd' products to use or abstract this computer program for the sole      */
/* purpose of implementing a product based on One Convergence Devices Pvt   */
/* Ltd products. No other rights to reproduce, use, or disseminate this     */
/* computer program, whether in part or in whole, are granted.              */
/****************************************************************************/

/****************************************************************************
*  File: sip_ct_parser.c
*
*  Description: File contains functions to parse sip headers and related
*  information by sipalg module.
*
*  Author: sriphaneesh.n@oneconvergence.com
*
*  Revision History:
*
*****************************************************************************/

#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/kernel.h>
#include <linux/in.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>

//#define DEBUG
#include "sip_ct_helpers.h"

/*---------------------------------------------------------------------------*/

/* A generic token and value structure definition. By using this struct we can
 * define mapping between string constants with their associated int values.
 */

typedef struct
{
    const char    *token;
    int            toklen;
    int            tokval;
} token_value_t;


/* Token_value_t definition of SIP methods which we are processing in sipalg.
 * String are defined by RFC, but values are implementation specific.
 */

/* TF1: Adding the support of SUBSCIBE,NOTIFY & OPTIONS in methods */

token_value_t sip_methods[] =
{
    { "INVITE",       sizeof("INVITE"),    SIP_INVITE },
    { "ACK",          sizeof("ACK"),       SIP_ACK },
    { "BYE",          sizeof("BYE"),       SIP_BYE },
    { "REGISTER",     sizeof("REGISTER"),  SIP_REGISTER },
    { "SUBSCRIBE",    sizeof("SUBSCRIBE"), SIP_SUBSCRIBE },
    { "NOTIFY",       sizeof("NOTIFY"),    SIP_NOTIFY },
    { "OPTIONS",      sizeof("OPTIONS"),   SIP_OPTIONS },
    { "CANCEL",       sizeof("CANCEL"),    SIP_CANCEL },
    {  NULL,          sizeof(""),          SIP_IGNORE }
};


/* Token_value_t definition of SIP headers which we are interested in sipalg.
 * String are defined by RFC, but values are implementation specific.
 */

token_value_t sip_hdrs[] =
{
    { "Via:",           sizeof("Via:"),           SIP_VIA_HDR },
    { "Contact:",       sizeof("Contact:"),       SIP_CONTACT_HDR },
    { "Record-Route:",  sizeof("Record-Route:"),  SIP_RROUTE_HDR },
    { "Route:",         sizeof("Route:"),         SIP_ROUTE_HDR },
    { "From:",          sizeof("From:"),          SIP_FROM_HDR },
    { "To:",            sizeof("To:"),            SIP_TO_HDR },
    { "Cseq:",          sizeof("CSeq:"),          SIP_CSEQ_HDR },
    { "Call-ID:",       sizeof("Call-ID:"),       SIP_CALLID_HDR },
    { "Content-Length:",sizeof("Content-Length:"),SIP_CONTENT_LEN },
    { "Content-Type:",  sizeof("Content-Type:"),  SIP_CONTENT_TYPE },
    { "Expires:",       sizeof("Expires:"),       SIP_EXPIRE_TIME },
    {  NULL,            sizeof(""),               SIP_UNKNOWN_HDR }
};


/* This is simple strcasecmp function. I could not link it from standard lib,
 * so reproducing same function here.
 */

int my_stricmp(const char *s1, const char *s2)
{
    int    c1, c2;

    do
    {
        c1 = tolower(*s1++);
        c2 = tolower(*s2++);
    } while ((c1 == c2) && (c1 != '\0'));

    return(c1-c2);
}


/* This is simple strncasecmp function. I could not link it from standard lib,
 * so reproducing same function here.
 */

int my_strnicmp(const char *s1, const char *s2, size_t n)
{
    int    c1, c2;

    if (n==0)
        return 0;

    do
    {
        c1 = tolower(*s1++);
        c2 = tolower(*s2++);
    } while ((c1 == c2) && (c1 != '\0') && (--n));

    return(c1-c2);
}


/* Converts dotted decimal ip addresses (ip address strings) into 32 bit
 * network byte order numbers. Same as standard inet_aton library function.
 */

int my_inet_aton(char *ipstr, u_int32_t *ipaddr)
{
    u_int32_t  nbyte, value;


    nbyte = 0;

    while (nbyte < 4)
    {
        /* Dont go ahead, if first char is not a digit */

        if (!isdigit(*ipstr))
            break;

        /* IP adresses will not have three digits continuously */

        if (isdigit(*(ipstr+1)) && isdigit(*(ipstr+2)) && isdigit(*(ipstr+3)))
            break;

        /* This function is used in netfilter, but I dont know much about it */
 
        value = simple_strtoul(ipstr, (char **) &ipstr, 10);

        if (value > 255)
            break;

        ((u_int8_t*)ipaddr)[nbyte] = value;

        nbyte ++;

        if (*ipstr != '.')
            break;

        ipstr ++;
    }

    /* Could not collect required bytes? */

    if (nbyte != 4)
        return(FAILURE);

    return(SUCCESS);
}


/* This function return tokens from the given strings/sip headers.
 * Something like standard strtok function, but written to suit our requirement.
 */

char rbuff[256]; //This is used with in my_strtok funtion
                 //Don't separate and Don't use it for other purpose

char *my_strtok(char *pstr, char *dlstr, char *dlchar,
                char **starts, char **ends, char **pnext)
{
    int    i, j;


    /* This check is for careless callers. Please dont delete this check.
       Removing this check is hazardous! Whole parser takes avantage of it */

    if (!pstr)
    {
        *pnext = NULL;
        return   NULL;
    }

    /* Skip leading spaces */

    while ((*pstr == ' ') || (*pstr == '\t'))
        pstr++;

    /* End of string is identified by any one of '\0' '\r' '\n' */
    /* Find the first occurance of delimiter char */

    *dlchar = '\0';
    j       = 0;

    while ((*pstr != '\0') && (*pstr != '\r') && (*pstr != '\n'))
    {
        i = 0;

        while ((dlstr[i] != '\0') && (*pstr != dlstr[i]))
        {
            i++;
        }

        if (dlstr[i] == '\0')
        {
            if (!j)
                *starts = pstr;

            rbuff[j] = *pstr;
            pstr     ++;
        }
        else /* Found it! matching delimiter is notified to caller by dlchar */
        {
            *ends = pstr-1;

            rbuff[j] = '\0';
            *dlchar  = dlstr[i];
            *pnext   = pstr+1;  // This is the pointer to start with next time
            break;
        }

        j ++;
    }

    /* Encountered end of string char; Caller will get to know that by pnext */

    if ((*pstr == '\0') || (*pstr == '\r') || (*pstr == '\n'))
    {
        *ends = pstr-1;

        rbuff[j] = '\0';
        *pnext   = NULL;
    }

    if (j > 0)
    {
        /* Bonus :-) Skip trailing spaces also */

        j --;

        while ((rbuff[j] == ' ') || (rbuff[j] == '\t'))
        {
            rbuff[j] = '\0';
            j --;
        }

        return(rbuff);
    }
    else
        return(NULL);
}


/* Given a pointer to the header/message and end of message pointer, this
 * function return length of the header and gives the next header pointer.
 * SIP headers are sequence of strings terminated by "\r\n" (By RFC).
 */

static inline int sip_ct_get_hdrlen(char *phdr, char *pend, char **pnext)
{
    int    hdrlen;

    hdrlen = 0;

    while ((phdr < pend) && (*phdr != '\r'))
    {
        phdr ++;
        hdrlen ++;
    }

    if ((phdr >= pend) || (*(phdr+1) != '\n'))
    {
        *pnext = NULL;
        return(0);
    }

    *pnext = phdr+2;

    return(hdrlen+2);
}


/* This function searches token_value_t type data for the specified strings
 * and retruns the associated int value.
 */

static inline int sip_ct_search_token(char *pstr, token_value_t toklist[])
{
    int index;


    index = 0;

    while (toklist[index].token)
    {
        if (!my_strnicmp(toklist[index].token, pstr, toklist[index].toklen-1))
        {
            DEBUGP("Found %s\n", toklist[index].token);
            break;
        }
        index++;
    }

    return(toklist[index].tokval);
}


/* This function parses request line for ip address and port.
 *
 */

int sip_ct_parse_request_line(char *hdrptr, sip_hdr_info_t *reqinfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;

    DEBUGP("SIPALG: Parsing request line\n");

    /* Skipping Method sip: userinfo@ */

    ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "@: ", &dlchar, &starts, &ends, &pnext);

    /* Following will be true if userinfo is present in sip uri;
       Ignore userinfo and again look for host:port */

    if (dlchar == '@')
        ptok = my_strtok(pnext, ": \r", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if (my_inet_aton(ptok, &reqinfo->ipaddr1) != FAILURE)
        {
            reqinfo->ip1begin = starts;
            reqinfo->ip1end   = ends;
        }
    }

    /* Following will be true if port is specified in host:port */

    if (dlchar == ':')
    {
        ptok = my_strtok(pnext, " \r", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            reqinfo->port1 = (u_int16_t) simple_strtol(ptok, NULL, 10);
            reqinfo->port1 = htons(reqinfo->port1);
            reqinfo->port1begin = starts;
            reqinfo->port1end   = ends;
        }
    }
    else /* If there is no port we know what it is */
        reqinfo->port1 = htons(SIP_CT_PORT);

    return(SUCCESS);
}


/* This function parses via for ip address, port, protocol and if at all there
 * is something required for sipalg.
 */

int sip_ct_parse_via_hdr(char *hdrptr, sip_hdr_info_t *viainfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;

 
    DEBUGP("SIPALG: Parsing via header\n");

    /* Skipping till Via: SIP/2.0/ */

    ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "/", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "/", &dlchar, &starts, &ends, &pnext);

    /* Now we get transport protocol */

    ptok = my_strtok(pnext, " ", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if (!my_stricmp(ptok, "UDP"))
            viainfo->proto1 = IPPROTO_UDP;
        else if (!my_stricmp(ptok, "TCP"))
            viainfo->proto1 = IPPROTO_TCP;
        else
            return(FAILURE);
    }

    /* Hope this is the host:port specification */
    
    ptok = my_strtok(pnext, ":;", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if (my_inet_aton(ptok, &viainfo->ipaddr1) != FAILURE)
        {
            viainfo->ip1begin = starts;
            viainfo->ip1end   = ends;
        }
    }

    if (dlchar == ':')
    {
        ptok = my_strtok(pnext, ";", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            viainfo->port1 = (u_int16_t) simple_strtol(ptok, NULL, 10);
            viainfo->port1 = htons(viainfo->port1);
            viainfo->port1begin = starts;
            viainfo->port1end   = ends;
        }
    }
    else /* If there is no port we know what it is */
        viainfo->port1 = htons(SIP_CT_PORT);

    /* Just trying to find out whether maddr is specified or not */

    while ((pnext != NULL) && (dlchar != '\r') && (dlchar != ','))
    {
        ptok = my_strtok(pnext, "=;,\r", &dlchar, &starts, &ends, &pnext);

        if ((ptok) && (dlchar == '='))
        {
            if (!my_stricmp(ptok, "maddr"))
            {
                ptok = my_strtok(pnext, ";,\r", &dlchar,&starts,&ends,&pnext);

                if ((ptok) && (my_inet_aton(ptok,&viainfo->ipaddr2)!=FAILURE))
                {
                    viainfo->ip2begin = starts;
                    viainfo->ip2end   = ends;
                }

                break;
            }
        }
    }
 
    return(SUCCESS);
}


/* This function parses contact header for ip address, port, protocol and if
 * at all there is something required for sipalg.
 */

int sip_ct_parse_contact_hdr(char *hdrptr, sip_hdr_info_t *continfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Parsing contact header\n");

    /* Skipping Contact: Display Name <sip: userinfo */

    ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "<", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "@:;>", &dlchar, &starts, &ends, &pnext);

    /* Following will be true if userinfo is present in sip uri;
       Ignore userinfo and again look for host:port */

    if (dlchar == '@')
        ptok = my_strtok(pnext, ":;>", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if (my_inet_aton(ptok, &continfo->ipaddr1) != FAILURE)
        {
            continfo->ip1begin = starts;
            continfo->ip1end   = ends;
        }
    }

    /* Following will be true if port is specified in host:port */

    if (dlchar == ':')
    {
        ptok = my_strtok(pnext, ";>", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            continfo->port1 = (u_int16_t) simple_strtol(ptok, NULL, 10);
            continfo->port1 = htons(continfo->port1);

            continfo->port1begin = starts;
            continfo->port1end   = ends;
        }
    }
    else
        continfo->port1 = htons(SIP_CT_PORT);  /* We already know this! */


    /* Just trying to find out whther transport is specified or not */

    while ((pnext != NULL) && (dlchar != '>'))
    {
        ptok = my_strtok(pnext, "=;>", &dlchar, &starts, &ends, &pnext);

        if ((ptok) && (dlchar == '='))
        {
            if (!my_stricmp(ptok, "transport"))
            {
                ptok = my_strtok(pnext, ";>", &dlchar, &starts, &ends, &pnext);

                if (ptok)
                {
                    if (!my_stricmp(ptok, "udp"))
                        continfo->proto1 = IPPROTO_UDP;
                    else if (!my_stricmp(ptok, "tcp"))
                        continfo->proto1 = IPPROTO_TCP;
                    else
                        return(FAILURE);
                }
            }
            else if (!my_stricmp(ptok, "maddr"))
            {
                ptok = my_strtok(pnext, ";>", &dlchar, &starts, &ends, &pnext);

                if ((ptok) && (my_inet_aton(ptok,&continfo->ipaddr2)!=FAILURE))
                {
                    continfo->ip2begin = starts;
                    continfo->ip2end   = ends;
                }
            }
        }
    }

    /* Trying to get expire time negotiated with proxy */

    /* First skip the address specification */

    while ((pnext != NULL) && (dlchar != '>'))
    {
        ptok = my_strtok(pnext, ">", &dlchar, &starts, &ends, &pnext);
    }

    while (pnext != NULL)
    {
        ptok = my_strtok(pnext, "=;", &dlchar, &starts, &ends, &pnext);

        if ((ptok) && (dlchar == '=') && (!my_stricmp(ptok, "EXPIRES")))
        {
            ptok = my_strtok(pnext, ";", &dlchar, &starts, &ends, &pnext);

            if (ptok)
            {
                continfo->expires = simple_strtol(ptok, NULL, 10);

                /* Zero indicates time is not negotiated, so +1 manipulation */

                continfo->expires += 1;
            }
        }
    }

    return(SUCCESS);
}


/* This function parses record route header for ip address, port, protocol
 * and if at all there is something required for sipalg.
 */

int sip_ct_parse_rroute_hdr(char *hdrptr, sip_hdr_info_t *rrinfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Parsing record-route header\n");

    /* Skipping Recor-Route: Display Name <sip: userinfo */

    ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "<", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, ":", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "@:;>", &dlchar, &starts, &ends, &pnext);

    /* Following will be true if userinfo is present in sip uri;
       Ignore userinfo and again look for host:port */

    if (dlchar == '@')
        ptok = my_strtok(pnext, ":;>", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if (my_inet_aton(ptok, &rrinfo->ipaddr1) == FAILURE)
        {
            /* TODO: May be it is a domain name, then it is the place we have
               to add name resolver code. Lot of brain storming ahead. */

            DEBUGP("Decoding ip address failed\n");

            return(FAILURE);
        }
    }

    /* Following will be true if port is specified in host:port */

    if (dlchar == ':')
    {
        ptok = my_strtok(pnext, ";>", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            rrinfo->port1 = (u_int16_t) simple_strtol(ptok, NULL, 10);
            rrinfo->port1 = htons(rrinfo->port1);
        }
    }
    else
        rrinfo->port1 = htons(SIP_CT_PORT);  /* We already know this! */

    /* Just trying to find out whther transport is specified or not */

    while ((pnext != NULL) && (dlchar != '>'))
    {
        ptok = my_strtok(pnext, "=;>", &dlchar, &starts, &ends, &pnext);

        if ((ptok) && (dlchar == '=') && (!my_stricmp(ptok, "transport")))
        {
            ptok = my_strtok(pnext, ";>", &dlchar, &starts, &ends, &pnext);

            if (ptok)
            {
                if (!my_stricmp(ptok, "udp"))
                    rrinfo->proto1 = IPPROTO_UDP;
                else if (!my_stricmp(ptok, "tcp"))
                    rrinfo->proto1 = IPPROTO_TCP;
                else
                    return(FAILURE);
            }
        }
    }

    return(SUCCESS);
}


/* This function parses sdp media description field for port and protocol.
 *
 */

int sip_ct_parse_mediadescr(char *hdrptr, sip_hdr_info_t *sdpinfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Parsing sdp media description\n");

    /* Skipping "m= media-field" */

    ptok = my_strtok(hdrptr, "=", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        sdpinfo->port1 = (u_int16_t) simple_strtol(ptok, NULL, 10);
        sdpinfo->port1 = htons(sdpinfo->port1);

        sdpinfo->port1begin = starts;
        sdpinfo->port1end   = ends;

        ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            if (!my_stricmp(ptok, "RTP/AVP"))
            {
                sdpinfo->proto1 = IPPROTO_UDP;
                return(SUCCESS);
            }

            /* TODO: other than RTP/AVP */
        }
    }

    DEBUGP("SIPALG: Parsing sdp mediadescr failed\n");
    return(FAILURE);
}


/* This function parses sdp connection information field for ip address.
 *
 */

int sip_ct_parse_conninfo(char *hdrptr, sip_hdr_info_t *sdpinfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Parsing sdp connection information\n");

    /* Skipping "c= conn-field nettype" */

    ptok = my_strtok(hdrptr, "=", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if ( my_inet_aton(ptok, &sdpinfo->ipaddr1) == SUCCESS )
        {
            sdpinfo->ip1begin = starts;
            sdpinfo->ip1end   = ends;
            return(SUCCESS);
        }

        /* TODO: if we get sip uri here */
    }

    DEBUGP("SIPALG: Parsing sdp conninfo failed\n");
    return(FAILURE);
}


/* This function parses sdp origin information field for ip address.
 *
 */

int sip_ct_parse_origininfo(char *hdrptr, sip_hdr_info_t *sdpinfo)
{
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Parsing sdp origin information\n");

    /* Skipping "c= conn-field nettype" */

    ptok = my_strtok(hdrptr, "=", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        if ( my_inet_aton(ptok, &sdpinfo->ipaddr2) == SUCCESS )
        {
            sdpinfo->ip2begin = starts;
            sdpinfo->ip2end   = ends;
            return(SUCCESS);
        }

        /* TODO: if we get sip uri here */
    }

    DEBUGP("SIPALG: Parsing sdp origin info failed\n");
    return(FAILURE);
}


/* This function parses sdp fileds. Primary purpose of this funcion is to get
 * the media connection details like RTP ipaddr, port, protocol and others.
 * Additionally keeps record of pointers to the adp fields which contain lan
 * ip addreses.
 */

int sip_ct_parse_sdp(sip_msg_info_t *msginfo, sip_hdr_info_t *sdpinfo)
{
    char   *hdrptr, *nexthdr, *sdpend;
    int    ret, index, hdrlen;


    DEBUGP("SIPALG: Parsing sdp payload\n");

    ret    = SUCCESS;
    hdrptr = msginfo->sdpptr;
    sdpend = hdrptr+msginfo->sdplen-1;
    index  = msginfo->hdrcount;

    while ((hdrptr) && (ret != FAILURE))
    {
        hdrlen = sip_ct_get_hdrlen(hdrptr, sdpend, &nexthdr);

        if (!hdrlen)
        {
//          printk(KERN_WARNING "End of sdp %p\n", nexthdr);
            break;
        }

        switch (*hdrptr)
        {
        case 'm':

            /* Media description field; Get RTP port and protocol and keep
               the pointer in hdrlist for NAT */

            msginfo->hdrlist[index].hdrtype = SIP_SDP_MEDIADESCR;
            msginfo->hdrlist[index].hdrptr  = hdrptr;
            msginfo->hdrlist[index].hdrlen  = hdrlen;
            msginfo->hdrcount++;

            index ++;

            ret = sip_ct_parse_mediadescr(hdrptr, sdpinfo);

        break;

        case 'c':

            /* Connection information field; Get ip address and add pointer
               to the pointer for NAT */

            msginfo->hdrlist[index].hdrtype = SIP_SDP_CONNINFO;
            msginfo->hdrlist[index].hdrptr  = hdrptr;
            msginfo->hdrlist[index].hdrlen  = hdrlen;
            msginfo->hdrcount++;

            index   ++;

            ret = sip_ct_parse_conninfo(hdrptr, sdpinfo);

        break;

        case 'o':

            /* Owner information field; Used to identify Lan2Lan */

            msginfo->hdrlist[index].hdrtype = SIP_SDP_ORIGININFO;
            msginfo->hdrlist[index].hdrptr  = hdrptr;
            msginfo->hdrlist[index].hdrlen  = hdrlen;
            msginfo->hdrcount++;

            index   ++;

            ret = sip_ct_parse_origininfo(hdrptr, sdpinfo);

        break;
        }

        /* No space to keep more header pointers;
           This should not happen. Increase size and rebuild code */

        if (index == SIP_CT_MAX_HDRS)
        {
            DEBUGP("SIPALG: Header list is full\n");
            ret = FAILURE;
        }

        hdrptr = nexthdr;
    }

    return ret;
}


/*
 *
 */

int sip_ct_parse_callid(sip_msg_info_t *msginfo, char **pcidstr)
{
    int    index;

    char   dlchar, *starts, *ends, *pnext, *ptok;

    sip_hdr_list_t    *hdrlist;


    DEBUGP("SIPALG: Parsing call id\n");


    index   = 0;
    hdrlist = msginfo->hdrlist;

    while ((index < SIP_CT_MAX_HDRS) && (hdrlist->hdrtype != SIP_CALLID_HDR))
    {
        index ++;
        hdrlist ++;
    }

    if (hdrlist->hdrtype == SIP_CALLID_HDR)
    {
        /* Skipping upto "Call-Id : " */

        ptok = my_strtok(hdrlist->hdrptr, ":", &dlchar, &starts, &ends, &pnext);
        ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            *pcidstr = ptok;
            return(SUCCESS);
        }
    }

    return(FAILURE);
}


int sip_ct_parse_status(char *hdrptr)
{
    int    ret;

    char   dlchar, *starts, *ends, *pnext, *ptok;


    DEBUGP("SIPALG: Parsing status line\n");

    /* Skipping upto "SIP/2.0 " */

    ptok = my_strtok(hdrptr, " ", &dlchar, &starts, &ends, &pnext);
    ptok = my_strtok(pnext, " ", &dlchar, &starts, &ends, &pnext);

    if (ptok)
    {
        ret = simple_strtol(ptok, NULL, 10);
        return(ret);
    }

    return(FAILURE);
}


/* This function parses information that requires or helpful for immediate
 * processing. It also parses generic information like session parameters,
 * state info and etc.
 */

int sip_ct_quick_parse(int hdrtype, char *hdrptr,
                       char **mhend, sip_msg_info_t *msginfo)
{
    int     ret;
    char    *ptok, *pnext, dlchar, *starts, *ends;


    DEBUGP("SIPALG: Quick parse message header\n");

    ret = FAILURE;

    switch (hdrtype)
    {
    case SIP_CSEQ_HDR:

        /* Here we get the sip method from a response message */

        if (msginfo->msgtype == SIP_STATUS)
        {
            DEBUGP("SIPALG: Parsing cseq for method\n");

            /* Skipping upto "CSeq : seqnum " */

            ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
            ptok = my_strtok(pnext, " \t", &dlchar, &starts, &ends, &pnext);
            ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

            if (ptok)
            {
                msginfo->method = sip_ct_search_token(ptok, sip_methods);
                ret = SUCCESS;

                /* TODO : What if method is unknown? */
            }
        }
        else
            ret = SUCCESS;

    break;

    case SIP_EXPIRE_TIME:

        DEBUGP("SIPALG: Parsing expire time\n");

        /* Skipping upto "Expires : " */

        ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
        ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {

            ret = simple_strtol(ptok, NULL, 10);

            /* Zero indicates time is not negotiated, so +1 manipulation */

            msginfo->expires = ret + 1;
            

            ret = SUCCESS;
        }

    break;

    case SIP_CONTENT_LEN:

        /* Separte msg header and body (for now sdp only), so that sdp body
           parsing can be taken care by sdp parser */

        DEBUGP("SIPALG: Parsing content length\n");

        /* Skipping upto "Content-Length : " */

        ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
        ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            ret = simple_strtol(ptok, NULL, 10);

            if ((ret >= 0) && (ret < msginfo->msglen))
            {
                *mhend = *mhend - ret;
                msginfo->sdplen = ret;

                if (ret > 0)
                {
                    msginfo->sdpptr = *mhend + 1;
                    ret = TRUE;
                }
                else
                    ret = SUCCESS;
            }
        }

    break;

    case SIP_CONTENT_TYPE:

        /* This code is not important for now, may be required for future */

        DEBUGP("SIPALG: Parsing content type\n");

        /* Skipping upto "Content-Type : " */

        ptok = my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
        ptok = my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

        if (ptok)
        {
            ret = strcmp(ptok, "application/sdp");

            if (ret == 0)
            {
                msginfo->conttype = SIP_SDP_CONTENT;
                ret = SUCCESS;
            }

            ret = SUCCESS;

            /* TODO: What if other than sdp */
        }

    break;

    case SIP_RROUTE_HDR:

        /* This flag is required while creating expects */

        DEBUGP("SIPALG: Record route is present\n");

        msginfo->rrflag = 1;
        ret = TRUE;

    break;

    default:

        ret = TRUE;

    break;
    }

    return ret;
}


/* This is the main parser function for sipalg. It parses whole message and
 * prepares required information to go ahead. It also prepares a list of key
 * headers which require further processing.
 */

int sip_ct_parse_msg(char *msgptr, int msglen, sip_msg_info_t *msginfo)
{
    int    ret, index, hdrlen, hdrtype, prevhdr;

    char   *hdrptr, *nexthdr, *msgend;


    DEBUGP("SIPALG: Parsing message header\n");

    /* Classify message as request or status; Very important for us */

    if ((msginfo->msglen) && (*msgptr == '\r') && (*(msgptr+1) == '\n'))
    {
        DEBUGP("SIPALG: Keepalive message\n");

        msginfo->msgtype = SIP_KEEPALIVE;
        return(SUCCESS);
    }

    if (msginfo->msglen < 8)
    {
        msginfo->msgtype = SIP_KEEPALIVE;
        return(SUCCESS);
    }

    if (strncmp(msgptr, "SIP", 3) != 0)
    {
        DEBUGP("SIPALG: Request message\n");

        msginfo->msgtype = SIP_REQUEST;
        msginfo->method  = sip_ct_search_token(msgptr, sip_methods);

        /* TODO : What if method is unknown? */
    }
    else
    {
        msginfo->msgtype = SIP_STATUS;
        msginfo->status  = sip_ct_parse_status(msgptr);

        DEBUGP("SIPALG: Status message %d\n", msginfo->status);
    }

    ret     = SUCCESS;
    index   = 0;
    prevhdr = 0;
    nexthdr = NULL;
    msgend  = msgptr+msglen-1;

    /* Add request or status line to the hdrlist; It may be required for nat */

    hdrlen = sip_ct_get_hdrlen(msgptr, msgend, &nexthdr);

    msginfo->hdrlist[index].hdrtype = SIP_REQUEST_STATUS;
    msginfo->hdrlist[index].hdrptr  = msgptr;
    msginfo->hdrlist[index].hdrlen  = hdrlen;
    msginfo->hdrcount ++;
    index ++;

    /* Go through the entire message and add required pointers to the hdrlist */

    while ((ret != FAILURE) && (nexthdr))
    {
        hdrptr  = nexthdr;

        /* Get length of the current header, and we also get nexthdr pointer */

        hdrlen  = sip_ct_get_hdrlen(hdrptr, msgend, &nexthdr);

        /* No more headers for parsing */

        if (hdrlen == 0)
            break;

        /* Check and to the hdrlist if we are interested at this header;
           Unknown means, we are not interested at this header */

        hdrtype = sip_ct_search_token(hdrptr, sip_hdrs);

        /* We may find multiple headers of same type Via/Route/Record-route
           and etc. Our interest is only first occurance of those headers,
           rest all can be ignored */

        if ((hdrtype != SIP_UNKNOWN_HDR) && (hdrtype != prevhdr))
        {
            /* Get information that require immediate attention (Eg. msgend) */
 
            ret = sip_ct_quick_parse(hdrtype, hdrptr, &msgend, msginfo);

            /* Quick parse says, add pointer to hdrlist for later processing */

            if (ret == TRUE)
            {
                msginfo->hdrlist[index].hdrtype = hdrtype;
                msginfo->hdrlist[index].hdrptr  = hdrptr;
                msginfo->hdrlist[index].hdrlen  = hdrlen;
                msginfo->hdrcount ++;
                prevhdr = hdrtype;
                index ++;

                /* No space to keep more header pointers;
                   This should not happen. Increase size and rebuild code */

                if (index == SIP_CT_MAX_HDRS)
                {
                    DEBUGP("SIPALG: Header list is full\n");
                    ret = FAILURE;
                }
            }
        }
    }

    /* Quick parse found sdp payload, keep the pointer for later processing */

    if ((ret != FAILURE) && (msginfo->sdpptr))
    {
        msginfo->hdrlist[index].hdrtype = SIP_SDP_CONTENT;
        msginfo->hdrlist[index].hdrptr  = msginfo->sdpptr;
        msginfo->hdrlist[index].hdrlen  = msginfo->sdplen;
        msginfo->hdrcount ++;

        /* No space to keep more header pointers;
           This should not happen. Increase size and rebuild code */

        if (index == SIP_CT_MAX_HDRS)
        {
            DEBUGP("SIPALG: Header list is full\n");
            ret = FAILURE;
        }
    }

    return(ret);
}
