/* Copyright (c) 2008, TeamF1, Inc. */

/*
modification history
--------------------
01a,24sep08,sat Added Hardware offload support
*/

/****************************************************************************/
/*                                                                          */
/*                                                                          */
/*                        Copyright 2006 - 2010                             */
/*                    One Convergence Devices Pvt Ltd                       */
/*                         ALL RIGHTS RESERVED                              */
/*                                                                          */
/* Permission is hereby granted to licensees of One Convergence Devices Pvt */
/* Ltd' products to use or abstract this computer program for the sole      */
/* purpose of implementing a product based on One Convergence Devices Pvt   */
/* Ltd products. No other rights to reproduce, use, or disseminate this     */
/* computer program, whether in part or in whole, are granted.              */
/****************************************************************************/

/****************************************************************************
*  File: sip_nat_helpers.c
*
*  Description: File contains sipalg module nat functions for Linux Netfilter
*
*  Author: sriphaneesh.n@oneconvergence.com
*
*  Revision History:
*
*****************************************************************************/

#include <linux/in.h>
#include <linux/netfilter.h>
#include <linux/interrupt.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>

//#define DEBUG
#include "sip_ct_helpers.h"


/* Module's global data */

extern struct ip_conntrack_helper *psip_ct_tcp_helpers;
extern struct ip_conntrack_helper *psip_ct_udp_helpers;
extern struct ip_conntrack_helper sip_ct_rtp_helpers;
extern spinlock_t sip_ct_lock;

/*---------------------------------------------------------------------------*/

void sip_nat_setup_info(struct ip_conntrack *ct,
                        struct ip_conntrack_expect *exp)
{
    struct ip_nat_range range;

    BUG_ON(ct->status & IPS_NAT_DONE_MASK);

    if (!exp->saved_proto.udp.port)
    {
        range.flags  = IP_NAT_RANGE_MAP_IPS;
        range.min_ip = range.max_ip
                     = ct->master->tuplehash[!exp->dir].tuple.dst.ip;

        ip_nat_setup_info(ct, &range, NF_IP_POST_ROUTING);
    }
    else
    {
        range.flags  = IP_NAT_RANGE_PROTO_SPECIFIED | IP_NAT_RANGE_MAP_IPS;
        range.min_ip = range.max_ip
                     = ct->master->tuplehash[!exp->dir].tuple.src.ip;
        range.min    = range.max = exp->saved_proto;

        ip_nat_setup_info(ct, &range, NF_IP_PRE_ROUTING);
    }
}


/* This function is going to be called by nefilter when singalling expectation
 * is initiating connection from Internet and when nat is enabled on this
 * session. Again we should force these packets to go through the sip ct helper
 * function.
 */

static void sip_nat_expectfn(struct ip_conntrack *new,
                             struct ip_conntrack_expect *exp1)
{

    DEBUGP("SIPALG: Nat expectfn for singalling\n");

    /* We know this flow is signalling, initialize helper on this ct */

    write_lock_bh(&ip_conntrack_lock);

    if (exp1->tuple.dst.protonum == IPPROTO_TCP)
        new->helper = psip_ct_tcp_helpers;
    else
        new->helper = psip_ct_udp_helpers;

    write_unlock_bh(&ip_conntrack_lock);

    /* If this expectation is not perminent, remove expect from our list. 
       Right now all sipalg expects are permanent, so it will never be true */

    if ((exp1->flags & IP_CT_EXPECT_PERMANENT) != IP_CT_EXPECT_PERMANENT)
    {
        sip_ct_info_t    *alginfo;
        sip_exp_info_t    *expinfo;

        spin_lock_bh(&sip_ct_lock);

        alginfo = (sip_ct_info_t*) exp1->master->help.ct_alg_info;

        if (alginfo == NULL)
        {
            spin_unlock_bh(&sip_ct_lock);
            return;
        }

        expinfo = sip_ct_get_expinfo(exp1, alginfo);

        if (expinfo == NULL)
        {
            spin_unlock_bh(&sip_ct_lock);
            return;
        }

        write_lock_bh(&ip_conntrack_lock);
        expinfo->exp1->flags = 0;
        write_unlock_bh(&ip_conntrack_lock);

        ip_conntrack_expect_put(expinfo->exp1);
        expinfo->exp1 = NULL;

        spin_unlock_bh(&sip_ct_lock);
    }

    /* This function sets up the reqired nat on this connection */

    sip_nat_setup_info(new, exp1);

    return;
}


/* This function is going to be called by nefilter when media expectation
 * is initiating connection from Internet and when the connection is going
 * through nat.
 */

static void sip_nat_rtp_expectfn(struct ip_conntrack *new,
                                 struct ip_conntrack_expect *exp1)
{
    int    usedby, callid;

    sip_ct_info_t    *alginfo;
    sip_exp_info_t    *expinfo;


    DEBUGP("SIPALG: Nat expectfn for rtp media\n");

    write_lock_bh(&ip_conntrack_lock);

    new->helper = &sip_ct_rtp_helpers;

    write_unlock_bh(&ip_conntrack_lock);

    /* Trying to associate this media connection with the call. Association
       is useful to update media activity in session's call list */

    spin_lock_bh(&sip_ct_lock);

    alginfo = (sip_ct_info_t*) exp1->master->help.ct_alg_info;

    if (alginfo == NULL)
    {
        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    expinfo = sip_ct_get_expinfo(exp1, alginfo);

    if (!expinfo)
    {
        spin_unlock_bh(&sip_ct_lock);
        return;
    }

    /* If this expectation is not perminent, remove expect from our list.
       Right now all sipalg expects are permanent, so it will never be true */

    if ((exp1->flags & IP_CT_EXPECT_PERMANENT) != IP_CT_EXPECT_PERMANENT)
    {
        write_lock_bh(&ip_conntrack_lock);
        expinfo->exp1->flags = 0;
        write_unlock_bh(&ip_conntrack_lock);

        expinfo->exp1 = NULL;
        ip_conntrack_expect_put(exp1);
    }

    /* Find who is expecting this media in session's calllist */

    usedby = expinfo->usedby;

    callid = 1;

    while (callid <= SIP_CT_MAX_CALLS)
    {
        if (CHECK_BIT(callid, usedby))
            break;
        callid ++;
    }

    /* Assign callid in this ct to update media activity in that callinfo */

    if (callid <= SIP_CT_MAX_CALLS)
    {
        DEBUGP("SIPALG: Media for callid %d\n", callid);

        new->help.ct_alg_info = (unsigned long) callid;
    }
    else
    {
        DEBUGP("SIPALG: Don't know callid of media\n");
    }

    spin_unlock_bh(&sip_ct_lock);

    /* This function sets up the reqired nat on this connection */

    sip_nat_setup_info(new, exp1);

    return;
}


/* This function creates expect for natted connections with the translated ip
 * address and port numbers. And also updates port mappings with respect to
 * different connections. This function is called only with outbound messages.
 */

    /* Expectation address and ports are changed to nat address and port for
       inbound connections, and a different expectation function is used for
       both in and outbound connections if nat is enabled on this session */

int sip_nat_create_expect(int exptype, struct ip_conntrack_expect *exp1,
                          sip_exp_info_t *expinfo, sip_msg_info_t *msginfo,
                          sip_ct_info_t *alginfo)
{
    u_int16_t    newport;

    int    ret;
 

    DEBUGP("SIPALG: Creating nat expect\n");

    /* Direction of the expected connection with respect to master is required
       to apply nat policy on the expected connection */

    if (alginfo->initdir == msginfo->msgdir)
        exp1->dir = IP_CT_DIR_REPLY;
    else
        exp1->dir = IP_CT_DIR_ORIGINAL;

    if (exptype == SIP_CT_EXP_SIP)
        exp1->expectfn = sip_nat_expectfn;
    else
        exp1->expectfn = sip_nat_rtp_expectfn;

    /* Save original port for directing new connection to the original port */

//  exp1->saved_proto.udp.port = exp1->tuple.dst.u.udp.port;

    /* Every connection in this session have to use this natip only */

//  exp1->tuple.dst.ip = alginfo->natip;

    if (msginfo->msgdir == SIP_CT_OUTBOUND)
    {
        exp1->saved_proto.udp.port = exp1->tuple.dst.u.udp.port;
        exp1->tuple.dst.ip         = alginfo->natip;
    }
    else
        exp1->saved_proto.udp.port = 0;

    /* Use the same port if available, otherwise create new port */

    newport = ntohs(exp1->tuple.dst.u.udp.port);

    if ((exptype == SIP_CT_EXP_RTCP) && (msginfo->msgdir == SIP_CT_OUTBOUND))
        newport = ntohs(msginfo->rtcpnatport);

    ret = FAILURE;

    for (; newport != 0; newport++)
    {
        exp1->tuple.dst.u.udp.port = htons(newport);

        if ((ret = ip_conntrack_expect_related(exp1)) == SUCCESS)
            break;
#ifdef CONFIG_TF1_HW_OFFLOAD
        DEBUGP("Dynamic port registration %d %d\n",exptype,exp1->tuple.dst.u.udp.port);
        hwOffloadRegister(exp1->tuple.dst.u.udp.port,1);
#endif

        if ((msginfo->msgdir == SIP_CT_INBOUND) || (exptype == SIP_CT_EXP_RTCP))
            break;

        if (exptype == SIP_CT_EXP_RTP)
            newport++;
    }

    if (ret != SUCCESS)
    {
        DEBUGP("SIPALG: Nat expect create failed\n");
        return(FAILURE);
    }

    /* Update nat port in expect list */

    if (msginfo->msgdir == SIP_CT_OUTBOUND)
        expinfo->natport = exp1->tuple.dst.u.udp.port;

//  printk(KERN_WARNING " *** New port %d ***\n", ntohs(expinfo->natport));

    return(SUCCESS);
}


/* Given a port this function returns mapped port correspoinding to the
 * specified port. Port map is already built at the time of expect creation.
 */

static u_int16_t sip_nat_get_portmap(int msgdir, sip_hdr_info_t *hdrinfo,
                                     sip_ct_info_t *alginfo)
{
    int    index;

    sip_exp_info_t    *expinfo;


    index   = 0;
    expinfo = alginfo->explist;

    if (msgdir == SIP_CT_OUTBOUND)
        {
        /* We need natted port mapped to the original port */

        while (index < SIP_CT_MAX_EXPS)
            {
            if (expinfo->origtuple.dst.ip == hdrinfo->ipaddr1)
                {
                if (expinfo->origtuple.dst.u.udp.port == hdrinfo->port1)
                    return(expinfo->natport); 
                }

            index   ++;
            expinfo ++;
            }
        }
    else
        {
        /* We need original port mapped to the natted port */

        while (index < SIP_CT_MAX_EXPS)
            {
            if (expinfo->natport == hdrinfo->port1)
                return(expinfo->origtuple.dst.u.udp.port); 

            index   ++;
            expinfo ++;
            }
        }

    DEBUGP("SIPALG: Port mapping is not available\n");
    //  return(0);
    return(hdrinfo->port1);
}


/* This function updates content length field in the sip message, if there is
 * a difference after mangling sdp content.
 */

int sip_nat_mangle_content_len(sip_msg_info_t *msginfo, int difflen)
{
    int    matchoff, matchlen, newlen, index;

    char    *hdrptr, *pnext, *starts, *ends, dlchar;
    char    newbuff[8];


    DEBUGP("SIPALG: Mangling content len\n");

    /* Parser already adds content len field pointer to hdrlist if sdp content
       is present in the message. Search hdrlist for this pointer */

    index = 0; pnext = starts = ends = NULL;

    while ((index < SIP_CT_MAX_HDRS) &&
           (msginfo->hdrlist[index].hdrtype != SIP_CONTENT_LEN))
        index++;

    /* We dont know location of the length in message, parse hdr for that */

    if (msginfo->hdrlist[index].hdrtype == SIP_CONTENT_LEN)
    {
        hdrptr = msginfo->hdrlist[index].hdrptr;

        /* Skipping upto "Content-Length : " */

        my_strtok(hdrptr, ":", &dlchar, &starts, &ends, &pnext);
        my_strtok(pnext, "\r\n", &dlchar, &starts, &ends, &pnext);

        if ((!starts) || (!ends))
            goto badreturn;

        matchoff = starts - msginfo->msgptr;
        matchlen = ends - starts + 1;

        /* Adjust length and write back to the message */

        newlen = sprintf(newbuff, "%u", msginfo->sdplen + difflen);

        msginfo->msglendiff += (newlen - matchlen);

        /* Call netfilter function to mangle packet, it returns 1 on success */

        if (ip_nat_mangle_packet(msginfo->pskb, msginfo->ct, msginfo->ctinfo,
                                         matchoff, matchlen, newbuff, newlen))
            return(SUCCESS);
    }

badreturn:

    DEBUGP("SIPALG: Mangling content length failed\n");

    return(FAILURE);
}


/* This function replaces local ip address and ports with nat ip address and
 * ports that are carried in sip sdp content. Hdrinfo is supposed to have
 * start and end pointers to the fields which needs to be modified.
 */

int sip_nat_mangle_sdp(sip_hdr_info_t *hdrinfo,
                       sip_msg_info_t *msginfo,
                       sip_ct_info_t *alginfo,
                       struct ip_conntrack_expect *exp1)
{
    int    matchoff, matchlen, newlen;
    int    index, done, ret;

    u_int32_t    newip;
    u_int16_t    newport;

    char    newbuff[25];


    DEBUGP("SIPALG: Mangling sdp content\n");

    index = msginfo->hdrcount; ret = SUCCESS; done = 0;

    while ((ret != FAILURE) && ((index--) >= 0) && (done < 2))
    {
        switch (msginfo->hdrlist[index].hdrtype)
        {
        case SIP_SDP_MEDIADESCR:

            /* Mangle 'm' field with new port number */

            if ((hdrinfo->port1begin) && (hdrinfo->port1end))
            {
                matchoff = hdrinfo->port1begin - msginfo->msgptr;
                matchlen = hdrinfo->port1end - hdrinfo->port1begin + 1;
            }
            else
            {
                ret = FAILURE;
                break;
            }

            /* Get new port number to write into sdp content */

            if (exp1)
                newport = exp1->saved_proto.udp.port;
            else
                newport = sip_nat_get_portmap(msginfo->msgdir,hdrinfo,alginfo);

            if (!newport)
            {
                ret = FAILURE;
                break;
            }

            newlen   = sprintf(newbuff, "%u", ntohs(newport));

            msginfo->sdplendiff += (newlen - matchlen);
            msginfo->msglendiff += (newlen - matchlen);

            /* Calling netfilter mangle, it returns 1 on success */

            if (!ip_nat_mangle_packet(msginfo->pskb, msginfo->ct,
                                      msginfo->ctinfo, matchoff, matchlen,
                                                          newbuff, newlen))
                ret = FAILURE;

            done ++;

        break;

        case SIP_SDP_CONNINFO:

            /* Mangle 'c' field with ip address */

            if (!hdrinfo->ipaddr1)  //Call hold message
                break;
            
            if ((hdrinfo->ip1begin) && (hdrinfo->ip1end))
            {
                matchoff = hdrinfo->ip1begin - msginfo->msgptr;
                matchlen = hdrinfo->ip1end - hdrinfo->ip1begin + 1;
            }
            else
            {
                ret = FAILURE;
                break;
            }

            /* Get and format new ip address to write into sdp content */

            if (exp1)
                newip = hdrinfo->ipaddr2;
            else
                newip = alginfo->natip;

            newlen = sprintf(newbuff, "%u.%u.%u.%u", NIPQUAD(newip));

            msginfo->sdplendiff += (newlen - matchlen);
            msginfo->msglendiff += (newlen - matchlen);

            /* Calling netfilter mangle, it returns 1 on success */

            if (!ip_nat_mangle_packet(msginfo->pskb, msginfo->ct,
                                      msginfo->ctinfo, matchoff, matchlen,
                                                          newbuff, newlen))
                ret = FAILURE;

            done ++;

        break;

            /* Actually we have to mangle owner info 'o' field also;
               But to support lan2lan calls in proxy setup, we are leaving 
               original ip in 'o' field as it is */
        }
    }

    if (ret == FAILURE)
        DEBUGP("SIPALG: Mangling sdp header failed\n");

    return(ret);
}


/* This function replaces local ip address and ports with nat ip address and
 * ports that are carried in sip message headers (and vice versa). Hdrinfo is
 * supposed to have start and end pointers to the fields which needs to be
 * modified.
 */

int sip_nat_mangle_hdr(int hdrtype, sip_hdr_info_t *hdrinfo,
                       sip_msg_info_t *msginfo, sip_ct_info_t *alginfo)
{
    int    matchlen, matchoff, newlen;

    u_int32_t    newip;
    u_int16_t    newport;

    char    newbuff[25];


    DEBUGP("SIPALG: Mangling message header\n");

    /* While sending out to internet, replace lan ip:port with nat ip:port
       And while sending into lan replace nat ip:port with lan ip:port */

    if (msginfo->msgdir == SIP_CT_OUTBOUND)
        newip = alginfo->natip;
    else
        newip = alginfo->origip;

    /* If maddr field exist in the packet, ip2begin and ip2end pointers points
       to that field in the msg. We have to replace it with newip address */

    if ((hdrinfo->ip2begin) && (hdrinfo->ip2end))
    {
        matchoff = hdrinfo->ip2begin - msginfo->msgptr;
        matchlen = hdrinfo->ip2end - hdrinfo->ip2begin + 1;

        newlen = sprintf(newbuff, "%u.%u.%u.%u", NIPQUAD(newip));

        /* Call netfilter function to mangle packet, it returns 1 on success */

        msginfo->msglendiff += (newlen - matchlen);

        if (!ip_nat_mangle_packet(msginfo->pskb, msginfo->ct, msginfo->ctinfo,
                                          matchoff, matchlen, newbuff, newlen))
            goto badreturn;
    }

    /* While sending out to internet, replace lan ip:port with natip:port
       And while sending into lan replace natip:port with original ip:port */

    newport = sip_nat_get_portmap(msginfo->msgdir, hdrinfo, alginfo);

    if (!newport)
        goto badreturn;

    /* There are chances to have a sip uri in place of ip address. In such
       situations we have to replace only port field. Some times port is also
       not specified with address field. In this case alg assumes default port,
       and creates expect. But nat wont guarantee default port, so without
       second thought write always ip:port */

    matchoff = matchlen = newlen = 0;

    if ((hdrinfo->ip1begin) && (hdrinfo->ip1end))
    {
        matchoff = hdrinfo->ip1begin - msginfo->msgptr;
        matchlen = hdrinfo->ip1end - hdrinfo->ip1begin + 1;

        /* Port is present or not, but we always write ip:port */

        newlen = sprintf(newbuff, "%u.%u.%u.%u:%u", NIPQUAD(newip),
                                                   ntohs(newport));
    }

    if ((hdrinfo->port1begin) && (hdrinfo->port1end))
    {
        if (!matchoff)
        {
            matchoff = hdrinfo->port1begin - msginfo->msgptr;
            matchlen = hdrinfo->port1end - hdrinfo->port1begin + 1;

            /* There is no ip, so we have to replace only port */

            newlen = sprintf(newbuff, "%u", ntohs(newport));
        }
        else
            matchlen = hdrinfo->port1end - hdrinfo->ip1begin + 1;
    }

    /* Looks like we dont have any thing to mangle */

    if (!matchlen)
    {
        //TODO: We may have to handle both ip&port are not mentioned
        //We need to change parser to handle this.
        return(SUCCESS);
    }

    msginfo->msglendiff += (newlen - matchlen);

    /* Call netfilter function to mangle packet, it returns 1 on success */

    if (ip_nat_mangle_packet(msginfo->pskb, msginfo->ct, msginfo->ctinfo,
                                     matchoff, matchlen, newbuff, newlen))
        return(SUCCESS);

badreturn:

    DEBUGP("SIPALG: Mangling message header failed\n");

    return(FAILURE);
}
