/*
** Copyright (C) 2002-2008 Sourcefire, Inc.
** Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License Version 2 as
** published by the Free Software Foundation.  You may not use, modify or
** distribute this program under any other version of the GNU General
** Public License.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* $Id$ */
/* Modification History
 * --------------------
 */

/* spo_alert_teamf1 
 * 
 * Purpose:
 *
 * This module sends alerts to the teamf1 service.
 *
 * Arguments:
 *   
 * Logging mechanism?
 *
 * Effect:
 *
 * Alerts are written to the teamf1 service with in the facility indicated by
 * the module arguments.
 *
 * Comments:
 *
 * First try
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <stdlib.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "decode.h"
#include "event.h"
#include "rules.h"
#include "plugbase.h"
#include "spo_plugbase.h"
#include "debug.h"
#include "parser.h"
#include "mstring.h"
#include "util.h"
#include "strlcatu.h"
#include "strlcpyu.h"

#include "snort.h"
#include "spo_alert_teamf1.h"
#include "inline.h"

#define SYSLOG_BUF  1024
#define UMI_BUF     SYSLOG_BUF + 128

typedef struct _Teamf1Data
{
    int     facility;
    int     priority;
    int     options;
    int     sockFamily;
    int     sockFd;
    char    msgBuf[UMI_BUF];
} Teamf1Data;

void AlertTeamf1Init(char *);
Teamf1Data *ParseTeamf1Args(char *);
void AlertTeamf1(Packet *, char *, void *, Event *);
void AlertTeamf1CleanExit(int, void *);
void AlertTeamf1Restart(int, void *);


void umiIoctl
    (
    Teamf1Data *    pData,
    int             destId,
    int             cmd,
    void *          pInBuf,
    int             inLen,
    void *          pOutBuf,
    int             outLen
    )
    {
    char *          pBuf = pData->msgBuf;
    UMI_REQ_INFO *  pUmiReq;

    pUmiReq = (UMI_REQ_INFO *)pBuf;
    pBuf +=  sizeof (UMI_REQ_INFO);

    memcpy ((char *)pBuf, (char *)&inLen, sizeof (int));
    pBuf += sizeof (unsigned int);

    memcpy ((void *)pBuf, (void *)pInBuf, inLen);
    pBuf += inLen;

    memcpy ((char *)pBuf, (char *)&outLen, sizeof (int));
    pBuf += sizeof (unsigned int);

    if (pOutBuf)
        {
        memcpy(pBuf , pOutBuf, outLen);
        pBuf += outLen;
        }

    pUmiReq->cmd    = cmd;
    pUmiReq->srcId  = UMI_COMP_IPS;
    pUmiReq->reqOpt = UMI_IOCTL_OPT_NO_RESULT;
    pUmiReq->reqId  = 0;
    pUmiReq->req    = 1;

    sendto (pData->sockFd, pUmiReq, pBuf - (char *)pUmiReq, 0,
             (struct sockaddr *) &destId, sizeof(UMI_COMP_ID));
    }

/*
 * Function: sendMsg()
 *
 * Purpose: Send the message to TeamF1 Logging daemon
 *
 * Arguments: pointer to Teamf1Data, pointer to message
 *
 * Returns: void function
 *
 */
void sendMsg 
    (
    Teamf1Data *    pData,
    char *          pMsg
    )
    {
    LOGGING_LOG_REQ logReq; 
    int             len;

    if (pData->sockFd == -1)
        return;

    logReq.compId = UMI_COMP_IPS;
    logReq.logLevel = pData->priority;
    logReq.text[0] = 0;
    len = strlen (pMsg) + 1;

    umiIoctl (pData, UMI_COMP_LOGGING, LOGGING_CMD_LOG,
             (void *)&logReq, sizeof (LOGGING_LOG_REQ), (void *)pMsg, len);
    }

/*
 * Function: SetupTeamf1()
 *
 * Purpose: Registers the output plugin keyword and initialization 
 *          function into the output plugin list.  This is the function that
 *          gets called from InitOutputPlugins() in plugbase.c.
 *
 * Arguments: None.
 *
 * Returns: void function
 *
 */
void AlertTeamf1Setup(void)
{
    /* link the preprocessor keyword to the init function in 
       the preproc list */
    RegisterOutputPlugin("alert_teamf1", NT_OUTPUT_ALERT, AlertTeamf1Init);
    DEBUG_WRAP(DebugMessage(DEBUG_INIT,"Output plugin: Alert-Teamf1 is setup...\n"););
}


/*
 * Function: AlertTeamf1Init(char *)
 *
 * Purpose: Calls the argument parsing function, performs final setup on data
 *          structs, links the preproc function into the function list.
 *
 * Arguments: args => ptr to argument string
 *
 * Returns: void function
 *
 */
void AlertTeamf1Init(char *args)
{
    Teamf1Data *data;
    DEBUG_WRAP(DebugMessage(DEBUG_INIT, "Output: Alert-Teamf1 Initialized\n"););

    pv.alert_plugin_active = 1;

    /* parse the argument list from the rules file */
    data = ParseTeamf1Args(args);

    data->sockFd = socket (data->sockFamily, 0, UMI_COMP_UMI);

    DEBUG_WRAP(DebugMessage(DEBUG_INIT,"Linking teamf1 alert function to call list...\n"););

    /* Set the preprocessor function into the function list */
    AddFuncToOutputList(AlertTeamf1, NT_OUTPUT_ALERT, data);
    AddFuncToCleanExitList(AlertTeamf1CleanExit, data);
    AddFuncToRestartList(AlertTeamf1Restart, data);
}



/*
 * Function: ParseTeamf1Args(char *)
 *
 * Purpose: Process the preprocessor arguements from the rules file and 
 *          initialize the preprocessor's data struct.  This function doesn't
 *          have to exist if it makes sense to parse the args in the init 
 *          function.
 *
 * Arguments: args => argument list
 *
 * Returns: void function
 *
 */
Teamf1Data *ParseTeamf1Args(char *args)
{
    Teamf1Data *data;

    data = (Teamf1Data *)SnortAlloc(sizeof(Teamf1Data));

    /* default values for teamf1 output */
    data->options = 0;
    data->facility = 0;
    data->priority = LOGGING_LEVEL_ALERT;

    if (args)
        {
        data->sockFamily = atoi (args);
        }
    return data;
}


/*
 * Function: PreprocFunction(Packet *)
 *
 * Purpose: Perform the preprocessor's intended function.  This can be
 *          simple (statistics collection) or complex (IP defragmentation)
 *          as you like.  Try not to destroy the performance of the whole
 *          system by trying to do too much....
 *
 * Arguments: p => pointer to the current packet data struct 
 *
 * Returns: void function
 *
 */
extern OptTreeNode *otn_tmp;
void AlertTeamf1(Packet *p, char *msg, void *arg, Event *event)
{
    char sip[16];
    char dip[16];
    char pri_data[STD_BUF];
    char ip_data[STD_BUF];
    char event_data[STD_BUF];
    char event_string[SYSLOG_BUF];
    Teamf1Data *data = (Teamf1Data *)arg;

    if (iv.reject)
        strcpy (event_string, "*REJECT*");
    else if (iv.drop)
        strcpy (event_string, "*DROP*");
    else
        strcpy (event_string, "*ALERT*");

    /* Remove this check when we support IPv6 below. */
    /* sip and dip char arrays need to change size for IPv6. */
    if (!IS_IP4(p))
    {
        return;
    }

    if(p && IPH_IS_VALID(p))
    {
        if (strlcpy(sip, inet_ntoa(GET_SRC_ADDR(p)), sizeof(sip)) >= sizeof(sip))
            return;

        if (strlcpy(dip, inet_ntoa(GET_DST_ADDR(p)), sizeof(dip)) >= sizeof(dip))
            return;

        if(event != NULL)
        {
            if( SnortSnprintf(event_data, STD_BUF, "[%lu:%lu:%lu] ", 
                              (unsigned long) event->sig_generator,
                              (unsigned long) event->sig_id, 
                              (unsigned long) event->sig_rev) != SNORT_SNPRINTF_SUCCESS )
                return ;

            if(  strlcat(event_string, event_data, SYSLOG_BUF) >= SYSLOG_BUF)
                return ;
        }

        if(msg != NULL)
        {
           if( strlcat(event_string, msg, SYSLOG_BUF) >= SYSLOG_BUF )
                return ;
        }
        else
        {
           if(strlcat(event_string, "ALERT", SYSLOG_BUF) >= SYSLOG_BUF)
                return ;
        }

        if(otn_tmp != NULL)
        {
            if(otn_tmp->sigInfo.classType)
            {
                if( otn_tmp->sigInfo.classType->name )
                {
                    if( SnortSnprintf(pri_data, STD_BUF-1, " [Classification: %s] "
                                      "[Priority: %d]:", 
                                      otn_tmp->sigInfo.classType->name,
                                      otn_tmp->sigInfo.priority) != SNORT_SNPRINTF_SUCCESS )
                        return ;
                }
                if( strlcat(event_string, pri_data, SYSLOG_BUF) >= SYSLOG_BUF)
                    return ;
            }
            else if(otn_tmp->sigInfo.priority != 0)
            {
                if( SnortSnprintf(pri_data, STD_BUF, "[Priority: %d]:", 
                                  otn_tmp->sigInfo.priority) != SNORT_SNPRINTF_SUCCESS )
                   return ;

                if( strlcat(event_string, pri_data, SYSLOG_BUF) >= SYSLOG_BUF)
                    return;
            }
        }

        if((GET_IPH_PROTO(p) != IPPROTO_TCP &&
                    GET_IPH_PROTO(p) != IPPROTO_UDP) || 
                p->frag_flag)
        {
            if(!pv.alert_interface_flag)
            {
                if( protocol_names[GET_IPH_PROTO(p)] )
                {
                    if( SnortSnprintf(ip_data, STD_BUF, " {%s} %s -> %s\n",  
                                      protocol_names[GET_IPH_PROTO(p)],
                                      sip, dip) != SNORT_SNPRINTF_SUCCESS )
                        return;
                }
            }
            else
            {
                if( protocol_names[GET_IPH_PROTO(p)] && PRINT_INTERFACE(pv.interface) )
                {
                    if( SnortSnprintf(ip_data, STD_BUF, " <%s> {%s} %s -> %s\n",  
                                      PRINT_INTERFACE(pv.interface), 
                                      protocol_names[GET_IPH_PROTO(p)],
                                      sip, dip) != SNORT_SNPRINTF_SUCCESS )
                        return ;
                }
            }
        }
        else
        {
            if(pv.alert_interface_flag)
            {
               if( protocol_names[GET_IPH_PROTO(p)] && PRINT_INTERFACE(pv.interface) )
               {
                   if( SnortSnprintf(ip_data, STD_BUF, " <%s> {%s} %s:%i -> %s:%i\n",
                                     PRINT_INTERFACE(pv.interface), 
                                     protocol_names[GET_IPH_PROTO(p)], sip,
                                     p->sp, dip, p->dp) != SNORT_SNPRINTF_SUCCESS )
                       return ;
               }
            }
            else
            {
               if( protocol_names[GET_IPH_PROTO(p)] )
               {
                   if( SnortSnprintf(ip_data, STD_BUF, " {%s} %s:%i -> %s:%i\n",
                                     protocol_names[GET_IPH_PROTO(p)], sip, p->sp, 
                                     dip, p->dp) != SNORT_SNPRINTF_SUCCESS )
                       return ;
               }
            }
        }

        if( strlcat(event_string, ip_data, SYSLOG_BUF) >= SYSLOG_BUF)
            return;

        sendMsg(data, event_string);

    }
    else  
    {
        sendMsg(data , msg == NULL ? "ALERT!\n" : msg);
    }

    return;

}


void AlertTeamf1CleanExit(int signal, void *arg)
{
    Teamf1Data *data = (Teamf1Data *)arg;
    DEBUG_WRAP(DebugMessage(DEBUG_LOG, "AlertTeamf1CleanExit\n"););
    /* free memory from Teamf1Data */
    if(data)
        free(data);
}

void AlertTeamf1Restart(int signal, void *arg)
{
    Teamf1Data *data = (Teamf1Data *)arg;
    DEBUG_WRAP(DebugMessage(DEBUG_LOG, "AlertTeamf1RestartFunc\n"););
    /* free memory from Teamf1Data */
    if(data)
        free(data);
}
